//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "ExportSettings.h"
#include "Utilities.h"
#include "PropertyKeys.h"

using namespace OgreMax;


//Implementation---------------------------------------------------------------
ExportSettings::ExportSettings() 
{
	this->exportMeshes = true;  
	this->exportMaterials = true;
    this->physicsExportType = NO_PHYSICS;
    this->exportVertexColors = false;
    this->upAxis = UP_AXIS_Y; //Y up is more in line with the way people usually use Ogre
    this->scale = GetMasterScale(UNITS_METERS);
    this->scaledWorldUnitsPerMeter = 1.0f;
    this->texCoord = UV;
    this->noMaterialName = _T("NoMaterial");
    this->sortBoneAssignments = true;
	
    this->exportTextMesh = true;
    this->exportBinaryMesh = true;	    
	this->generateTangents = true;
    this->generateEdgeLists = false;
	this->generateLod = false;    
    this->numLodLevels = 0;
	this->lodDistance = 500.0f;	
    this->lodReductionType = LOD_PERCENT;
	this->lodPercentReduction = 20.0f;
	this->lodVertexReduction = 0;    
    this->vertexColorFormat = DIRECT3D;
    this->endian = Ogre::Serializer::ENDIAN_NATIVE;	

    //Remaining settings should be set manually
}

ExportSettings::~ExportSettings()
{
}

void ExportSettings::Initialize(const KeyedValues& properties)
{
    TSTR value;

    //General Exporter ExportSettings------------------------------------------------

    //Mesh file export enabled?
    if (properties.GetValue(PropertyKeys::EXPORT_MESHES, value))
        this->exportMeshes = Utilities::ToBool(value);
    
	//Enable material file export?
    if (properties.GetValue(PropertyKeys::EXPORT_MATERIALS, value))
	    this->exportMaterials = Utilities::ToBool(value);

    //Physics export type
    if (properties.GetValue(PropertyKeys::PHYSICS_EXPORT_TYPE, value))
    {
        if (Utilities::IsSameNoCase(value, _T("none")))
            this->physicsExportType = NO_PHYSICS;
        if (Utilities::IsSameNoCase(value, _T("opal")))
            this->physicsExportType = OPAL_PHYSICS;
    }

	//Enable vertex color export?
    if (properties.GetValue(PropertyKeys::EXPORT_VERTEX_COLORS, value))
	    this->exportVertexColors = Utilities::ToBool(value);	

    //The up axis of the exported file
    if (properties.GetValue(PropertyKeys::UP_AXIS, value))
    {
        if (Utilities::IsSameNoCase(value, _T("z")))	        
            this->upAxis = UP_AXIS_Z;
        else
            this->upAxis = UP_AXIS_Y;
    }

    //Set scale to use when exporting
    if (properties.GetValue(PropertyKeys::SCALE, value))
    {
        //Get extra scale if specified. The extra scale will be specified as the 
        //denominator in a fraction, such as "meters/1000"
        float extraScale = 1;
        int divideIndex = Utilities::FindDivide(value);        
        if (divideIndex != -1)
        {
            //Split the value into its two parts
            TSTR units = value.Substr(0, divideIndex);
            Utilities::TrimWhitespace(units);

            TSTR extraScaleDenominatorText = value.Substr(divideIndex + 1, value.length() - divideIndex - 1);
            Utilities::TrimWhitespace(extraScaleDenominatorText);
            
            //Parse the extra scale and use it if it isn't zero
            float extraScaleDenominator = atof(extraScaleDenominatorText);
            if (extraScaleDenominator != 0)
                extraScale = 1 / extraScaleDenominator;

            //Put the units back into "value"
            value = units;
        }
        
        //Get regular scale
        if (Utilities::IsSameNoCase(value, _T("active")))
            this->scale = GetMasterScale(GetUnitDisplayType());
        else if (Utilities::IsSameNoCase(value, _T("meters"))||Utilities::IsSameNoCase(value, _T("")))
            this->scale = GetMasterScale(UNITS_METERS);
        else if (Utilities::IsSameNoCase(value, _T("kilometers")))
            this->scale = GetMasterScale(UNITS_KILOMETERS);
        else if (Utilities::IsSameNoCase(value, _T("feet")))
            this->scale = GetMasterScale(UNITS_FEET);
        else if (Utilities::IsSameNoCase(value, _T("miles")))
            this->scale = GetMasterScale(UNITS_MILES);
        else
        {
            this->scale = Utilities::ToFloat(value);
            if (this->scale == 0)
                this->scale = 1;
        }

        //Multiply in the extra scaling factor
        this->scale *= extraScale;                
        this->scaledWorldUnitsPerMeter = this->scale / GetMasterScale(UNITS_METERS);
    }
    
	//Set the interpretation of texture coordinates
    if (properties.GetValue(PropertyKeys::TEX_COORD_2D, value))
    {
        if (Utilities::IsSameNoCase(value, _T("VW")))
            this->texCoord = VW;
        else if (Utilities::IsSameNoCase(value, _T("WU")))
            this->texCoord = WU;
        else
            this->texCoord = UV;
    }
	
    //Default material name to use when none is provided for a mesh
    if (properties.GetValue(PropertyKeys::NO_MATERIAL_NAME, value))
	    this->noMaterialName = value;

    //Sort bone assignments?
    if (properties.GetValue(PropertyKeys::SORT_BONE_ASSIGNMENTS, value))
        this->sortBoneAssignments = Utilities::ToBool(value);

    
	//Binary mesh export settings----------------------------------------------

    //Enable mesh text file export?
    if (properties.GetValue(PropertyKeys::EXPORT_TEXT_MESH, value))
	    this->exportTextMesh = Utilities::ToBool(value);

    //Enable mesh binary file export?
    if (properties.GetValue(PropertyKeys::EXPORT_BINARY_MESH, value))
	    this->exportBinaryMesh = Utilities::ToBool(value);

    //Tangent vector generation
    if (properties.GetValue(PropertyKeys::GENERATE_TANGENTS, value))
	    this->generateTangents = Utilities::ToBool(value);	
	
    //Edge lists
    if (properties.GetValue(PropertyKeys::GENERATE_EDGE_LISTS, value))
        this->generateEdgeLists = Utilities::ToBool(value);	

	//Generate LoD?
    if (properties.GetValue(PropertyKeys::GENERATE_LOD, value))
        this->generateLod = Utilities::ToBool(value);

    //Number of LoD levels
    if (properties.GetValue(PropertyKeys::NUM_LOD_LEVELS, value))
	    this->numLodLevels = Utilities::ToUInt(value);	
	
	//LoD distance
    if (properties.GetValue(PropertyKeys::LOD_DISTANCE, value))
	    this->lodDistance = Utilities::ToFloat(value);	
	
    //Use percent reduction?
    if (properties.GetValue(PropertyKeys::LOD_REDUCTION_TYPE, value))
    {
        if (Utilities::IsSameNoCase(value, _T("vertex")))
            this->lodReductionType = LOD_VERTEX;
        else
            this->lodReductionType = LOD_PERCENT;
    }

	//LoD percent reduction
    if (this->lodReductionType == LOD_PERCENT)
    {
        if (properties.GetValue(PropertyKeys::LOD_PERCENT_REDUCTION, value) ||
            properties.GetValue(PropertyKeys::LOD_REDUCTION_AMOUNT, value))
        {
	        this->lodPercentReduction = Utilities::ToFloat(value);
        }
    }

	//LoD vertex reduction
    if (this->lodReductionType == LOD_VERTEX)
    {
        if (properties.GetValue(PropertyKeys::LOD_VERTEX_REDUCTION, value) ||
            properties.GetValue(PropertyKeys::LOD_REDUCTION_AMOUNT, value))
        {
	        this->lodVertexReduction = Utilities::ToInt(value);
        }
    }

    //vertex colour format
    if (properties.GetValue(PropertyKeys::VERTEX_COLOR_FORMAT, value))
    {
        if (Utilities::IsSameNoCase(value, _T("OPENGL")))
            this->vertexColorFormat = OPENGL;
        else
            this->vertexColorFormat = DIRECT3D;
    }

	//endianness
    if (properties.GetValue(PropertyKeys::ENDIAN, value))
    {
        if (Utilities::IsSameNoCase(value, _T("BIG")))
            this->endian = Ogre::Serializer::ENDIAN_BIG;
        else if (Utilities::IsSameNoCase(value, _T("LITTLE")))
            this->endian = Ogre::Serializer::ENDIAN_LITTLE;
        else
            this->endian = Ogre::Serializer::ENDIAN_NATIVE;
    }
	
	//Paths--------------------------------------------------------------------
    
    //Export path
    if (properties.GetValue(PropertyKeys::MESH_PATH, value))
    {
	    this->exportPath = value;
        Utilities::EnsureTrailingPathSeparator(this->exportPath);
    }

    //Mesh export path
    if (properties.GetValue(PropertyKeys::MESH_PATH, value))
    {
	    this->meshPath = value;
        Utilities::EnsureTrailingPathSeparator(this->meshPath);
    }

    //Material export path
    if (properties.GetValue(PropertyKeys::MATERIAL_PATH, value))
    {
	    this->materialPath = value;
        Utilities::EnsureTrailingPathSeparator(this->materialPath);
    }

    //Model export path
    if (properties.GetValue(PropertyKeys::MODEL_PATH, value))
    {
	    this->modelPath = value;
        Utilities::EnsureTrailingPathSeparator(this->modelPath);
    }

    //Log file path
    if (properties.GetValue(PropertyKeys::LOG_PATH, value))
    {
	    this->logPath = value;
        Utilities::EnsureTrailingPathSeparator(this->logPath);
    }

    //Laterial export filename
    if (properties.GetValue(PropertyKeys::MATERIAL_FILE_NAME, value))
	    this->materialFileName = value;        
}

void ExportSettings::MakeExportPaths(const TSTR& exportPath, const TSTR& exportFileName)
{
    if (this->exportPath.isNull())
        this->exportPath = exportPath;
    
    this->meshPath = Utilities::CreateExportDirectory(this->exportPath, this->meshPath);
    this->materialPath = Utilities::CreateExportDirectory(this->exportPath, this->materialPath);
    this->modelPath = Utilities::CreateExportDirectory(this->exportPath, this->modelPath);
    this->logPath = Utilities::CreateExportDirectory(this->exportPath, _T("exportlogs\\"));    

    //Set material file name
    if (this->materialFileName.isNull())
    {
        this->materialFileName = this->materialPath;
        this->materialFileName += Utilities::GetFileNameNoExtension(exportFileName);
        this->materialFileName = Utilities::MakeMaterialFileName(this->materialFileName);
    }
    else
    {
        this->materialFileName = Utilities::JoinPath(this->materialPath, this->materialFileName);
        Utilities::CreateFileDirectoryRecursive(this->materialFileName);
    }
}