
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "MeshAnimationSettingsDialog.h"
#include "RemoveBoneTranslationsDialog.h"
#include "Utilities.h"
#include "ObjectSettings.h"
#include "CommonSettings.h"
#include "PluginGlobals.h"
#include "PropertyValues.h"
#include "StringSplitter.h"
#include "IncludeExcludeBoneAnimationDialog.h"
#include "resource.h"

using namespace OgreMax;


//Globals----------------------------------------------------------------------
static const int timeScaleComboStringIDs[] = {IDS_NONE, IDS_SCALE, IDS_LENGTH, IDS_NORMALIZE};

static const int sampleTypeComboStringIDs[] = {IDS_USE_OBJECT_SETTING, IDS_SECONDS_BETWEEN_SAMPLES, IDS_SAMPLES_PER_SECOND};
enum {NO_SAMPLE_RATE_INDEX, SAMPLE_INTERVAL_INDEX, SAMPLE_RATE_INDEX, SAMPLE_TYPE_COUNT};


//Local functions--------------------------------------------------------------
static const TSTR& ChooseTimeScaleValue(const TSTR& scale, const TSTR& length)
{
    if (!Utilities::IsNull(scale))
        return scale;
    else if (!Utilities::IsNull(length))
        return length;

    //It doesn't matter what we return since everything's empty
    return scale;
}

static const TSTR& ChooseAnimationSampleValue
    (
    const TSTR& intervalValue, 
    const TSTR& rateValue
    )
{
    if (!Utilities::IsNull(intervalValue))
        return intervalValue;
    else if (!Utilities::IsNull(rateValue))
        return rateValue;
    
    //It doesn't matter what we return since everything's empty
    return intervalValue;
}

static int GetTimeScaleTypeValueComboStringIndex(const TSTR& scale, const TSTR& length)
{
    if (Utilities::IsSameNoCase(scale, PropertyValues::TimeScale::NORMALIZE) || 
        Utilities::IsSameNoCase(length, PropertyValues::TimeScale::NORMALIZE))
        return 3;
    else if (!Utilities::IsNull(scale))
        return 1;
    else if (!Utilities::IsNull(length))
        return 2;
    else
        return 0;
}

static int GetSampleTypeValueComboStringIndex(const TSTR& intervalValue, const TSTR& rateValue)
{
    if (Utilities::IsNull(intervalValue) && 
        Utilities::IsNull(rateValue))
    {
        return NO_SAMPLE_RATE_INDEX;
    }
    else if (!Utilities::IsNull(intervalValue))
        return SAMPLE_INTERVAL_INDEX;
    else
        return SAMPLE_RATE_INDEX;
}

static void SetSampleValue(TSTR& sampleInterval, TSTR& sampleRate, int sampleTypeIndex, const TSTR& sampleValue)
{
    sampleInterval = sampleRate = _T("");

    if (sampleTypeIndex == SAMPLE_INTERVAL_INDEX)
        sampleInterval = sampleValue;
    else if (sampleTypeIndex == SAMPLE_RATE_INDEX)
        sampleRate = sampleValue;
}

static void SetTimeScale(TSTR& scale, TSTR& length, int timeScaleIndex, const TSTR& value)
{
    scale = length = _T("");
    if (timeScaleIndex == 1)
        scale = value;
    else if (timeScaleIndex == 2)
        length = value;
    else if (timeScaleIndex == 3)
        scale = PropertyValues::TimeScale::NORMALIZE;
}

static bool TimeScaleComboIndexRequiresTimeValue(int index)
{
    switch (index)
    {
        case 0: return false;
        case 1: return true;
        case 2: return true;
        case 3: return false;
    }

    return false;
}

static bool SamplingComboIndexRequiresTimeValue(int index)
{
    return index != 0;
}


//Implementation---------------------------------------------------------------
MeshAnimationSettingsDialog::MeshAnimationSettingsDialog(MeshAnimationSettings& animSettings) : settings(animSettings)
{
}

MeshAnimationSettingsDialog::~MeshAnimationSettingsDialog()
{
}

int MeshAnimationSettingsDialog::DoModal(HWND hWndParent)
{
    return DialogBase::DoModal(IDD_MESH_ANIMATION_SETTINGS_DIALOG, hWndParent);
}

static const TSTR BONE_KEY(_T("bone"));
static const TSTR BONE_TREE_KEY(_T("boneTree"));
static const TSTR INCLUDE_TEXT(_T("include"));
static const TSTR EXCLUDE_TEXT(_T("exclude"));

void MeshAnimationSettingsDialog::GetGUIData()
{
    //Animation name edit
    GetDlgItemText(IDC_MESH_ANIMATION_NAME_EDIT, this->settings.name);

    //Time type combo
    HWND timeScaleComboHwnd = GetDlgItem(IDC_TIME_SCALE_COMBO);
    int selectedTimeScaleIndex = ComboBox_GetCurSel(timeScaleComboHwnd);
    
    //Time edit
    TSTR timeValue;
    GetDlgItemText(IDC_TIME_EDIT, timeValue);
    SetTimeScale(this->settings.scale, this->settings.length, selectedTimeScaleIndex, timeValue);

    //Sample type combo
    HWND samplingComboHwnd = GetDlgItem(IDC_SAMPLING_TYPE_COMBO);
    int selectedSamplingTypeIndex = ComboBox_GetCurSel(samplingComboHwnd);

    //Sampling edit
    TSTR sampling;
    GetDlgItemText(IDC_SAMPLING_EDIT, sampling);    
    SetSampleValue(this->settings.sampleInterval, this->settings.sampleRate, selectedSamplingTypeIndex, sampling);

    //Copy-first-key-to-last check
    this->settings.copyFirstKeyToLast = IsDlgButtonCheckedText(IDC_COPY_FIRST_KEY_TO_LAST_CHECK);
    
    //Include/exclude bones edit
    TSTR exportBonesText;
    GetDlgItemText(IDC_EXPORT_BONES_EDIT, exportBonesText);
    StringSplitter exportBones(exportBonesText, _T('\n'));
    this->settings.exportBones.clear();
    this->settings.exportBones.reserve(exportBones.size());
    BoneAllow allowBone;
    TSTR bonePrefixKey;
    for (size_t i = 0; i < exportBones.size(); i++)
    {
        if (Utilities::ParsePrefixedKeyValueLine(exportBones[i], allowBone.name, allowBone.allow, &bonePrefixKey))
        {
            allowBone.allow = Utilities::IsSameNoCase(allowBone.allow, INCLUDE_TEXT) ? PropertyValues::TRUE_TEXT : PropertyValues::FALSE_TEXT;
        
            if (Utilities::IsSameNoCase(bonePrefixKey, BONE_KEY))
            {
                allowBone.tree = PropertyValues::FALSE_TEXT;
                this->settings.exportBones.push_back(allowBone);
            }
            else if (Utilities::IsSameNoCase(bonePrefixKey, BONE_TREE_KEY))
            {
                allowBone.tree = PropertyValues::TRUE_TEXT;
                this->settings.exportBones.push_back(allowBone);
            }
        }
    }
}

void MeshAnimationSettingsDialog::SetGUIData()
{
    //Animation name edit
    SetDlgItemText(IDC_MESH_ANIMATION_NAME_EDIT, this->settings.name);

    //Time type combo
    HWND timeScaleComboHwnd = GetDlgItem(IDC_TIME_SCALE_COMBO);
    for (int i = 0; i < COUNT_OF(timeScaleComboStringIDs); i++)
        ComboBox_AddString(timeScaleComboHwnd, PluginGlobals::GetString(timeScaleComboStringIDs[i]));
    ComboBox_SetCurSel(timeScaleComboHwnd, GetTimeScaleTypeValueComboStringIndex(this->settings.scale, this->settings.length));
    
    //Time edit
    SetDlgItemText(IDC_TIME_EDIT, ChooseTimeScaleValue(this->settings.scale, this->settings.length));

    //Sample type combo
    HWND samplingComboHwnd = GetDlgItem(IDC_SAMPLING_TYPE_COMBO);
    for (int i = 0; i < SAMPLE_TYPE_COUNT; i++)
        ComboBox_AddString(samplingComboHwnd, PluginGlobals::GetString(sampleTypeComboStringIDs[i]));
    ComboBox_SetCurSel(samplingComboHwnd, GetSampleTypeValueComboStringIndex(this->settings.sampleInterval, this->settings.sampleRate));    

    //Sampling edit
    SetDlgItemText(IDC_SAMPLING_EDIT, ChooseAnimationSampleValue(this->settings.sampleInterval, this->settings.sampleRate));

    //Copy-first-key-to-last check
    CheckDlgButton(IDC_COPY_FIRST_KEY_TO_LAST_CHECK, ObjectSettings::Values::CopyFirstKeyToLast(this->settings.copyFirstKeyToLast));
    
    //Remove bone translations edit
    TSTR removeBoneTranslationsText;
    CommonSettings::Write(removeBoneTranslationsText, this->settings.removeBoneTranslations, PluginGlobals::CONTROL_NEWLINE);
    SetDlgItemText(IDC_REMOVE_BONE_TRANSLATIONS_EDIT, removeBoneTranslationsText);    

    //Exclude/include bones edit
    TSTR exportBonesText;
    TSTR allow;
    for (size_t i = 0; i < this->settings.exportBones.size(); i++)
    {
        allow = Utilities::ToBool(this->settings.exportBones[i].allow) ? INCLUDE_TEXT : EXCLUDE_TEXT;
        if (Utilities::ToBool(this->settings.exportBones[i].tree))
            Utilities::AppendPrefixedKeyAndValue(exportBonesText, BONE_TREE_KEY, this->settings.exportBones[i].name, allow, PluginGlobals::CONTROL_NEWLINE);
        else
            Utilities::AppendPrefixedKeyAndValue(exportBonesText, BONE_KEY, this->settings.exportBones[i].name, allow, PluginGlobals::CONTROL_NEWLINE);
    }
    SetDlgItemText(IDC_EXPORT_BONES_EDIT, exportBonesText);    

    //Final updates
    UpdateTimeScaleEdit();
    UpdateSamplingTypeEdit();
}

void MeshAnimationSettingsDialog::UpdateTimeScaleEdit()
{
    HWND timeScaleComboHwnd = GetDlgItem(IDC_TIME_SCALE_COMBO);
    int selectedTimeScaleIndex = ComboBox_GetCurSel(timeScaleComboHwnd);
    bool show = TimeScaleComboIndexRequiresTimeValue(selectedTimeScaleIndex);
    ShowWindow(GetDlgItem(IDC_TIME_EDIT), show ? TRUE : FALSE);
}

void MeshAnimationSettingsDialog::UpdateSamplingTypeEdit()
{
    HWND samplingComboHwnd = GetDlgItem(IDC_SAMPLING_TYPE_COMBO);
    int selectedSamplingTypeIndex = ComboBox_GetCurSel(samplingComboHwnd);
    bool show = SamplingComboIndexRequiresTimeValue(selectedSamplingTypeIndex);
    ShowWindow(GetDlgItem(IDC_SAMPLING_EDIT), show ? TRUE : FALSE);
}

INT_PTR MeshAnimationSettingsDialog::DialogProc(UINT message, WPARAM wParam, LPARAM lParam, bool& handled)
{
    switch (message)
    {
        case WM_INITDIALOG:
        {
            //Set window title
            AppendWindowText(_T(" - "), this->settings.name);
            
            //Center in parent window
            CenterWindow();

            SetGUIData();

            handled = true;

            break;
        }
        case WM_COMMAND:
        {
            int id = LOWORD(wParam);
            int code = HIWORD(wParam);

            switch (id)
            {
                case IDOK:
                {
                    GetGUIData();

                    handled = true;
                    EndDialog(this->hDlg, TRUE);
                    return TRUE;
                }
                case IDCANCEL:
                {
                    handled = true;
                    EndDialog(this->hDlg, FALSE);
                    return FALSE;
                }
                case IDC_REMOVE_BONE_TRANSLATIONS_ADD_BUTTON:
                {
                    switch (code)
                    {
                        case BN_CLICKED:
                        {
                            RemoveBoneTranslationsDialog dialog;
                            if (dialog.DoModal(this->hDlg))
                            {
                                TSTR removeBoneTranslationsText;
                                GetDlgItemText(IDC_REMOVE_BONE_TRANSLATIONS_EDIT, removeBoneTranslationsText);
                                Utilities::AppendKeyAndValue(removeBoneTranslationsText, dialog.boneName, dialog.xyz, PluginGlobals::CONTROL_NEWLINE);
                                SetDlgItemText(IDC_REMOVE_BONE_TRANSLATIONS_EDIT, removeBoneTranslationsText);
                            }

                            handled = true;
                            break;
                        }
                    }
                    break;
                }
                case IDC_TIME_SCALE_COMBO:
                {
                    switch (code)
                    {
                        case CBN_SELCHANGE:
                        {
                            UpdateTimeScaleEdit();
                            handled = true;                            
                            break;
                        }
                    }
                    break;
                }
                case IDC_SAMPLING_TYPE_COMBO:
                {
                    switch (code)
                    {
                        case CBN_SELCHANGE:
                        {
                            UpdateSamplingTypeEdit();
                            handled = true;                            
                            break;
                        }
                    }
                    break;
                }
                case IDC_ADD_BONE_BUTTON:
                {
                    switch (code)
                    {
                        case BN_CLICKED:
                        {
                            IncludeExcludeBoneAnimationDialog dialog;
                            if (dialog.DoModal(this->hDlg))
                            {
                                const TSTR& prefix = dialog.isTree ? BONE_TREE_KEY : BONE_KEY;
                                const TSTR& allow = dialog.include ? INCLUDE_TEXT : EXCLUDE_TEXT;
                                
                                TSTR exportBonesText;
                                GetDlgItemText(IDC_EXPORT_BONES_EDIT, exportBonesText);    
                                Utilities::AppendPrefixedKeyAndValue(exportBonesText, prefix, dialog.boneName, allow, PluginGlobals::CONTROL_NEWLINE);
                                SetDlgItemText(IDC_EXPORT_BONES_EDIT, exportBonesText);    
                            }

                            handled = true;
                            break;
                        }
                    }
                    break;
                }                
            }

            break;
        }        
    }

    return 0;
}