
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "MeshExporter.h"
#include "MaterialExporter.h"
#include "OgreXmlMeshConverter.h"
#include "OgreMaxSceneExporter.h"
#include "Vertex.h"
#include "Utilities.h"
#include "PluginGlobals.h"
#include "PropertyKeys.h"
#include "MeshAnimationSettingsSerializer.h"
#include "resource.h"
#include<direct.h>
using namespace OgreMax;

#define DEFAULT_VERTEX_ANIMATION_INTERVAL .3333f


//Implementation---------------------------------------------------------------
MeshExporter::MeshExporter
    (
    const KeyedValues& sceneProps,
    const ExportSettings& settings, 
    const InitialNodeStates& initNodeStates,
    IGameScene* gameInterface,
    Interface* maxInterface
    ) : 
    ExporterBase(settings, gameInterface, maxInterface),
    sceneProperties(sceneProps),
    initialNodeStates(initNodeStates)
{	
}

MeshExporter::~MeshExporter()
{
}
std::string modelDir(std::string input) {
	std::string::size_type where=input.find_last_of("\\/");
	if (where!=std::string::npos) {
		std::string tmp=input.substr(0,where)+"/models";
		_mkdir(tmp.c_str());
		input=tmp+input.substr(where);
	}else{ 
		_mkdir ("models");
		input="models\\"+input;
	}
	return input;
}
bool MeshExporter::ExportMeshes(const std::list<GameNodeAndObject>& meshNodes, MaterialMap& materials)
{
    std::list<ExportedMesh> meshes;
    if (ExportMeshes(meshNodes, meshes, materials)) 
    {
	    //Export successful

        //Write each mesh file
        for (std::list<ExportedMesh>::iterator meshItem = meshes.begin(); 
            meshItem != meshes.end(); 
            ++meshItem)
        {
            //Simplified access
            const ExportedMesh& exportedMesh = *meshItem;

            //Write mesh XML
            stdofstream meshOut; 
			std::string fn=modelDir(exportedMesh.meshFileName.data());
			std::string fn2;
		    meshOut.open(fn.c_str(), std::ios::out);
		    meshOut << exportedMesh.meshXml.data();
		    meshOut.close();

            //Write mesh skeleton XML
            if (exportedMesh.HasSkeleton())
            {
                stdofstream skeletonOut;  
				fn2=modelDir(exportedMesh.skeletonFileName.data());
		        skeletonOut.open(fn2.c_str(), std::ios::out);
		        skeletonOut << exportedMesh.skeletonXml.data();
		        skeletonOut.close();
            }

            //Convert XML files to binary files
            if (this->exportSettings.exportBinaryMesh)
            {
                if (exportedMesh.HasSkeleton())
                    ConvertXmlSkeletonToBinary(exportedMesh);
                ConvertXmlMeshToBinary(exportedMesh);
            }

            //Remove XML files if necessary
            if (!this->exportSettings.exportTextMesh)
            {
                if (exportedMesh.HasSkeleton())
                    DeleteFile(fn2.c_str());
                DeleteFile(fn.c_str());
            }
	    }

        return true;
    }
    else
        return false;
}

bool MeshExporter::ExportMeshes
    (
    const std::list<GameNodeAndObject>& meshNodes, 
    std::list<ExportedMesh>& meshes, 
    MaterialMap& materials
    )
{
    int nodeCount = (int)meshNodes.size();
	if (nodeCount == 0) 
    	return true;

    try 
    {
        stdstringstream meshOut;
        stdstringstream skeletonOut;

        //Write the meshes                
        for (std::list<GameNodeAndObject>::const_iterator meshNodeIterator = meshNodes.begin(); 
            meshNodeIterator != meshNodes.end(); 
            ++meshNodeIterator)
		{               
            IGameNode* meshNode = meshNodeIterator->gameNode;

            InitialNodeState* initialNodeState = this->initialNodeStates.find(meshNode->GetMaxNode())->second;
            const KeyedValues& meshNodeProperties = initialNodeState->properties;
            
            TSTR vertexAnimationType;
            meshNodeProperties.GetValue(PropertyKeys::VERTEX_ANIMATION_TYPE, vertexAnimationType);

            float vertexAnimationSampleInterval = 0;
            float skeletonAnimationSampleInterval = 0;
            Utilities::GetMeshAnimationSampleIntervals(meshNodeProperties, vertexAnimationSampleInterval, skeletonAnimationSampleInterval);
            
            //Determine if the faces need to be flipped. This will only be true on
            //a node that has been mirrored
            bool flipFaces = Utilities::IsMirrored(meshNode);
            
            //Make sure mesh can be exported
            TSTR meshName = Utilities::GetMeshName(meshNode, initialNodeState->properties);            
            bool canExportMesh = 
                Utilities::CanExportMesh(meshName, this->sceneProperties) &&
                this->exportedMeshNames.find(meshName.data()) == this->exportedMeshNames.end();
            if (!canExportMesh)
                continue;

            //Get various objects from the node
			IGameMaterial* material = meshNode->GetNodeMaterial();                
            IGameMesh* gameMesh = (IGameMesh*)meshNodeIterator->gameObject;
            bool morphOnly = Utilities::IsSameNoCase(vertexAnimationType, _T("morphOnly"));
            bool useMorphVertexAnimation = morphOnly || Utilities::IsSameNoCase(vertexAnimationType, _T("morph"));
            IGameSkin* gameSkin = morphOnly ? NULL : gameMesh->GetIGameSkin();
                        
            //Write mesh header to output
            WriteFileHeader(meshOut);

            //Write submeshes for material
            int submeshIndex = 0;
            std::list<Submesh> submeshes;
            if (material && material->IsMultiType())
            {
                //Material has submaterials
                for (int submaterialIndex = 0; 
                    submaterialIndex < material->GetSubMaterialCount(); 
                    submaterialIndex++)
                {
                    IGameMaterial* submaterial = material->GetSubMaterial(submaterialIndex);
                    
                    //Add material to output materials
                    TSTR submaterialName = Utilities::MakeMaterialName(submaterial, this->exportSettings.noMaterialName);
                    if (Utilities::CanExportMaterial(submaterialName, this->sceneProperties) && 
                        materials.find(submaterialName.data()) == materials.end()) 
                    {
                        materials[submaterialName.data()] = submaterial;				
                    }

                    //Write the submesh to output
                    Submesh submesh;
                    if (WriteSubmesh(meshOut, meshNode, gameMesh, gameSkin, submaterial, submaterialName, flipFaces, submesh.reindexedVertices, submesh.exportedVertexIndices))
                    {                        
                        submesh.name = meshName + _T("_") + Utilities::ToString(submaterialIndex);
                        submesh.index = submeshIndex++;
                        submeshes.push_back(submesh);
                    }
                }
            }
            else
            {
                //No material or a standard material

                //Add material to output materials
                TSTR materialName = Utilities::MakeMaterialName(material, this->exportSettings.noMaterialName);
                if (Utilities::CanExportMaterial(materialName, this->sceneProperties) && 
                    materials.find(materialName.data()) == materials.end()) 
                {
                    materials[materialName.data()] = material;				
                }

                //Write the submesh to output
                Submesh submesh;
                if (WriteSubmesh(meshOut, meshNode, gameMesh, gameSkin, material, materialName, flipFaces, submesh.reindexedVertices, submesh.exportedVertexIndices))
                {                    
                    submesh.name = meshName;
                    submesh.index = submeshIndex++;
                    submeshes.push_back(submesh);
                }
            }

            //Hold onto the mesh name so another mesh with the same name isn't exported later
            this->exportedMeshNames[meshName.data()] = meshName.data();
            
            //Handle skeleton
            TSTR skeletonFileName;
            TSTR skeletonXmlFileName;
            if (gameSkin != NULL)
            {
                //Get skeleton name
                TSTR skeletonName;
                if (!meshNodeProperties.GetValue(PropertyKeys::SKELETON_NAME, skeletonName))
                    skeletonName = meshName;

                //Create file names
                skeletonFileName = Utilities::MakeSkeletonFileName(skeletonName, false);
                skeletonXmlFileName = Utilities::MakeSkeletonFileName(skeletonName, true);

                //Write the skeleton if it hasn't already been written
                if (this->exportedSkeletonNames.find(skeletonName.data()) == this->exportedSkeletonNames.end())
                {
                    WriteSkeleton(skeletonOut, meshNode, gameSkin, skeletonAnimationSampleInterval);
                    
                    this->exportedSkeletonNames[skeletonName.data()] = skeletonName.data();
                }
            }

            //Finish writing mesh            
	        WriteFileFooter(meshOut, gameMesh, meshNode, useMorphVertexAnimation, morphOnly, skeletonFileName, vertexAnimationSampleInterval, submeshes);
            stdstring meshOutputString = meshOut.str();
            meshOut.str(_T("")); //Reset mesh output stream

            stdstring meshSkeletonOutputString = skeletonOut.str();
            skeletonOut.str(_T("")); //Reset skeleton output stream

	        //Add exported mesh to output list
            ExportedMesh exportedMesh;
            exportedMesh.name = meshName;
            exportedMesh.meshFileName = this->exportSettings.meshPath + Utilities::MakeMeshFileName(meshName, true);
            exportedMesh.meshXml = meshOutputString.c_str();
            exportedMesh.meshNode = meshNode;
            exportedMesh.skeletonFileName = skeletonXmlFileName;            
            exportedMesh.skeletonXml = meshSkeletonOutputString.c_str();
            exportedMesh.hasTextureCoords = gameMesh->GetNumberOfTexVerts() > 0;
            meshes.push_back(exportedMesh);
		}

		return true;
	}
    catch (Ogre::Exception e)
    {
        TSTR errorFormat = PluginGlobals::GetString(IDS_UNEXPECTED_EXPORT_ERROR_FORMAT);
        TSTR message;
        message.printf(errorFormat, e.getFullDescription().c_str());

        MessageBox
            (
            this->maxInterface->GetMAXHWnd(), 
            message,
            PluginGlobals::GetString2(IDS_EXPORT_ERROR_CAPTION),
            MB_OK | MB_ICONEXCLAMATION
            );

        return false;
    }
	catch (...) 
    {
		MessageBox
            (
            this->maxInterface->GetMAXHWnd(), 
            PluginGlobals::GetString(IDS_UNEXPECTED_EXPORT_ERROR),
            PluginGlobals::GetString2(IDS_EXPORT_ERROR_CAPTION),
            MB_OK | MB_ICONEXCLAMATION
            );

		return false;
	}
}

void MeshExporter::WriteFileHeader(stdostream& out) 
{
	//write the XML header tags
	out << _T("<?xml version=\"1.0\"?>") << std::endl;
	out << _T("<mesh>") << std::endl;

	//*************** Export Submeshes ***************
	out << _T("\t<submeshes>") << std::endl;

	out.precision(6);
	out << std::fixed;
}

void MeshExporter::WriteFileFooter
    (
    stdostream& out, 
    IGameMesh* gameMesh, 
    IGameNode* gameNode, 
    bool useMorphVertexAnimation,
    bool morphOnly,
    const TSTR& skeletonFileName, 
    float vertexAnimationSampleInterval,    
    std::list<Submesh>& submeshes
    ) 
{
	out << _T("\t</submeshes>") << std::endl;
	//*************** End Submeshes Export ***********

    //*************** Export Submesh Names ***************
	out << _T("\t<submeshnames>") << std::endl;

	for (std::list<Submesh>::iterator submesh = submeshes.begin(); submesh != submeshes.end(); ++submesh)
    	out << _T("\t\t<submeshname name=\"") << submesh->name.data() << _T("\" index=\"") << submesh->index << _T("\" />") << std::endl;
	
	out << _T("\t</submeshnames>") << std::endl;
	//*************** End Submesh Names Export ***********

    if (useMorphVertexAnimation)
    {
        if (vertexAnimationSampleInterval == 0)
             vertexAnimationSampleInterval = DEFAULT_VERTEX_ANIMATION_INTERVAL;

        WriteMorphAnimationsBySampling
            (
            out, 
            gameMesh, 
            gameNode,
            submeshes,
            morphOnly,
            vertexAnimationSampleInterval
            );        
    }
    else
    {
        IGameMorpher* gameMorpher = Utilities::GetMorpher(gameMesh);
        WriteMorphPosesAndAnimations
            (
            out, 
            gameMorpher, 
            gameMesh, 
            gameNode, 
            submeshes,
            vertexAnimationSampleInterval
            );
    
        //Skeleton link
        if (!Utilities::IsNull(skeletonFileName))
            out << _T("\t<skeletonlink name=\"") << skeletonFileName.data() << "\" />" << std::endl;
    }
    
	out << _T("</mesh>") << std::endl;
}

bool MeshExporter::WriteSubmesh
    (
    stdostream& out, 
    IGameNode* gameNode,
    IGameMesh* mesh, 
    IGameSkin* skin,
    IGameMaterial* gameMaterial, 
    TSTR& materialName,
    bool flipFaces,
    ReindexedVertices& reindexedVertices,
    std::vector<int>& exportedVertexIndices
    ) 
{	
    InitialNodeStates::const_iterator initialNodeStateIterator = this->initialNodeStates.find(gameNode->GetMaxNode());
    InitialNodeState* initialNodeState = initialNodeStateIterator->second;

    int faceCount = 0;
    
    VertexList vertexList;

    std::vector<int> vertexIndices;
    vertexIndices.reserve(initialNodeState->unmorphedInitialVertices.size() * 2);

    reindexedVertices.resize(initialNodeState->unmorphedInitialVertices.size());

    exportedVertexIndices.reserve(mesh->GetNumberOfFaces() * 2);

    bool simpleMesh = false;
    TSTR simpleMeshText;
    if (initialNodeState->properties.GetValue(PropertyKeys::SIMPLE_MESH, simpleMeshText))
        simpleMesh = Utilities::ToBool(simpleMeshText);

    int faceOrder[3] = {0, 1, 2};
    if (flipFaces)
        std::swap(faceOrder[1], faceOrder[2]);

    Tab<int> texMapChannels = mesh->GetActiveMapChannelNum();
    const int maxMappingChannels = 8;
    if (texMapChannels.Count() > maxMappingChannels)
        texMapChannels.Resize(maxMappingChannels);

    for (int faceIndex = 0; faceIndex < mesh->GetNumberOfFaces(); faceIndex++)
    {
        //Process face only if its material matches the one currently being
        //exported. This logic works whether or not there is a material
        IGameMaterial* faceMaterial = mesh->GetMaterialFromFace(faceIndex);
        if (faceMaterial == gameMaterial)
        {
            //Increment face count
            faceCount++;

            //Get face
            FaceEx* face = mesh->GetFace(faceIndex);

            //Get all the vertices for the face
	        for (int vertexIndex = 0; vertexIndex < 3; vertexIndex++) 
            {
                int vi = faceOrder[vertexIndex];
                Point3 p = initialNodeState->unmorphedInitialVertices[face->vert[vi]];
                
		        Vertex v(p.x, p.y, p.z);

		        if (this->exportSettings.exportVertexColors) 
                {
			        Point3 c = mesh->GetColorVertex(face->vert[vi]);
			        float a = mesh->GetAlphaVertex(face->vert[vi]);

			        v.SetColor(c.x, c.y, c.z, a);
		        }

                //Get normal
                Point3 n = initialNodeState->unmorphedInitialNormals.GetNormal(face->vert[vi], faceIndex, face->smGrp);
                v.SetNormal(n.x, n.y, n.z);
                
		        //Get each set of texcoords for this vertex
		        for (int texMapIndex = 0; texMapIndex < texMapChannels.Count(); texMapIndex++) 
                {
			        Point3 tv;
			        DWORD indices[3];

			        if (mesh->GetMapFaceIndex(texMapChannels[texMapIndex], faceIndex, indices))
				        tv = mesh->GetMapVertex(texMapChannels[texMapIndex], indices[vi]);
			        else
				        tv = mesh->GetMapVertex(texMapChannels[texMapIndex], face->vert[vi]);

			        v.AddTexCoord(texMapChannels[texMapIndex], tv.x, tv.y, tv.z);
		        }

                bool isNew;
		        int actualVertexIndex = vertexList.Add(v, &isNew, simpleMesh);
                vertexIndices.push_back(actualVertexIndex);

                if (isNew)
                    exportedVertexIndices.push_back(face->vert[vi]);

                ReindexedVertex& reindexedVertex = reindexedVertices[face->vert[vi]];
                if (std::find(reindexedVertex.begin(), reindexedVertex.end(), actualVertexIndex) == reindexedVertex.end())
                    reindexedVertex.push_back(actualVertexIndex);
	        }
        }
    }

    if (faceCount > 0)
    {
        int vertCount = (int)vertexList.GetSize();

        out << _T("\t\t<submesh ");
    	
	    if (materialName.length() > 0)
		    out << _T("material=\"") << materialName << _T("\" ");

	    out << _T("usesharedvertices=\"false\" use32bitindexes=\"") << (vertCount > 65535 ? _T("true") : _T("false")) << _T("\" operationtype=\"triangle_list\">") << std::endl;

	    //*************** Export Face List ***************
	    out << _T("\t\t\t<faces count=\"") << faceCount << _T("\">") << std::endl;
        for (int faceIndex = 0; faceIndex < faceCount; faceIndex++) 
        {
		    out << _T("\t\t\t\t<face");
    		
		    for (int vertexIndex = 0; vertexIndex < 3; vertexIndex++) 
            {
			    int actualVertexIndex = vertexIndices[faceIndex * 3 + vertexIndex];
			    out << _T(" v") << vertexIndex + 1 << _T("=\"") << actualVertexIndex << _T("\"");
		    }

		    out << _T(" />") << std::endl;
	    }

	    out << _T("\t\t\t</faces>") << std::endl;
	    //*************** End Export Face List ***************


	    //*************** Export Geometry ***************
	    out << _T("\t\t\t<geometry vertexcount=\"") << vertCount << _T("\">") << std::endl;

	    //*************** Export Vertex Buffer ***************
	    bool exportNormals = true;
        bool exportVertexColors = this->exportSettings.exportVertexColors;

	    out << std::boolalpha;
	    out << _T("\t\t\t\t<vertexbuffer positions=\"true\" normals=\"") << exportNormals << _T("\" colours_diffuse=\"") << exportVertexColors << _T("\" texture_coords=\"") << texMapChannels.Count() << _T("\"");
    	
        for (int texChannelIndex = 0; texChannelIndex < texMapChannels.Count(); texChannelIndex++)
		    out << _T(" texture_coords_dimensions_") << texChannelIndex << _T("=\"2\"");
    	
	    out << _T(">") << std::endl;

	    for (int vertexIndex = 0; vertexIndex < vertCount; vertexIndex++) 
        {
		    const Vertex& v = vertexList.GetFront();

		    out << _T("\t\t\t\t\t<vertex>") << std::endl;
		    out << std::showpoint;

            Ogre::Vector3 position = v.position * this->exportSettings.scale;
		    out << _T("\t\t\t\t\t\t<position x=\"") << position.x << _T("\" y=\"") << position.y << _T("\" z=\"") << position.z << _T("\" />") << std::endl;

		    if (this->exportSettings.exportVertexColors) 
            {
			    out << _T("\t\t\t\t\t\t<colour_diffuse value=\"\t") << v.color.r << _T("\t") << v.color.g << _T("\t") << v.color.b << _T("\" />") << std::endl;
		    }

		    if (exportNormals) 
            {
                out << _T("\t\t\t\t\t\t<normal x=\"") << v.normal.x << _T("\" y=\"") << v.normal.y << _T("\" z=\"") << v.normal.z << _T("\" />") << std::endl;
		    }

		    //output the tex coords for each map used
		    for (int ti = 0; ti < texMapChannels.Count(); ti++) 
            {
			    int texMap = texMapChannels[ti];

			    const Ogre::Vector3& uvw = v.GetUVW(texMap); 

			    switch (this->exportSettings.texCoord) 
                {
				    case OgreMax::UV:
					    out << _T("\t\t\t\t\t\t<texcoord u=\"") << uvw.x << _T("\" v=\"") << (1.0 - uvw.y) << _T("\" />") << std::endl; 
					    break;
				    case OgreMax::VW:
					    out << _T("\t\t\t\t\t\t<texcoord v=\"") << uvw.y << _T("\" w=\"") << (1.0 - uvw.z) << _T("\" />") << std::endl; 
					    break;
				    case OgreMax::WU:
					    out << _T("\t\t\t\t\t\t<texcoord w=\"") << uvw.z << _T("\" u=\"") << (1.0 - uvw.x) << _T("\" />") << std::endl; 
					    break;
			    }
		    }
    		
		    out << std::noshowpoint;
		    out << _T("\t\t\t\t\t</vertex>") << std::endl;
		    vertexList.Pop();
	    }

	    out << _T("\t\t\t\t</vertexbuffer>") << std::endl;
	    //*************** End Export Vertex Buffer ***************

	    out << _T("\t\t\t</geometry>") << std::endl;
	    //*************** End Export Geometry ***********

	    //Skin
        WriteSubmeshBoneAssignments(out, gameNode, skin, mesh, reindexedVertices);
        
        out << _T("\t\t</submesh>") << std::endl;
    }

    return faceCount > 0;
}

struct BoneInfluencedVertex
{
    BoneInfluencedVertex()
    {
        this->exported = false;
    }

    void AddAssignment(int vertexIndex, int boneIndex, float weight)
    {
        for (size_t i = 0; i < this->assignments.size(); i++)
        {
            if (this->assignments[i].vertexIndex == vertexIndex && 
                this->assignments[i].boneIndex == boneIndex)
            {
                this->assignments[i].weight += weight;                
                return;
            }
        }

        this->assignments.push_back(Assignment(vertexIndex, boneIndex, weight));
    }

    struct Assignment
    {
        Assignment(int vertexIndex, int boneIndex, float weight)
        {
            this->vertexIndex = vertexIndex;
            this->boneIndex = boneIndex;
            this->weight = weight;
        }

        bool operator < (const Assignment& a) const
        {
            return this->vertexIndex < a.vertexIndex;
        }

        int vertexIndex;
        int boneIndex;
        float weight;
    };
    std::vector<Assignment> assignments;

    bool exported;
};

void MeshExporter::WriteSubmeshBoneAssignments
    (
    stdostream& out, 
    IGameNode* gameNode,
    IGameSkin* gameSkin, 
    IGameMesh* gameMesh, 
    const ReindexedVertices& reindexedVertices
    )
{
    if (gameSkin == NULL)
        return;

    //If the object that has the skin applied is not a mesh then the number of skinned vertices
    //will likely not be the same as the number of mesh vertices. This is because 
    //IPhysiqueExport (and, by association, IGameSkin) use the Object::GetPoint() method
    //which does not necessarily evaluate to the same points as the mesh
    //In this case, we have to search for the GetPoint() vertices closest to the
    //ones in the mesh when exporting each skinned vertex. This won't actually produce
    //a very nice exported result, but it will at least allow the user to see something
    bool skinPointsValid = gameSkin->GetNumOfSkinnedVerts() == gameMesh->GetNumberOfVerts();    
    Object* maxObject = skinPointsValid ? NULL : gameNode->GetMaxNode()->EvalWorldState(0).obj;

    //Used for tracking which vertices have been exported. 
    //This is needed because when we do a search for the closest vertices, we 
    //will likely find the same vertex more than once.
    //It's also used to combine vertex bone weightings since Max has a tendency to 
    //occasionally create more than one bone assignment with the same bone for the same vertex
    std::vector<BoneInfluencedVertex> exportedVertices;
    exportedVertices.resize(gameMesh->GetNumberOfVerts());

    std::vector<BoneInfluencedVertex::Assignment> sortedAssignments;
    if (this->exportSettings.sortBoneAssignments)
        sortedAssignments.reserve(gameMesh->GetNumberOfVerts());

    //Loop through all the vertices, writing out skinning data as we go
    if (gameSkin->GetNumOfSkinnedVerts() > 0) 
    {
        out << _T("\t\t\t<boneassignments>") << std::endl;

        for (int skinVertexIndex = 0; skinVertexIndex < gameSkin->GetNumOfSkinnedVerts(); skinVertexIndex++) 
        {
            int meshVertexIndex;
            if (skinPointsValid)
                meshVertexIndex = skinVertexIndex;
            else
            {
                float closest;
                
                Point3 p = maxObject->GetPoint(skinVertexIndex);
                Utilities::ConvertRawVertex(p, this->exportSettings.upAxis);
                meshVertexIndex = Utilities::GetClosestVertexIndex(gameMesh, p, &closest);
            }

            //Get the bone indices for this vertex
            int vertexBoneCount = gameSkin->GetNumberOfBones(skinVertexIndex);
            if (!exportedVertices[meshVertexIndex].exported && !reindexedVertices[meshVertexIndex].empty() && vertexBoneCount > 0) 
            {
                for (int boneIndex = 0; boneIndex < vertexBoneCount; boneIndex++) 
                {
			        //Get weight per bone influence
			        int vertexBoneIndex = gameSkin->GetBoneIndex(gameSkin->GetBone(skinVertexIndex, boneIndex));
			        float vertexWeight = gameSkin->GetWeight(skinVertexIndex, boneIndex);

                    if (!Utilities::AlmostZero(vertexWeight))
                    {
                        for (size_t reindexed = 0; reindexed < reindexedVertices[meshVertexIndex].size(); reindexed++)
                            exportedVertices[meshVertexIndex].AddAssignment(reindexedVertices[meshVertexIndex][reindexed], vertexBoneIndex, vertexWeight);
                    }
		        }

                for (size_t i = 0; i < exportedVertices[meshVertexIndex].assignments.size(); i++)
                {
                    const BoneInfluencedVertex::Assignment& a = exportedVertices[meshVertexIndex].assignments[i];
                    
                    if (this->exportSettings.sortBoneAssignments)
                        sortedAssignments.push_back(a);
                    else
                        out << _T("\t\t\t\t<vertexboneassignment vertexindex=\"") << a.vertexIndex << _T("\" boneindex=\"") << a.boneIndex << _T("\" weight=\"") << a.weight << _T("\" />") << std::endl;
                }

                exportedVertices[meshVertexIndex].exported = true;
	        }
	    }

        if (this->exportSettings.sortBoneAssignments)
        {
            std::sort(sortedAssignments.begin(), sortedAssignments.end());
            for (size_t i = 0; i < sortedAssignments.size(); i++)
            {
                const BoneInfluencedVertex::Assignment& a = sortedAssignments[i];
                out << _T("\t\t\t\t<vertexboneassignment vertexindex=\"") << a.vertexIndex << _T("\" boneindex=\"") << a.boneIndex << _T("\" weight=\"") << a.weight << _T("\" />") << std::endl;
            }
        }

        out << _T("\t\t\t</boneassignments>") << std::endl;
    }
}

void MeshExporter::WriteSkeleton(stdostream& out, IGameNode* meshNode, IGameSkin* gameSkin, float sampleInterval)
{
    int boneCount = gameSkin->GetTotalSkinBoneCount();
    int boneIndex;

    std::vector<GMatrix> initialBoneTransforms;
    initialBoneTransforms.resize(boneCount);

    out << _T("<?xml version=\"1.0\"?>") << std::endl;
            
    out << std::showpoint;

    out << _T("<skeleton>") << std::endl;
        out << _T("\t<bones>") << std::endl;
            for (boneIndex = 0; boneIndex < boneCount; boneIndex++)
            {
                INode* boneNode = gameSkin->GetBone(boneIndex);
                IGameNode* gameBoneNode = this->gameInterface->GetIGameNode(boneNode);
                
                //Get the node's initial parent transform
                IGameNode* gameParentBoneNode = gameBoneNode->GetNodeParent();
                GMatrix parentTransform;
                if (gameParentBoneNode == NULL || !gameSkin->GetInitBoneTM(gameParentBoneNode, parentTransform))
                    gameSkin->GetInitSkinTM(parentTransform);
                
                //Get the node's initial transform
                GMatrix transform;
                gameSkin->GetInitBoneTM(gameBoneNode, transform);                    
                
                //Get the node's initial transform relative to the parent
                transform = transform * parentTransform.Inverse();

                initialBoneTransforms[boneIndex] = transform;
                
                Point3 position = transform.Translation() * this->exportSettings.scale;
                AngAxis angAxis(transform.Rotation());                    
                Point3 scale = transform.Scaling();

                out << _T("\t\t<bone id=\"") << boneIndex << _T("\" name=\"") << gameBoneNode->GetName() << _T("\">") << std::endl;

                out << _T("\t\t\t<position x=\"") << position.x << _T("\" y=\"") << position.y << _T("\" z=\"") << position.z << _T("\" />") << std::endl;
                out << _T("\t\t\t<rotation angle=\"") << -angAxis.angle << _T("\">") << std::endl;
                    out << _T("\t\t\t\t<axis x=\"") << angAxis.axis.x << _T("\" y=\"") << angAxis.axis.y << _T("\" z=\"") << angAxis.axis.z << _T("\" />") << std::endl;
                out << _T("\t\t\t</rotation>") << std::endl;
                out << _T("\t\t\t<scale x=\"") << scale.x << _T("\" y=\"") << scale.y << _T("\" z=\"") << scale.z << _T("\" />") << std::endl;
                
                out << _T("\t\t</bone>") << std::endl;
            }
        out << _T("\t</bones>") << std::endl;

        out << _T("\t<bonehierarchy>") << std::endl;
            for (boneIndex = 0; boneIndex < boneCount; boneIndex++)
            {
                INode* boneNode = gameSkin->GetBone(boneIndex);
                INode* boneParentNode = boneNode->GetParentNode();
                      
                if (!boneParentNode->IsRootNode() && Utilities::IsBone(boneParentNode, gameSkin))
                    out << _T("\t\t<boneparent bone=\"") << boneNode->GetName() << _T("\" parent=\"") << boneParentNode->GetName() << _T("\" />") << std::endl;
            }
        out << _T("\t</bonehierarchy>") << std::endl;

        WriteSkeletonBoneAnimations(out, meshNode, gameSkin, sampleInterval, initialBoneTransforms);

    out << _T("</skeleton>") << std::endl;

    out << std::noshowpoint;
}

struct BoneTransform
{
    BoneTransform() : translation(0,0,0), scale(1,1,1)
    {
    }

    bool IsNull() const
    {
        return 
            Utilities::AlmostZero(this->angAxis.angle) && 
            Utilities::AlmostZero(this->translation) && 
            Utilities::AlmostZero(this->scale);
    }

    TimeValue time;
    Point3 translation;
    AngAxis angAxis;
    Point3 scale;
};

struct BoneAllowParsed
{
    INode* node;
    bool allow;
    bool tree;
};

static bool IsBoneExportable(INode* node, std::vector<BoneAllowParsed>& exportBones)
{
    bool exportable = true;

    for (size_t i = 0; i < exportBones.size(); i++)
    {
        if (exportBones[i].tree)
        {
            if (exportBones[i].node == node || Utilities::HasDescendant(exportBones[i].node, node))
                exportable = exportBones[i].allow;
        }
        else
        {
            if (exportBones[i].node == node)
                exportable = exportBones[i].allow;
        }
    }       

    return exportable;
}

void MeshExporter::WriteSkeletonBoneAnimations
    (
    stdostream& out, 
    IGameNode* meshNode, 
    IGameSkin* gameSkin, 
    float sampleInterval,
    std::vector<GMatrix>& initialBoneTransforms
    )
{
    NoteTrack* noteTrack = Utilities::GetAnimationNoteTrack(gameSkin->GetMaxModifier());
    DefNoteTrack* defNoteTrack = (DefNoteTrack*)noteTrack;

    out << _T("\t<animations>") << std::endl;

        //Make sure there are a non-zero, even number of keys
        int keyCount = defNoteTrack ? defNoteTrack->NumKeys() : 2;
        if (keyCount > 0 && keyCount % 2 == 0)
        {
            for (int keyIndex = 0; keyIndex < keyCount; keyIndex += 2)
            {
                Interval animationInterval;
                TSTR animationName;
                float animationScale = 1;
                bool copyFirstKeyToLast = false;
                float animationSampleInterval = sampleInterval;
                KeyedValues animationProperties;
                std::vector<BoneAllowParsed> exportBones;
                if (noteTrack != NULL)
                {
                    animationProperties.Parse(defNoteTrack->keys[keyIndex]->note);
                    animationInterval = Interval(defNoteTrack->keys[keyIndex]->time, defNoteTrack->keys[keyIndex + 1]->time);

                    animationProperties.GetValue(PropertyKeys::ANIMATION_NAME, animationName);
                    
                    animationScale = Utilities::GetAnimationScale(animationProperties, animationInterval);

                    copyFirstKeyToLast = Utilities::GetCopyFirstKeyToLast(animationProperties);

                    MeshAnimationSettingsSerializer meshAnimationSettings;
                    meshAnimationSettings.Read(defNoteTrack->keys[keyIndex]->note);
                    BoneAllowParsed boneAllow;    
                    for (int allowBoneIndex = 0; allowBoneIndex < meshAnimationSettings.settings.exportBones.size(); allowBoneIndex++)
                    {
                        boneAllow.node = this->maxInterface->GetINodeByName(meshAnimationSettings.settings.exportBones[allowBoneIndex].name);
                        if (boneAllow.node != NULL)
                        {
                            boneAllow.allow = Utilities::ToBool(meshAnimationSettings.settings.exportBones[allowBoneIndex].allow);
                            boneAllow.tree = Utilities::ToBool(meshAnimationSettings.settings.exportBones[allowBoneIndex].tree);
                            exportBones.push_back(boneAllow);
                        }
                    }

                    animationSampleInterval = Utilities::GetAnimationSampleInterval(animationProperties);
                    if (animationSampleInterval == 0)
                        animationSampleInterval = sampleInterval;
                }
                else
                {
                    animationInterval = this->maxInterface->GetAnimRange();
                    animationName = meshNode->GetName();
                }
                float animationLength = TicksToSec(animationInterval.Duration());

                out << _T("\t\t<animation name=\"") << animationName << _T("\" length=\"") << (animationLength * animationScale) << _T("\">") << std::endl;
                    out << _T("\t\t\t<tracks>") << std::endl;

                        int boneCount = gameSkin->GetTotalSkinBoneCount();                        
                        for (int boneIndex = 0; boneIndex < boneCount; boneIndex++)
                        {
                            INode* boneNode = gameSkin->GetBone(boneIndex);
                            IGameNode* gameBoneNode = this->gameInterface->GetIGameNode(boneNode);

                            //Determine if the bone should be skipped
                            if (!IsBoneExportable(boneNode, exportBones))
                                continue;
                            
                            //Get the key times
                            std::vector<TimeValue> keyTimes;
                            if (animationSampleInterval > 0)
                            {
                                Utilities::GetSampledKeyTimes(keyTimes, animationSampleInterval, animationLength, animationInterval.Start());
                            }
                            else
                            {
                                std::vector<TimeValue> allKeyTimes;
                                Utilities::GetPRSKeyTimes(gameBoneNode, allKeyTimes, POSITION_KEY | ROTATION_KEY | SCALE_KEY);

                                Utilities::GetTimesInInterval(keyTimes, allKeyTimes, animationInterval);
                            }
                            
                            //Exit early if there are no animations for the bone
                            if (keyTimes.size() <= 1)
                                continue;

                            Point3 boneTranslationMask(1,1,1);
                            KeyedValues boneProperties;
                            Utilities::GetNodeUserProperties(boneNode, boneProperties);
                            TSTR ignoreBoneTranslationText;
                            TSTR boneSpecificTranslationKey;
                            boneSpecificTranslationKey.printf(PropertyKeys::REMOVE_BONE_TRANSLATION_FORMAT.data(), boneNode->GetName());
                            if (animationProperties.GetValue(boneSpecificTranslationKey, ignoreBoneTranslationText) || 
                                boneProperties.GetValue(PropertyKeys::REMOVE_BONE_TRANSLATION, ignoreBoneTranslationText))
                            {
                                boneTranslationMask = Utilities::ParseXYZMask(ignoreBoneTranslationText, 0, 1);
                                if (this->exportSettings.upAxis == UP_AXIS_Y)
                                    std::swap(boneTranslationMask.y, boneTranslationMask.z);
                            }
                                                        
                            //Get all the transforms for the current bone's animation                            
                            std::vector<BoneTransform> transforms;
                            transforms.reserve(keyTimes.size());
                            BoneTransform boneTransform;
                            for (size_t keyIndex = 0; keyIndex < keyTimes.size(); keyIndex++)
                            {
                                TimeValue keyTime = keyTimes[keyIndex];
                                TimeValue sampleTime = (copyFirstKeyToLast && keyIndex == keyTimes.size() - 1) ? keyTimes[0] : keyTime;
                                
                                //Get the parent transform at the current time
                                IGameNode* gameParentBoneNode = gameBoneNode->GetNodeParent();
                                GMatrix parentTransform;
                                if (gameParentBoneNode != NULL && Utilities::IsBone(gameParentBoneNode, gameSkin))
                                    parentTransform = gameParentBoneNode->GetWorldTM(sampleTime);
                                else
                                    parentTransform = meshNode->GetWorldTM(sampleTime);

                                //Get the bone transform relative to the parent at the current time
                                GMatrix transform = gameBoneNode->GetWorldTM(sampleTime) * parentTransform.Inverse();
                                
                                //Add the key to the transforms collection
                                //The keys are relative to the bone's initial transform
                                boneTransform.time = keyTime;             
                                boneTransform.translation = boneTranslationMask * (transform.Translation() - initialBoneTransforms[boneIndex].Translation()) * this->exportSettings.scale;
                                Quat rotationKey = transform.Rotation() * initialBoneTransforms[boneIndex].Rotation().Inverse();
                                boneTransform.angAxis = AngAxis(rotationKey);
                                boneTransform.scale = transform.Scaling() / initialBoneTransforms[boneIndex].Scaling();
                                if (animationSampleInterval == 0 || keyIndex == 0 || !boneTransform.IsNull() || transforms.size() > 1)
                                    transforms.push_back(boneTransform);
                            }
                                                        
                            //Exit early if there are no transforms
                            if (transforms.size() <= 1)
                                continue;

                            //Export all the animation keys
                            out << _T("\t\t\t\t<track bone=\"") << gameBoneNode->GetName() << _T("\">") << std::endl;
                                out << _T("\t\t\t\t\t<keyframes>") << std::endl;                                    
                                    for (size_t transformIndex = 0; transformIndex < transforms.size(); transformIndex++)
                                    {
                                        const Point3& positionKey = transforms[transformIndex].translation;
                                        const AngAxis& angAxis = transforms[transformIndex].angAxis;
                                        const Point3& scale = transforms[transformIndex].scale;
                                        
                                        out << _T("\t\t\t\t\t\t<keyframe time=\"") << (TicksToSec(transforms[transformIndex].time - animationInterval.Start()) * animationScale) << _T("\">") << std::endl;
                                            out << _T("\t\t\t\t\t\t\t<translate x=\"") << positionKey.x << _T("\" y=\"") << positionKey.y << _T("\" z=\"") << positionKey.z << _T("\" />") << std::endl;
                                            out << _T("\t\t\t\t\t\t\t<rotate angle=\"") << -angAxis.angle << _T("\">") << std::endl;
                                                out << _T("\t\t\t\t\t\t\t\t<axis x=\"") << angAxis.axis.x << _T("\" y=\"") << angAxis.axis.y << _T("\" z=\"") << angAxis.axis.z << _T("\" />") << std::endl;
                                            out << _T("\t\t\t\t\t\t\t</rotate>") << std::endl;
                                            out << _T("\t\t\t\t\t\t\t<scale x=\"") << scale.x << _T("\" y=\"") << scale.y << _T("\" z=\"") << scale.z << _T("\" />") << std::endl;
                                        out << _T("\t\t\t\t\t\t</keyframe>") << std::endl;
                                    }
                                out << _T("\t\t\t\t\t</keyframes>") << std::endl;
                            out << _T("\t\t\t\t</track>") << std::endl;
                        }

                    out << _T("\t\t\t</tracks>") << std::endl;
                out << _T("\t\t</animation>") << std::endl;
            }
        }
        
    out << _T("\t</animations>") << std::endl;
}

struct VertexOffset
{
    int index;
    Point3 offset;
};

void MeshExporter::WriteMorphPosesAndAnimations
    (
    stdostream& out, 
    IGameMorpher* gameMorpher, 
    IGameMesh* gameMesh, 
    IGameNode* gameNode,
    std::list<Submesh>& submeshes,
    float sampleInterval
    )
{
    if (gameMorpher == NULL || gameMorpher->GetMorphType() != IGameMorpher::MORPHER_MODIFIER)
    {
        //The morpher is not of the proper type. It must be a modifier
        return;
    }

    InitialNodeState* initialNodeState = this->initialNodeStates.find(gameNode->GetMaxNode())->second;
        
    const int targetCount = gameMorpher->GetNumberOfMorphTargets();

    //Write the morph target nodes as poses
    out << _T("\t<poses>") << std::endl;    
    for (int targetIndex = 0; targetIndex < targetCount; targetIndex++)
    {
        IGameNode* morphTargetGameNode = gameMorpher->GetMorphTarget(targetIndex);
        IGameObject* morphTargetGameObject = morphTargetGameNode->GetIGameObject();

        if (morphTargetGameObject->GetIGameType() == IGameObject::IGAME_MESH)
        {
            IGameMesh* morphTargetGameMesh = (IGameMesh*)morphTargetGameObject;            
            Utilities::InitializeData(morphTargetGameObject, morphTargetGameNode);
            
            std::list<VertexOffset> offsets;
            for (std::list<Submesh>::iterator submesh = submeshes.begin(); submesh != submeshes.end(); ++submesh)
            {
                //Get all the pose morph offsets for the submesh
                int vertexCount = std::min(gameMesh->GetNumberOfVerts(), morphTargetGameMesh->GetNumberOfVerts());
                
                for (int index = 0; index < vertexCount; index++)
                {
                    //Check the vertex only if it's part of the submesh
                    if (submesh->ContainsVertex(index))
                    {
                        VertexOffset vertexOffset;
                        vertexOffset.offset = morphTargetGameMesh->GetVertex(index, true) - initialNodeState->unmorphedInitialVertices[index] * this->exportSettings.scale;
                        if (!Utilities::IsZero(vertexOffset.offset))
                        {
                            for (size_t reindexed = 0; reindexed < submesh->reindexedVertices[index].size(); reindexed++)
                            {
                                vertexOffset.index = submesh->reindexedVertices[index][reindexed];
                                offsets.push_back(vertexOffset);
                            }
                        }
                    }
                }

                //Write the pose and its offsets
                TSTR submeshPoseName = morphTargetGameNode->GetName();
                if (submeshes.size() > 1)
                {
                    submeshPoseName += _T("_");
                    submeshPoseName += Utilities::ToString(submesh->index);
                }

                out << _T("\t\t<pose target=\"submesh\" index=\"") << submesh->index << _T("\" name=\"") << submeshPoseName << _T("\">") << std::endl;
                
                for (std::list<VertexOffset>::iterator offset = offsets.begin(); offset != offsets.end(); ++offset)
                    out << _T("\t\t\t<poseoffset index=\"") << offset->index << _T("\" x=\"") << offset->offset.x << _T("\" y=\"") << offset->offset.y << _T("\" z=\"") << offset->offset.z << _T("\" />") << std::endl;
                
                out << _T("\t\t</pose>") << std::endl;

                offsets.clear();
            }
        }
    }
    out << _T("\t</poses>") << std::endl;

    //Write the morph animations
    NoteTrack* noteTrack = Utilities::GetAnimationNoteTrack(gameMorpher->GetMaxModifier());
    DefNoteTrack* defNoteTrack = (DefNoteTrack*)noteTrack;

    out << _T("\t<animations>") << std::endl;

        //Make sure there are a non-zero, even number of keys
        int keyCount = defNoteTrack ? defNoteTrack->NumKeys() : 2;
        if (keyCount > 0 && keyCount % 2 == 0)
        {
            for (int keyIndex = 0; keyIndex < keyCount; keyIndex += 2)
            {
                Interval animationInterval;
                TSTR animationName;                
                float animationScale = 1;
                bool copyFirstKeyToLast = false;
                float animationSampleInterval = sampleInterval;
                if (noteTrack != NULL)
                {
                    animationInterval = Interval(defNoteTrack->keys[keyIndex]->time, defNoteTrack->keys[keyIndex + 1]->time);

                    KeyedValues properties(defNoteTrack->keys[keyIndex]->note);
                    properties.GetValue(PropertyKeys::ANIMATION_NAME, animationName);

                    animationScale = Utilities::GetAnimationScale(properties, animationInterval);

                    copyFirstKeyToLast = Utilities::GetCopyFirstKeyToLast(properties);

                    animationSampleInterval = Utilities::GetAnimationSampleInterval(properties);
                    if (animationSampleInterval == 0)
                        animationSampleInterval = sampleInterval;
                }
                else
                {
                    animationInterval = this->maxInterface->GetAnimRange();
                    animationName = gameNode->GetName();
                }
                float animationLength = TicksToSec(animationInterval.Duration());
                
                out << _T("\t\t<animation name=\"") << animationName << _T("\" length=\"") << (animationLength * animationScale) << _T("\">") << std::endl;
                    out << _T("\t\t\t<tracks>") << std::endl;
                    
                        //Get the key times for the animation
                        std::vector<TimeValue> keyTimes;
                        if (animationSampleInterval > 0)
                        {
                            Utilities::GetSampledKeyTimes(keyTimes, animationSampleInterval, animationLength, animationInterval.Start());                                
                        }
                        else
                        {                        
                            std::vector<TimeValue> allKeyTimes;
                            for (int targetIndex = 0; targetIndex < targetCount; targetIndex++)
                            {
                                IGameNode* gameTargetNode = gameMorpher->GetMorphTarget(targetIndex);
                                IGameControl* gameWeightControl = gameMorpher->GetMorphWeight(targetIndex);

                                Utilities::GetFloatKeyTimes(gameWeightControl->GetMaxControl(IGAME_FLOAT), allKeyTimes);
                            }
                            Utilities::GetTimesInInterval(keyTimes, allKeyTimes, animationInterval);
                        }
                        
                        //Write the keyframes for each submesh
                        for (std::list<Submesh>::iterator submesh = submeshes.begin(); submesh != submeshes.end(); ++submesh)
                        {
                            out << _T("\t\t\t\t<track target=\"submesh\" index=\"") << submesh->index << _T("\" type=\"pose\">") << std::endl;
                                out << _T("\t\t\t\t\t<keyframes>") << std::endl;
                                    //Export all the keys                                
                                    for (size_t keyIndex = 0; keyIndex < keyTimes.size(); keyIndex++)
                                    {
                                        bool wroteKeyframe = false;
                                        TimeValue keyTime = keyTimes[keyIndex];
                                        TimeValue sampleTime = (copyFirstKeyToLast && keyIndex == keyTimes.size() - 1) ? keyTimes[0] : keyTime;

                                        for (int targetIndex = 0; targetIndex < targetCount; targetIndex++)
                                        {
                                            IGameControl* gameWeightControl = gameMorpher->GetMorphWeight(targetIndex);
                                            float influence = 0;
                                            gameWeightControl->GetMaxControl(IGAME_FLOAT)->GetValue(sampleTime, &influence, FOREVER);
                                            if (influence > 0)
                                            {
                                                influence /= 100;
                                                if (!wroteKeyframe)
                                                {
                                                    out << _T("\t\t\t\t\t\t<keyframe time=\"") << (TicksToSec(keyTime - animationInterval.Start()) * animationScale) << _T("\">") << std::endl;
                                                    wroteKeyframe = true;
                                                }
                                                int poseIndex = (int)(targetIndex * submeshes.size() + submesh->index);
                                                out << _T("\t\t\t\t\t\t\t<poseref poseindex=\"") << poseIndex << _T("\" influence=\"") << influence << _T("\" />") << std::endl;
                                            }
                                        }
                                        
                                        if (wroteKeyframe)
                                            out << _T("\t\t\t\t\t\t</keyframe>") << std::endl;
                                    }
                                out << _T("\t\t\t\t\t</keyframes>") << std::endl;
                            out << _T("\t\t\t\t</track>") << std::endl;
                        }

                    out << _T("\t\t\t</tracks>") << std::endl;
                out << _T("\t\t</animation>") << std::endl;
            }
        }
        
    out << _T("\t</animations>") << std::endl;
}

struct SampledPoseAnimation
{
    Interval interval;
    TSTR name;
    std::vector<TimeValue> keyTimes;
    float sampleInterval;

    struct SubmeshPose
    {
        std::vector<int> posesIndices;
    };
    std::vector<SubmeshPose> submeshPoses;
};

void MeshExporter::WriteMorphAnimationsBySampling
    (
    stdostream& out, 
    IGameMesh* gameMesh, 
    IGameNode* gameNode,
    std::list<Submesh>& submeshes,
    bool allowSkinModifier,
    float sampleInterval
    )
{
    //Find the animation note track
    NoteTrack* noteTrack = Utilities::GetAnimationNoteTrack(gameNode->GetMaxNode()->GetObjectRef());
    DefNoteTrack* defNoteTrack = (DefNoteTrack*)noteTrack;
    int keyCount = defNoteTrack ? defNoteTrack->NumKeys() : 2;        

    //Get preliminary animation info
    std::vector<SampledPoseAnimation> animations;    
    float animationScale = 1;
    bool copyFirstKeyToLast = false;
    if (keyCount > 0 && keyCount % 2 == 0)
    {
        animations.resize(keyCount / 2);
        
        if (noteTrack != NULL)
        {
            //There are custom animations, probably multiple animations
            for (int keyIndex = 0; keyIndex < keyCount; keyIndex += 2)
            {
                SampledPoseAnimation& animation = animations[keyIndex / 2];
                animation.sampleInterval = sampleInterval;

                if (noteTrack != NULL)
                {
                    KeyedValues properties(defNoteTrack->keys[keyIndex]->note);
                    animation.interval = Interval(defNoteTrack->keys[keyIndex]->time, defNoteTrack->keys[keyIndex + 1]->time);                

                    properties.GetValue(PropertyKeys::ANIMATION_NAME, animation.name);

                    animationScale = Utilities::GetAnimationScale(properties, animation.interval);

                    copyFirstKeyToLast = Utilities::GetCopyFirstKeyToLast(properties);

                    animation.sampleInterval = Utilities::GetAnimationSampleInterval(properties);
                    if (animation.sampleInterval == 0)
                        animation.sampleInterval = sampleInterval;
                }
                else
                {
                    animation.interval = this->maxInterface->GetAnimRange();
                    animation.name = gameNode->GetName();
                }
            }
        }
        else
        {
            //No animations explicitly specified. Take the entire animation range            

            SampledPoseAnimation& animation = animations[0];
            
            animation.interval = this->maxInterface->GetAnimRange();
            animation.name = gameNode->GetName();
            animation.sampleInterval = sampleInterval;
        }
    }
    
    //Write morph animations
    out << _T("\t<animations>") << std::endl;
    for (size_t animationIndex = 0; animationIndex < animations.size(); animationIndex++)
    {
        SampledPoseAnimation& animation = animations[animationIndex];
        Utilities::GetSampledKeyTimes(animation.keyTimes, animation.sampleInterval, TicksToSec(animation.interval.Duration()), animation.interval.Start());

        out << _T("\t\t<animation name=\"") << animation.name << _T("\" length=\"") << (TicksToSec(animation.interval.Duration()) * animationScale) << _T("\">") << std::endl;
            out << _T("\t\t\t<tracks>") << std::endl;
            
                //Write the keyframes for each submesh
                for (std::list<Submesh>::iterator submesh = submeshes.begin(); submesh != submeshes.end(); ++submesh)
                {       
                    out << _T("\t\t\t\t<track target=\"submesh\" index=\"") << submesh->index << _T("\" type=\"morph\">") << std::endl;
                        out << _T("\t\t\t\t\t<keyframes>") << std::endl;

                        for (size_t keyIndex = 0; keyIndex < animation.keyTimes.size(); keyIndex++)
                        {
                            TimeValue keyTime = animation.keyTimes[keyIndex];
                            TimeValue sampleTime = (copyFirstKeyToLast && keyIndex == animation.keyTimes.size() - 1) ? animation.keyTimes[0] : keyTime;
                        
                            std::vector<Point3> meshVerts;
                            if (allowSkinModifier)
                                Utilities::EvalMeshVertices(gameNode->GetMaxNode(), sampleTime, this->exportSettings.upAxis, meshVerts);
                            else
                                Utilities::EvalUnskinnedMeshVertices(gameNode->GetMaxNode(), sampleTime, this->exportSettings.upAxis, meshVerts);
                                                            
                            out << _T("\t\t\t\t\t\t<keyframe time=\"") << (TicksToSec(keyTime - animation.interval.Start()) * animationScale) << _T("\">") << std::endl;
                            for (int vertexIndex = 0; vertexIndex < submesh->exportedVertexIndices.size(); vertexIndex++)
                            {
                                const Point3& position = meshVerts[submesh->exportedVertexIndices[vertexIndex]];
                                out << _T("\t\t\t\t\t\t\t<position x=\"") << position.x << _T("\" y=\"") << position.y << _T("\" z=\"") << position.z << _T("\" />") << std::endl;
                            }
                            out << _T("\t\t\t\t\t\t</keyframe>") << std::endl;
                        }

                        out << _T("\t\t\t\t\t</keyframes>") << std::endl;
                    out << _T("\t\t\t\t</track>") << std::endl;
                }
            out << _T("\t\t\t</tracks>") << std::endl;
        out << _T("\t\t</animation>") << std::endl;
    }
    out << _T("\t</animations>") << std::endl;
}

void MeshExporter::ConvertXmlSkeletonToBinary(const ExportedMesh& exportedMesh)
{
    //Get the input file's extension
    TSTR fileNameExtension = Utilities::GetExtension(exportedMesh.skeletonFileName);
    fileNameExtension = fileNameExtension.Substr(1, fileNameExtension.length() - 1); //Remove leading '.'
    fileNameExtension.toLower();    

    //Determine the output file's location and extension
    //The input file name is presumed to end with ".skeleton.xml", so remove
    //the ".xml" part to get the binary file name
    TSTR binaryFileName = Utilities::RemoveExtension(exportedMesh.skeletonFileName);    

    TSTR binaryFileNameExtension = Utilities::GetExtension(binaryFileName);
    binaryFileNameExtension = binaryFileNameExtension.Substr(1, binaryFileNameExtension.length() - 1); //Remove leading '.'
    binaryFileNameExtension.toLower();

    //Determine the log file name
    TSTR logFileName = this->exportSettings.logPath + Utilities::GetFileName(binaryFileName) + _T(".log");

    //Set the options
    Ogre::XmlOptions options;
    options.source = exportedMesh.skeletonFileName.data();
	options.source="models/"+options.source;
    options.dest = binaryFileName.data();
	options.dest="models/"+options.dest;
    options.sourceExt = fileNameExtension.data();
    options.destExt = binaryFileName.data();
    options.logFile = logFileName.data();   

    options.numLods = 0;
    options.lodDist = 0;
    options.usePercent = 0;
    options.lodPercent = 0;
    options.lodFixed = 0;
    options.nuextremityPoints = 0; //I don't know what this option does
    options.generateEdgeLists = false;
    options.generateTangents = false;
    options.tangentSemantic = Ogre::VES_TANGENT;
    options.reorganiseBuffers = false;
    options.optimiseAnimations = true;
    options.quietMode = false;
    options.d3d = false;
    options.gl = false;
    options.endian = this->exportSettings.endian;
 
    try
    {
        //Perform the conversion
        Ogre::XMLConverter::Convert(options);
    }
    catch (Ogre::Exception e)
    {
        //Should probably record the error message
    }
}

void MeshExporter::ConvertXmlMeshToBinary(const ExportedMesh& exportedMesh)
{
    //Get the input file's extension
    TSTR fileNameExtension = Utilities::GetExtension(exportedMesh.meshFileName);
    fileNameExtension = fileNameExtension.Substr(1, fileNameExtension.length() - 1); //Remove leading '.'
    fileNameExtension.toLower();    

    //Determine the output file's location and extension
    //The input file name is presumed to end with ".mesh.xml", so remove
    //the ".xml" part to get the binary file name
    TSTR binaryFileName = Utilities::RemoveExtension(exportedMesh.meshFileName);    

    TSTR binaryFileNameExtension = Utilities::GetExtension(binaryFileName);
    binaryFileNameExtension = binaryFileNameExtension.Substr(1, binaryFileNameExtension.length() - 1); //Remove leading '.'
    binaryFileNameExtension.toLower();

    //Determine the log file name
    TSTR logFileName = this->exportSettings.logPath + Utilities::GetFileName(binaryFileName) + _T(".log");

    //Get the mesh node's custom export settings, if any
    InitialNodeState* initialNodeState = this->initialNodeStates.find(exportedMesh.meshNode->GetMaxNode())->second;
    const KeyedValues& meshNodeProperties = initialNodeState->properties;
    ExportSettings meshNodeExportSettings;
    meshNodeExportSettings.Initialize(meshNodeProperties);
    
    //Set the options
    Ogre::XmlOptions options;
    options.source = exportedMesh.meshFileName.data();
	options.source=modelDir(options.source);
    options.dest = binaryFileName.data();
	options.dest=modelDir(options.dest);
    options.sourceExt = fileNameExtension.data();
    options.destExt = binaryFileName.data();
    options.logFile = logFileName.data();   

    TSTR generateLodText;
    meshNodeProperties.GetValue(PropertyKeys::GENERATE_LOD, generateLodText);
    bool generateLod;
    if (!generateLodText.isNull())
        generateLod = Utilities::ToBool(generateLodText);
    else
        generateLod = this->exportSettings.generateLod;
    
    if (generateLod)
    {        
        options.numLods = 
            meshNodeProperties.HasValue(PropertyKeys::NUM_LOD_LEVELS) ? 
            meshNodeExportSettings.numLodLevels : 
            this->exportSettings.numLodLevels;
        options.lodDist = 
            meshNodeProperties.HasValue(PropertyKeys::LOD_DISTANCE) ? 
            meshNodeExportSettings.lodDistance : 
            this->exportSettings.lodDistance;
        options.lodDist *= this->exportSettings.scale;
        options.usePercent = 
            meshNodeProperties.HasValue(PropertyKeys::LOD_REDUCTION_TYPE) ? 
            meshNodeExportSettings.lodReductionType == LOD_PERCENT : 
            this->exportSettings.lodReductionType == LOD_PERCENT;
        options.lodPercent = 
            (meshNodeProperties.HasValue(PropertyKeys::LOD_PERCENT_REDUCTION) || meshNodeProperties.HasValue(PropertyKeys::LOD_REDUCTION_AMOUNT)) ? 
            meshNodeExportSettings.lodPercentReduction : 
            this->exportSettings.lodPercentReduction;
        options.lodFixed = 
            (meshNodeProperties.HasValue(PropertyKeys::LOD_VERTEX_REDUCTION) || meshNodeProperties.HasValue(PropertyKeys::LOD_REDUCTION_AMOUNT)) ? 
            meshNodeExportSettings.lodVertexReduction : 
            this->exportSettings.lodVertexReduction;
    }
    else
    {
        options.numLods = 0;
        options.lodDist = 0;
        options.usePercent = 0;
        options.lodPercent = 0;
        options.lodFixed = 0;
    }

    options.nuextremityPoints = 0; //I don't know what this option does
    options.generateEdgeLists = this->exportSettings.generateEdgeLists;
    options.generateTangents = exportedMesh.hasTextureCoords && this->exportSettings.generateTangents;
    options.tangentSemantic = Ogre::VES_TANGENT;
    options.reorganiseBuffers = true;
    options.optimiseAnimations = true;
    options.quietMode = false;
    options.d3d = this->exportSettings.vertexColorFormat == DIRECT3D;
    options.gl = this->exportSettings.vertexColorFormat == OPENGL;
    options.endian = this->exportSettings.endian;
 
    try
    {
        //Perform the conversion
        Ogre::XMLConverter::Convert(options);
    }
    catch (Ogre::Exception e)
    {
        //Should probably record the error message
    }
}
