
#ifndef OgreMax_MeshExporter_INCLUDED
#define OgreMax_MeshExporter_INCLUDED


//Includes---------------------------------------------------------------------
#include "ExportSettings.h"
#include "ExporterBase.h"
#include "MaterialExporter.h"
#include "Utilities.h"
#include "MeshNormals.h"


//Classes----------------------------------------------------------------------
namespace OgreMax 
{
    /** Intiial node information, initialized before IGame has been initialized */
    struct InitialNodeState
    {
        /** Parsed node user properties */
        KeyedValues properties;

        /** The enabled/disabled state of the node object's modifier stack */
        std::vector<bool> modifierStackState;

        /** The vertices of the node mesh at frame 0. Used as a reference point when exporting poses */
        std::vector<Point3> unmorphedInitialVertices;

        MeshNormals unmorphedInitialNormals;
    };
    typedef std::map<INode*, InitialNodeState*> InitialNodeStates;

    class MeshExporter : public ExporterBase
	{		
	public:
        MeshExporter
            (
            const KeyedValues& sceneProperties,
            const ExportSettings& settings, 
            const InitialNodeStates& initialNodeStates,
            IGameScene* gameInterface,
            Interface* maxInterface
            );
		virtual ~MeshExporter();

        bool ExportMeshes(const std::list<GameNodeAndObject>& meshNodes, MaterialMap& materials);

	protected:        	
        struct ExportedMesh
        {
            bool HasSkeleton() const
            {
                return 
                    !Utilities::IsNull(this->skeletonFileName) && 
                    !Utilities::IsNull(this->skeletonXml);
            }

            TSTR name;
            TSTR meshFileName;
            TSTR meshXml;            
            TSTR skeletonFileName;
            TSTR skeletonXml;
            IGameNode* meshNode;
            bool hasTextureCoords;
        };

        typedef std::vector<int> ReindexedVertex;
        typedef std::vector<ReindexedVertex> ReindexedVertices;

        struct Submesh
        {
            bool ContainsVertex(int vertexIndex) const
            {
                return 
                    vertexIndex < (int)this->reindexedVertices.size() && 
                    !this->reindexedVertices[vertexIndex].empty();
            }

            TSTR name;
            int index;
            ReindexedVertices reindexedVertices;            
            std::vector<int> exportedVertexIndices;
        };

        bool ExportMeshes
            (
            const std::list<GameNodeAndObject>& meshNodes, 
            std::list<ExportedMesh>& meshes, 
            MaterialMap& materials
            );

		// mesh file stream functions
		void WriteFileHeader(stdostream& out);

		void WriteFileFooter
            (
            stdostream& out, 
            IGameMesh* gameMesh, 
            IGameNode* gameNode, 
            bool useMorphVertexAnimation,
            bool morphOnly,
            const TSTR& skeletonFileName, 
            float vertexAnimationSampleInterval,
            std::list<Submesh>& submeshes
            );

		bool WriteSubmesh
            (
            stdostream& out, 
            IGameNode* gameNode,
            IGameMesh* gameMesh, 
            IGameSkin* gameSkin,
            IGameMaterial* gameMaterial, 
            TSTR& materialName, 
            bool flipFaces,
            ReindexedVertices& reindexedVertices,
            std::vector<int>& exportedVertexIndices
            );

        void WriteSubmeshBoneAssignments(stdostream& out, IGameNode* gameNode, IGameSkin* skin, IGameMesh* mesh, const ReindexedVertices& reindexedVertices);

        void WriteSkeleton(stdostream& out, IGameNode* meshNode, IGameSkin* gameSkin, float sampleInterval);
        void WriteSkeletonBoneAnimations
            (
            stdostream& out, 
            IGameNode* meshNode, 
            IGameSkin* gameSkin, 
            float sampleInterval,
            std::vector<GMatrix>& initialBoneTransforms
            );
        
        void WriteMorphPosesAndAnimations
            (
            stdostream& out, 
            IGameMorpher* gameMorpher, 
            IGameMesh* gameMesh, 
            IGameNode* gameNode, 
            std::list<Submesh>& submeshes,
            float sampleInterval
            );

        void WriteMorphAnimationsBySampling
            (
            stdostream& out, 
            IGameMesh* gameMesh, 
            IGameNode* gameNode,
            std::list<Submesh>& submeshes,
            bool allowSkinModifier,
            float sampleInterval
            );

        void ConvertXmlSkeletonToBinary(const ExportedMesh& exportedMesh);
        void ConvertXmlMeshToBinary(const ExportedMesh& exportedMesh);        

	private:
        const KeyedValues& sceneProperties;
        const InitialNodeStates& initialNodeStates;

        std::map<TSTR, TSTR> exportedMeshNames;
        std::map<TSTR, TSTR> exportedSkeletonNames;
	};
}

#endif