
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "OgreMaxSceneExporter.h"
#include "PluginGlobals.h"
#include "Utilities.h"
#include "resource.h"
#include "PropertyKeys.h"
#include <fstream>

using namespace OgreMax;

#define HINGE_CLASS_ID Class_ID(0x754274c6, 0x4b6750eb)
#define POINT_POINT_CLASS_ID Class_ID(0x5fcc6503, 0x41095b28)

//Local classes-----------------------------------------------------------------
class OgreMaxSceneExporterClassDesc : public ClassDesc 
{
public:
	int IsPublic()						{return TRUE;}
	void* Create(BOOL loading = FALSE)	{return new OgreMaxSceneExporter;}
	const TCHAR* ClassName()			{return PluginGlobals::GetString(IDS_EXPORTER_CLASS_NAME);}
	SClass_ID SuperClassID()			{return SCENE_EXPORT_CLASS_ID;}
    Class_ID ClassID()					{return Class_ID(0xcee3d67, 0x34e062f);}	
    const TCHAR* Category()             {return PluginGlobals::GetString(IDS_PLUGINS_CATEGORY);}

public:
    static OgreMaxSceneExporterClassDesc& GetInstance()
    {
        static OgreMaxSceneExporterClassDesc instance;
        return instance;
    }
};


//Static initialization--------------------------------------------------------
const TimeValue OgreMaxSceneExporter::DefaultTimeValue = 0;
const int OgreMaxSceneExporter::DefaultFrameNumber = 0;


//Implementation----------------------------------------------------------------
OgreMaxSceneExporter::OgreMaxSceneExporter()
{    
    this->sceneXmlElement = NULL;
        this->nodesXmlElement = NULL;    
        this->externalsXmlElement = NULL;
        this->environmentXmlElement = NULL;
            this->skyBoxXmlElement = NULL;
            this->skyDomeXmlElement = NULL;
            this->skyPlaneXmlElement = NULL;
        this->terrainXmlElement = NULL;
}

OgreMaxSceneExporter::~OgreMaxSceneExporter()
{    
    Utilities::DeleteAll(this->modelExporters);
}

int OgreMaxSceneExporter::ExtCount()
{
	return 1;
}

const TCHAR* OgreMaxSceneExporter::Ext(int n)
{
	return _T("scene");
}

const TCHAR* OgreMaxSceneExporter::LongDesc()
{
    return PluginGlobals::GetString(IDS_EXPORTER_LONG_DESCRIPTION);
}

const TCHAR* OgreMaxSceneExporter::ShortDesc()
{
	return PluginGlobals::GetString(IDS_EXPORTER_SHORT_DESCRIPTION);
}

const TCHAR* OgreMaxSceneExporter::AuthorName()
{
    return PluginGlobals::versionInfo.GetCompanyName().data();
}

const TCHAR* OgreMaxSceneExporter::CopyrightMessage()
{
    return PluginGlobals::versionInfo.GetLegalCopyright().data();    
}

const TCHAR* OgreMaxSceneExporter::OtherMessage1()
{
	return _T("");
}

const TCHAR* OgreMaxSceneExporter::OtherMessage2()
{
	return _T("");
}

unsigned int OgreMaxSceneExporter::Version()
{
	return 100;
}

void OgreMaxSceneExporter::ShowAbout(HWND hWnd)
{    
    Utilities::ShowAbout();
}

int OgreMaxSceneExporter::DoExport
    (
    const TCHAR* fileName, 
    ExpInterface* exportInterface, 
    Interface* maxInterface, 
    BOOL suppressPrompts, 
    DWORD options
    )
{
    //Hold onto the export settings so they don't have to be passed around
    this->exportFileName = fileName;
    Utilities::MakeExtensionLowercase(this->exportFileName);
    this->exportPath = Utilities::GetDirectoryPath(this->exportFileName);
    this->maxInterface = maxInterface;	
    this->suppressPrompts = suppressPrompts == TRUE;
    this->options = options;
    
    //Initialize the exporter    
    bool initialized = Initialize();
    if (initialized)
    {
        //Export everything    
        ExportScene();            
        ExportMeshesAndMaterials();
        ExportPhysics();
        ExportModels();
    }

    //Shut down the exporter
    Shutdown();

    if (!initialized)
    {
        if (suppressPrompts)
        {
            //There was an error but no error was shown. Let Max know the export failed
            return IMPEXP_FAIL;
        }
        else
        {
            //An error was already shown. Return success so Max doesn't show its own error
            return IMPEXP_SUCCESS;
        }
    }
    else
        return IMPEXP_SUCCESS;
}

BOOL OgreMaxSceneExporter::SupportsOptions(int ext, DWORD options)
{	
    if (options & SCENE_EXPORT_SELECTED)
        return true;

	return FALSE;
}

ClassDesc* OgreMaxSceneExporter::GetClassDesc()
{
	return &OgreMaxSceneExporterClassDesc::GetInstance();
}

void OgreMaxSceneExporter::ShowError(const TSTR& errorMessage)
{
    MessageBox
        (
        this->maxInterface->GetMAXHWnd(), 
        errorMessage.data(), 
        PluginGlobals::GetString(IDS_EXPORT_ERROR_CAPTION), 
        MB_OK | MB_ICONEXCLAMATION
        );
}

TiXmlElement* OgreMaxSceneExporter::GetCurrentNodeXmlElement() 
{
    //Create the 'nodes' element if necessary
    if (!this->nodesXmlElement)
    {
        this->nodesXmlElement = (TiXmlElement*)this->sceneXmlElement->LinkEndChild(new TiXmlElement(_T("nodes")));
        this->nodeXmlElementStack.push_back(this->nodesXmlElement);
    }
    
    return this->nodeXmlElementStack.back();
}

bool OgreMaxSceneExporter::Initialize()
{
    //Parse the scene info and load any necessary settings
    InitializeSceneProperties();
    InitializeExportSettings();

    //Get all the modifier stack states and disable invalid modifiers
    int nodeCount = this->maxInterface->GetRootNode()->NumberOfChildren();
    for (int index = 0; index < nodeCount; index++)
    {
        if (!GetInitialNodeStates(this->maxInterface->GetRootNode()->GetChildNode(index)))
            return false;
    }
    
    //Initialize IGame interface
    if (!InitializeGameInterface())
        return false;

    return true;
}

void OgreMaxSceneExporter::Shutdown()
{
    //Clean up IGame interface
    this->gameInterface->ReleaseIGame();

    //Restore the modifier stack states
    for (InitialNodeStates::iterator state = this->initialNodeStates.begin(); 
        state != this->initialNodeStates.end(); 
        ++state)
    {
        Utilities::SetModifierStackState(state->first->GetObjectRef(), state->second->modifierStackState);        
        delete state->second;
    }    
}

void OgreMaxSceneExporter::InitializeSceneProperties()
{
    //Parse the summary info key=value properties
    this->sceneSummary = Utilities::GetSceneSummaryInfo();
    this->sceneProperties.Parse(this->sceneSummary);
}

void OgreMaxSceneExporter::InitializeExportSettings()
{
    //Initialize export settings from parsed scene properties
    this->exportSettings.Initialize(this->sceneProperties);

    //Generate all the export paths
    this->exportSettings.MakeExportPaths(this->exportPath, this->exportFileName);
    
    //Initialize physics exporter
    this->physicsExporter.SetPhysicsExportType(this->exportSettings.physicsExportType);
    
    //Determine world direction vectors
    this->rightVector = Point3(1,0,0);
    if (this->exportSettings.upAxis == UP_AXIS_Z)
    {
        this->forwardVector = Point3(0,1,0);
        this->upVector = Point3(0,0,1);
        this->upVectorPerpendicular = Point3(0,1,0);
    }
    else
    {
        this->forwardVector = Point3(0,0,1);
        this->upVector = Point3(0,1,0);        
        this->upVectorPerpendicular = Point3(0,0,-1);
    }
}

bool OgreMaxSceneExporter::InitializeGameInterface()
{
    this->gameInterface = GetIGameInterface();    

    //Set the 3DS Max-defined default properties file name
    TSTR propertiesFileName = _T("IGameProp.xml"); 
    this->gameInterface->SetPropertyFile(propertiesFileName);

    if (this->gameInterface->InitialiseIGame(false))
    {    
        //Successful initialization

        //Set up axis conversion
        IGameConversionManager* conversion = GetConversionManager();
        if (this->exportSettings.upAxis == UP_AXIS_Y)
            conversion->SetCoordSystem(IGameConversionManager::IGAME_OGL);
        else
            conversion->SetCoordSystem(IGameConversionManager::IGAME_MAX);

        //Set static frame
        this->gameInterface->SetStaticFrame(DefaultFrameNumber);

        return true;
    }
    else
    {
        //Failed to initialize

        //Get the path to the IGame properties file
        TSTR errorMessage;
        TSTR pluginConfigDirectory = this->maxInterface->GetDir(APP_PLUGCFG_DIR);

        //The 3DS Max 9 SDK documentation indicates the returned directory will have the
        //trailing path separator, but the documentation is either wrong or different 3DS Max 
        //versions behave differently. In any case, make sure there is a trailing separator
        Utilities::EnsureTrailingPathSeparator(pluginConfigDirectory);
        TSTR propertiesFilePath = pluginConfigDirectory + propertiesFileName;        

        //Determine the error message
        if (GetFileAttributes(propertiesFilePath) == INVALID_FILE_ATTRIBUTES)
        {
            //IGame properties file not present
            errorMessage.printf
                (
                PluginGlobals::GetString(IDS_EXPORT_IGAME_INITIALIZE_PROPERTIES_FILE_ERROR_MESSAGE_FORMAT), 
                (const TCHAR*)propertiesFilePath
                );
        }
        else
        {
            //Some other error
            errorMessage = PluginGlobals::GetString(IDS_EXPORT_IGAME_INITIALIZE_ERROR_MESSAGE);
        }

        ShowError(errorMessage);

        this->gameInterface->ReleaseIGame();
        
        return false;
    }
}

void OgreMaxSceneExporter::ExportScene()
{
    //Prepare the XML document into which the scene will be stored
    this->sceneXmlElement = (TiXmlElement*)this->sceneXmlDocument.LinkEndChild(new TiXmlElement(_T("scene")));
    this->sceneBindingsXmlElement = (TiXmlElement*)this->sceneBindingsXmlDocument.LinkEndChild(new TiXmlElement(_T("bindings")));

    ExportSceneProperties();
    ExportEnvironment();    
    ExportNodes();
    
    //Save generated scene XML document    
    this->sceneXmlDocument.SaveFile(this->exportFileName);

    if (!this->sceneBindingsXmlElement->NoChildren())
    {
        TSTR fileName = this->exportFileName + _T(".bindings");
        this->sceneBindingsXmlDocument.SaveFile(fileName);
    }
}

void OgreMaxSceneExporter::ExportMeshesAndMaterials()
{
    //Export meshes
    if (this->exportSettings.exportMeshes)
    {        
        MeshExporter meshExporter(this->sceneProperties, this->exportSettings, this->initialNodeStates, this->gameInterface, this->maxInterface);        
        meshExporter.ExportMeshes(this->entityMeshNodes, this->materials);
    }

    //Export materials
    if (this->exportSettings.exportMaterials)
    {
        MaterialExporter materialExporter(this->sceneProperties, this->exportSettings, this->gameInterface, this->maxInterface);
        materialExporter.ExportMaterials(this->materials);
    }
}

void OgreMaxSceneExporter::ExportPhysics()
{
    this->physicsExporter.Save(this->exportFileName);
}

void OgreMaxSceneExporter::ExportModels()
{
    TSTR fileName;

    for (std::list<ModelExporter*>::iterator model = this->modelExporters.begin();
        model != this->modelExporters.end();
        ++model)
    {
        fileName = this->exportSettings.modelPath;
        fileName += (*model)->GetModelName();
        fileName = Utilities::MakeModelName(fileName);

        (*model)->Save(fileName);
    }
}

void OgreMaxSceneExporter::ExportSceneProperties()
{       
    //Version
    this->sceneXmlElement->SetAttribute(_T("formatVersion"), _T("1.0"));
	
    //ID
    TSTR id;
    if (this->sceneProperties.GetValue(PropertyKeys::ID, id))
        this->sceneXmlElement->SetAttribute("id", id);
    
    //Scene manager
    TSTR sceneManager;
    if (this->sceneProperties.GetValue(PropertyKeys::SCENE_MANAGER, sceneManager))
        this->sceneXmlElement->SetAttribute(_T("sceneManager"), sceneManager);

    //Up axis
    TSTR upAxisText = _T("y");
    if (this->exportSettings.upAxis == UP_AXIS_Z)
        upAxisText = _T("z");
    this->sceneXmlElement->SetAttribute(_T("upAxis"), upAxisText);

    //Minimum Ogre version is 1.4 (Eihort)
    this->sceneXmlElement->SetAttribute(_T("minOgreVersion"), _T("1.4"));

    //Author
    TSTR author = Utilities::GetPropertyString(PROPSET_SUMMARYINFO, PIDSI_AUTHOR);
    if (author.isNull())
        author = PluginGlobals::GetString(IDS_XML_DOCUMENT_AUTHOR);
    this->sceneXmlElement->SetAttribute(_T("author"), author);

    //User data
    Utilities::ExportUserData(this->sceneXmlElement, this->sceneProperties, this->sceneSummary);

    //Physics simulator info
    if (this->physicsExporter.GetPhysicsExportType() == OPAL_PHYSICS)
    {        
        TiXmlElement* simulatorInfoXmlElement = this->physicsExporter.GetOpalSimulatorInfoElement();

        //Units per meter
        simulatorInfoXmlElement->SetAttribute(_T("unitsPerMeter"), Utilities::ToString(this->exportSettings.scaledWorldUnitsPerMeter));

        //Space type
        TSTR spaceType;
        if (this->sceneProperties.GetValue(PropertyKeys::PHYSICS_SPACE_TYPE, spaceType))
        {
            simulatorInfoXmlElement->SetAttribute(_T("spaceType"), spaceType);
            if (Utilities::IsSameNoCase(spaceType, _T("hash")))
            {
                TSTR hashLevel;
                if (this->sceneProperties.GetValue(PropertyKeys::PHYSICS_SPACE_HASH_MIN_LEVEL, hashLevel))
                    simulatorInfoXmlElement->SetAttribute(_T("hashMinLevel"), hashLevel);
                if (this->sceneProperties.GetValue(PropertyKeys::PHYSICS_SPACE_HASH_MAX_LEVEL, hashLevel))
                    simulatorInfoXmlElement->SetAttribute(_T("hashMaxLevel"), hashLevel);
            }
            else if (Utilities::IsSameNoCase(spaceType, _T("octree")))
            {
                TSTR octreeDepth;
                if (this->sceneProperties.GetValue(PropertyKeys::PHYSICS_SPACE_OCTREE_DEPTH, octreeDepth))
                    simulatorInfoXmlElement->SetAttribute(_T("octreeDepth"), octreeDepth);
            }
        }            

        //Contact group setup items
        TSTR contactGroupSetupCountText;
        if (this->sceneProperties.GetValue(PropertyKeys::CONTACT_GROUP_SETUP_COUNT, contactGroupSetupCountText))
        {        
            TiXmlElement* contactGroupsXmlElement = 
                (TiXmlElement*)simulatorInfoXmlElement->LinkEndChild(new TiXmlElement("ContactGroups"));

            TSTR contactGroupSetupKey;
            TSTR contactGroupSetup;

            int contactGroupSetupCount = _ttoi(contactGroupSetupCountText);
            for (int contactGroupSetupIndex = 0; contactGroupSetupIndex < contactGroupSetupCount; contactGroupSetupIndex++)
            {
                contactGroupSetupKey.printf(PropertyKeys::CONTACT_GROUP_SETUP_FORMAT.data(), contactGroupSetupIndex + 1);
                if (this->sceneProperties.GetValue(contactGroupSetupKey, contactGroupSetup))
                    Utilities::AddOpalContactGroupSetupElement(contactGroupsXmlElement, contactGroupSetup, this->sceneProperties);
            }
        }
    }
}

void OgreMaxSceneExporter::ExportEnvironment()
{
    this->environmentXmlElement = 
        (TiXmlElement*)this->sceneXmlElement->LinkEndChild(new TiXmlElement("environment"));

    //Environment range
    ExportEnvironmentRange();

    //Fog
    ExportEnvironmentFog();

    //Ambient color
    Point3 ambientColor = this->maxInterface->GetAmbient(DefaultTimeValue, FOREVER);
    Utilities::AddColorElement(this->environmentXmlElement, _T("colourAmbient"), ambientColor);

    //Background color
    Point3 backgroundColor = this->maxInterface->GetBackGround(DefaultTimeValue, FOREVER);
    Utilities::AddColorElement(this->environmentXmlElement, _T("colourBackground"), backgroundColor);

    //Shadows
    TSTR shadowTechnique;
    if (this->sceneProperties.GetValue(PropertyKeys::SHADOW_TECHNIQUE, shadowTechnique))
        ExportEnvironmentShadows(shadowTechnique);
}

void OgreMaxSceneExporter::ExportEnvironmentRange()
{
    //The near and far ranges are given as scene properties
    //As an alternative, the far range may be given by a user specified node, 
        
    //Near
    float environmentNear = 0;
    TSTR environmentNearText;
    if (this->sceneProperties.GetValue(PropertyKeys::ENVIRONMENT_NEAR, environmentNearText))
        environmentNear = Utilities::ToFloat(environmentNearText) * this->exportSettings.scale;

    //Far
    float environmentFar = -1;
    TSTR environmentFarText;
    if (this->sceneProperties.GetValue(PropertyKeys::ENVIRONMENT_FAR, environmentFarText))
        environmentFar = Utilities::ToFloat(environmentFarText) * this->exportSettings.scale;

    //See if a far node was specified. If so, use its bounding box to calculate the range
    TSTR environmentBoxName;
    if (this->sceneProperties.GetValue(PropertyKeys::ENVIRONMENT_BOX, environmentBoxName))
    {
        INode* environmentBoxMaxNode = this->maxInterface->GetINodeByName(environmentBoxName); 
        IGameNode* environmentBoxNode = (environmentBoxMaxNode) ? this->gameInterface->GetIGameNode(environmentBoxMaxNode) : NULL;
        if (environmentBoxNode)
        {
            this->ignoreNodes.push_back(environmentBoxNode->GetMaxNode());

            Box3 environmentBox;
            IGameObject* environmentBoxObject = environmentBoxNode->GetIGameObject();
            Utilities::InitializeData(environmentBoxObject, environmentBoxNode);
            if (Utilities::GetMeshLocalBoundingBox(environmentBoxObject, environmentBox))
                environmentFar = Utilities::GetBoxHalfSize(environmentBox) * this->exportSettings.scale;
        }
    }

    if (environmentNear >= 0 && environmentFar > environmentNear)
        Utilities::AddClippingElement(this->environmentXmlElement, environmentNear, environmentFar);
}

void OgreMaxSceneExporter::ExportEnvironmentFog()
{
    //Look for the first fog atmospheric
    int atmosphericCount = this->maxInterface->NumAtmospheric();
    for (int index = 0; index < atmosphericCount; index++)
    {
        Atmospheric* atmos = this->maxInterface->GetAtmospheric(index);
        if (atmos->ClassID() == Class_ID(FOG_CLASS_ID, 0)) 
        {
            //Found fog atmospheric
            StdFog* fog = (StdFog*)atmos;

            TiXmlElement* fogXmlElement = 
                (TiXmlElement*)this->environmentXmlElement->LinkEndChild(new TiXmlElement(_T("fog")));
			
            //Color
            Color diffuseColor = fog->GetColor(DefaultTimeValue);
            Utilities::AddColorElement(fogXmlElement, _T("colourDiffuse"), diffuseColor);

			if (fog->GetType() == 0)
            {
                //Standard fog

				//Mode
                TSTR mode;
                if (!this->sceneProperties.GetValue(PropertyKeys::FOG_MODE, mode))
                    mode = _T("linear");
                fogXmlElement->SetAttribute(_T("mode"), mode);
                
                //Linear range
                float start = fog->GetNear(DefaultTimeValue);
				fogXmlElement->SetAttribute(_T("linearStart"), Utilities::ToString(start));
                
                float end = fog->GetFar(DefaultTimeValue);
                fogXmlElement->SetAttribute(_T("linearEnd"), Utilities::ToString(end));

                //Exponential density
                TSTR density;
                if (this->sceneProperties.GetValue(PropertyKeys::FOG_EXP_DENSITY, density))
                    fogXmlElement->SetAttribute(_T("expDensity"), density);
            }

            break;
        }
    }
}

void OgreMaxSceneExporter::ExportEnvironmentShadows(const TSTR& shadowTechnique)
{
    TiXmlElement* shadowsXmlElement = 
        (TiXmlElement*)this->environmentXmlElement->LinkEndChild(new TiXmlElement(_T("shadows")));
	
    //Technique
    shadowsXmlElement->SetAttribute(_T("technique"), shadowTechnique.data());

    //Self shadowing
    TSTR selfShadow;
    if (this->sceneProperties.GetValue(PropertyKeys::SELF_SHADOW, selfShadow))
        shadowsXmlElement->SetAttribute(_T("selfShadow"), selfShadow);

    //Shadow color
    TSTR colorText;
    if (this->sceneProperties.GetValue(PropertyKeys::SHADOW_COLOR, colorText))
    {
        Point3 color = Utilities::ToPoint3(colorText);
        Utilities::AddColorElement(shadowsXmlElement, _T("colourShadow"), color);
    }

    //Texture size and count
    TSTR shadowTextureSize;
    bool hasShadowTextureSize = this->sceneProperties.GetValue(PropertyKeys::SHADOW_TEXTURE_SIZE, shadowTextureSize);
    TSTR shadowTextureCount;
    bool hasShadowTextureCount = this->sceneProperties.GetValue(PropertyKeys::SHADOW_TEXTURE_COUNT, shadowTextureCount);
    if (hasShadowTextureSize || hasShadowTextureCount)
    {
        TiXmlElement* shadowTexturesXmlElement = 
            (TiXmlElement*)shadowsXmlElement->LinkEndChild(new TiXmlElement(_T("shadowTextures")));

        //Texture size
        if (hasShadowTextureSize)
            shadowTexturesXmlElement->SetAttribute(_T("size"), shadowTextureSize);

        //Texture count
        if (hasShadowTextureCount)
            shadowTexturesXmlElement->SetAttribute(_T("count"), shadowTextureCount);
    }

	//Shadow camera setup
    TSTR shadowCameraType;    
    bool hasShadowCameraType = this->sceneProperties.GetValue(PropertyKeys::SHADOW_CAMERA_TYPE, shadowCameraType);
    TSTR shadowPlaneNodeName;
    bool hasShadowPlane = this->sceneProperties.GetValue(PropertyKeys::SHADOW_PLANE, shadowPlaneNodeName);
    INode* shadowPlaneMaxNode = this->maxInterface->GetINodeByName(shadowPlaneNodeName);
    if (hasShadowCameraType || (hasShadowPlane && shadowPlaneMaxNode))
    {
        TiXmlElement* shadowCameraSetupXmlElement = 
            (TiXmlElement*)shadowsXmlElement->LinkEndChild(new TiXmlElement(_T("shadowCameraSetup")));

        //Type
        if (hasShadowCameraType)
            shadowCameraSetupXmlElement->SetAttribute(_T("type"), shadowCameraType);

        //Plane, and possibly the type
        if (hasShadowPlane && shadowPlaneMaxNode)
        {
            IGameNode* shadowPlaneNode = this->gameInterface->GetIGameNode(shadowPlaneMaxNode);

            //Plane
            IGameObject* shadowPlaneObject = shadowPlaneNode->GetIGameObject();
            Utilities::InitializeData(shadowPlaneObject, shadowPlaneNode);
            Point4 plane;
            if (Utilities::GetMeshPlane(shadowPlaneObject, plane, true))
                Utilities::SetPlaneAttributes(shadowCameraSetupXmlElement, plane);

            //Set the shadow camera type if there is none
            if (!hasShadowCameraType)
                shadowCameraSetupXmlElement->SetAttribute(_T("type"), _T("planeOptimal"));
        }
    }
}

void OgreMaxSceneExporter::ExportNodes()
{
    bool exportSelected = (this->options & SCENE_EXPORT_SELECTED) != 0;

    std::vector<INode*> nodes;
    Utilities::GetTopLevelNodes(nodes, exportSelected);    

    //Export the nodes
    for (size_t index = 0; index < nodes.size(); index++)
        TraverseNode(this->gameInterface->GetIGameNode(nodes[index]));
}

bool OgreMaxSceneExporter::GetInitialNodeStates(INode* maxNode)
{   
    //Get node user properties
    KeyedValues properties;
    Utilities::GetNodeUserProperties(maxNode, properties);
        
    TSTR nodeExportType;
    properties.GetValue(PropertyKeys::TYPE, nodeExportType);
    if (!IgnoreNode(maxNode, nodeExportType))
    {
        //Make sure the node name is unique
        bool nodeUnique = this->initialNodeStatesByName.find(maxNode->GetName()) == this->initialNodeStatesByName.end();
        if (!nodeUnique)
        {
            if (this->suppressPrompts)
            {
                //Can't ask user for feedback. We assume this scene is in error
                return false;
            }

            TSTR title = PluginGlobals::GetString(IDS_EXPORT_ERROR_CAPTION);
            TSTR messageFormat = PluginGlobals::GetString(IDS_EXPORT_ERROR_DUPLICATE_NODE_NAME);        
            TSTR message;
            message.printf(messageFormat, maxNode->GetName());
            
            int result = Utilities::ShowMessage(message, title, MB_ICONQUESTION | MB_YESNOCANCEL);
            switch (result)
            {
                case IDYES: break; //User wants to export node
                case IDNO: return true; //User wants to skip the node and its children
                case IDCANCEL: return false; //User wants to cancel export process
            }
        }

        InitialNodeState* state = new InitialNodeState;
        this->initialNodeStates[maxNode] = state;
        if (nodeUnique)
            this->initialNodeStatesByName[maxNode->GetName()] = state;

        //Parse user properties
        state->properties = properties;

        //Get modifier stack state, to be restored later    
        Object* maxObject = maxNode->GetObjectRef();
        Utilities::GetModifierStackState(maxObject, state->modifierStackState);

        //Disable invalid modifiers    
        static const Class_ID modifierClassIDs[] =
            {
            Class_ID(OPTIMIZEOSM_CLASS_ID, 0) //The "optimize" modifier causes IGame to crash when InitializeData() is called
            };
        static const int modifierClassIDCount = COUNT_OF(modifierClassIDs);
        Utilities::SetModifiersState(maxObject, modifierClassIDs, modifierClassIDCount, false);

        //Get the unmorphed vertices at frame 0
        Utilities::EvalUnmorphedMeshVertices(maxNode, DefaultTimeValue, this->exportSettings.upAxis, state->unmorphedInitialVertices, &state->unmorphedInitialNormals);
                
        //Traverse child nodes
        bool result = true;
        int childCount = maxNode->NumberOfChildren();
        for (int childIndex = 0; childIndex < childCount && result; childIndex++)
            result = GetInitialNodeStates(maxNode->GetChildNode(childIndex));

        return result;
    }
    else
    {        
        return true;
    }
}

struct ChildNode
{
    ChildNode()
    {
        this->gameNode = NULL;
        this->idNumber = 0;
        this->order = 0;
    }

    IGameNode* gameNode;
    TSTR name;
    TSTR id;
    int idNumber;
    int order;
};

class ChildNodeSorter
{
public:
    enum SortType
    {
        SORT_NONE,
        SORT_NAME,
        SORT_ID,
        SORT_ID_NUMBER,
        SORT_ORDER
    };
    
    bool operator () (const ChildNode& n1, const ChildNode& n2) const
    {
        switch (sortType)
        {
            case ChildNodeSorter::SORT_NAME: return _tcsicmp(n1.name.data(), n2.name.data()) < 0; 
            case ChildNodeSorter::SORT_ID: return _tcsicmp(n1.id.data(), n2.id.data()) < 0; 
            case ChildNodeSorter::SORT_ID_NUMBER: return n1.idNumber < n2.idNumber; 
            case ChildNodeSorter::SORT_ORDER: return n1.order < n2.order;
            default: return true;
        }
    }

    static SortType sortType;
};
ChildNodeSorter::SortType ChildNodeSorter::sortType = ChildNodeSorter::SORT_NONE;

void OgreMaxSceneExporter::TraverseNode(IGameNode* gameNode)
{
    //Parse the node user property text
    InitialNodeStates::iterator initialNodeStateIterator = this->initialNodeStates.find(gameNode->GetMaxNode());
    if (initialNodeStateIterator != this->initialNodeStates.end())
    {    
        InitialNodeState* initialNodeState = initialNodeStateIterator->second;
        const KeyedValues& properties = initialNodeState->properties;
        
        //Get node type property
        TSTR nodeExportType;
        properties.GetValue(PropertyKeys::TYPE, nodeExportType);

        //If the node isn't an ignore node, process the node and its children
        if (!IgnoreNode(gameNode->GetMaxNode(), nodeExportType))
        {
            bool ignoreChildren = false;
            TiXmlElement* nodeXmlElement = ActOnNode(gameNode, properties, ignoreChildren);

            //Further determine whether children are to be ignored
            if (!ignoreChildren)
            {
                TSTR ignoreChildrenText;
                ignoreChildren = 
                    properties.GetValue(PropertyKeys::IGNORE_CHILDREN, ignoreChildrenText) &&
                    Utilities::ToBool(ignoreChildrenText);
            
                //Traverse child nodes if children are not to be ignored
                if (!ignoreChildren)
                {
                    //If a node XML element was created, push it onto the top of the
                    //node stack so it becomes the parent of all its child nodes
                    if (nodeXmlElement)
                        this->nodeXmlElementStack.push_back(nodeXmlElement);

                    //Determine child node sort type
                    TSTR childSorting;
                    properties.GetValue(PropertyKeys::CHILD_SORTING, childSorting);
                    if (Utilities::IsSameNoCase(childSorting, _T("name")))
                        ChildNodeSorter::sortType = ChildNodeSorter::SORT_NAME;                            
                    else if (Utilities::IsSameNoCase(childSorting, _T("id")))
                        ChildNodeSorter::sortType = ChildNodeSorter::SORT_ID;
                    else if (Utilities::IsSameNoCase(childSorting, _T("idNumber")))
                        ChildNodeSorter::sortType = ChildNodeSorter::SORT_ID_NUMBER;
                    else if (Utilities::IsSameNoCase(childSorting, _T("childOrder")))
                        ChildNodeSorter::sortType = ChildNodeSorter::SORT_ORDER;
                    else
                        ChildNodeSorter::sortType = ChildNodeSorter::SORT_NONE;

                    //Create the set that performs the child sorting
                    typedef std::multiset<ChildNode, ChildNodeSorter> SortedChildNodes;
                    SortedChildNodes childNodes;

                    //Collect children
                    int childCount = gameNode->GetChildCount();
                    for (int childIndex = 0; childIndex < childCount; childIndex++)
                    {
                        ChildNode childNode;
                        childNode.gameNode = gameNode->GetNodeChild(childIndex);
                        childNode.name = childNode.gameNode->GetName();
                        if (properties.GetValue(PropertyKeys::ID, childNode.id))
                            childNode.idNumber = Utilities::ToInt(childNode.id);
                        else
                        {
                            childNode.idNumber = childNode.gameNode->GetNodeID();
                            childNode.id.printf("%d", childNode.idNumber);
                        }
                        TSTR order;
                        if (properties.GetValue(PropertyKeys::CHILD_ORDER, order))
                            childNode.order = Utilities::ToInt(order);
                        childNodes.insert(childNode);
                    }

                    //Traverse sorted child nodes
                    for (SortedChildNodes::iterator childNode = childNodes.begin(); childNode != childNodes.end(); ++childNode)
                        TraverseNode(childNode->gameNode);
                    
                    //Remove the previously added XML element
                    if (nodeXmlElement)
                        this->nodeXmlElementStack.pop_back();
                }
            }
        }
    }
}

TiXmlElement* OgreMaxSceneExporter::ActOnNode(IGameNode* gameNode, const KeyedValues& properties, bool& ignoreChildren)
{
    IGameObject* gameObject = gameNode->GetIGameObject();
    Utilities::InitializeData(gameObject, gameNode);

    //Get node type property
    TSTR nodeExportType;
    if (!properties.GetValue(PropertyKeys::TYPE, nodeExportType))
        properties.GetValue(PropertyKeys::GRAPHICS_TYPE, nodeExportType);

    //Determine if the node is a model
    TSTR isModelText;
    bool isModel = 
        properties.GetValue(PropertyKeys::IS_MODEL, isModelText) && 
        Utilities::ToBool(isModelText);

    //The returned result
    TiXmlElement* nodeElement = NULL;

    //Act on the node
    if (isModel)
    {
        nodeElement = ExportModel(gameNode, gameObject, properties, nodeExportType, ignoreChildren);        
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("external")))
    {
        ExportExternal(gameNode, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("terrain")))
    {
        ExportTerrain(gameNode, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("skyBox")))
    {
        ExportSkyBox(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("skyDome")))
    {
        ExportSkyDome(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("skyPlane")))
    {
        ExportSkyPlane(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("modelInstance")))
    {
        nodeElement = ExportModelInstance(gameNode, properties);        
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("spring")))
    {
        ExportSpring(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (gameObject->GetMaxObject()->ClassID() == HINGE_CLASS_ID)
    {
        ExportHinge(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (gameObject->GetMaxObject()->ClassID() == POINT_POINT_CLASS_ID)
    {
        ExportPointPoint(gameNode, gameObject, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("gravity")))
    {
        ExportGravity(gameNode, properties);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("raysensor")))
    {
        ExportRaySensor(gameNode, properties);
        ignoreChildren = true;
    }
    else
    {
        //This is the only case that generates a 'node' element
        nodeElement = ExportNode(gameNode, gameObject, properties, nodeExportType, ignoreChildren);            
    }

    return nodeElement;
}

TiXmlElement* OgreMaxSceneExporter::ExportNode
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& nodeExportType,
    bool& ignoreChildren,
    TiXmlElement* nodeXmlElement,
    TiXmlElement* physicsSceneXmlElement,
    TiXmlElement* bindingsSceneXmlElement,
    size_t flags
    )
{
    Box3 localBox;
    gameObject->GetMaxObject()->GetDeformBBox(DefaultTimeValue, localBox);
    Point3 graphicsToPhysicsOffsetTranslation = localBox.Center() * this->exportSettings.scale;

    //Name
    TSTR name;
    if (!properties.GetValue(PropertyKeys::NAME, name))
        name = gameNode->GetName();

    //ID
    TSTR id;
    if (!properties.GetValue(PropertyKeys::ID, id))
        id = Utilities::ToString(gameNode->GetNodeID());

    //Position/rotation/scale    
    GMatrix localToWorldMatrix = GetNodeTM(gameNode, false);
    GMatrix localToParentMatrix = GetNodeTM(gameNode, true);
    Point3 position = localToParentMatrix.Translation() * this->exportSettings.scale;
    Point3 scaling = localToParentMatrix.Scaling();
    Quat rotation = localToParentMatrix.Rotation();

    if (Utilities::IsLightOrCamera(gameObject))
    {
        //If y is the current up axis, camera and light nodes must be 
        //rotated so that they continue to point in the direction of 
        //their negative z axis

        //The node rotation needs to be adjusted so that it always looks 
        //down its z axis            
        if (this->exportSettings.upAxis == UP_AXIS_Y)
        {
            AngAxis rotationAngAxis(1, 0, 0, DegToRad(90));
            rotation = Quat(rotationAngAxis) * rotation;                
        }             

        //Lights need to be face the opposite direction
        if (gameObject->GetIGameType() == IGameObject::IGAME_LIGHT)
        {
            AngAxis rotationAngAxis(1, 0, 0, DegToRad(180));
            rotation = Quat(rotationAngAxis) * rotation;
        }
    }

    //Export node if it isn't physics-only
    bool isPhysicsOnly = Utilities::IsSameNoCase(nodeExportType, _T("physics"));

    if (!isPhysicsOnly)
    {        
        if (nodeXmlElement == NULL)
        {
            nodeXmlElement = 
                (TiXmlElement*)GetCurrentNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("node")));
        }

        //Name
        nodeXmlElement->SetAttribute(_T("name"), name);

        //ID
        nodeXmlElement->SetAttribute(_T("id"), id);

        //Is target?
        TSTR isTarget = _T("true");
        if (!gameNode->IsTarget())
            isTarget = _T("false");
        nodeXmlElement->SetAttribute(_T("isTarget"), isTarget);

        //Visibility
        TSTR visibility;
        if (properties.GetValue(PropertyKeys::VISIBILITY, visibility))
            nodeXmlElement->SetAttribute(_T("visibility"), visibility);

        //Position/rotation/scale    
        if (!(flags & NO_NODE_TRANSFORM))
        {
            Utilities::AddXYZElement(nodeXmlElement, _T("scale"), scaling);            
            Utilities::AddXYZElement(nodeXmlElement, _T("position"), position);    
        }
        Utilities::AddRotationElement(nodeXmlElement, _T("rotation"), rotation);

        //Look/track target
        IGameNode* targetNode = targetNode = GetControllerLookAtNode(gameNode);
        if (!(flags & NO_NODE_TARGETS) && targetNode)
        {
            TSTR targetType;
            properties.GetValue(PropertyKeys::TARGET_TYPE, targetType);

            if (Utilities::IsSameNoCase(targetType, _T("look")))
            {
                //Look target
                TiXmlElement* targetNodeXmlElement = 
                    (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("lookTarget")));

                //Target node name
                targetNodeXmlElement->SetAttribute(_T("nodeName"), targetNode->GetName());

                //Local direction
                ExportLookAtDirection(gameNode, _T("localDirection"), targetNodeXmlElement);

                //Look at position
                GMatrix targetTransform = Utilities::GetNodeTM(targetNode, DefaultTimeValue, false);
                Point3 position = targetTransform.Translation() * this->exportSettings.scale;
                Utilities::AddXYZElement(targetNodeXmlElement, _T("position"), position);
            }
            else
            {
                //Track target
                TiXmlElement* targetNodeXmlElement = 
                    (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("trackTarget")));

                //Track node name
                targetNodeXmlElement->SetAttribute(_T("nodeName"), targetNode->GetName());

                //Local direction
                ExportLookAtDirection(gameNode, _T("localDirection"), targetNodeXmlElement);
            }
        }
        else if (gameNode->GetMaxNode()->GetTarget() && Utilities::IsLightOrCamera(gameObject))
        {
            //The node is a targeted light or camera
            //Export the target as a track target
            
            TiXmlElement* targetNodeXmlElement = 
                (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("trackTarget")));

            //Track node name
            targetNodeXmlElement->SetAttribute(_T("nodeName"), gameNode->GetMaxNode()->GetTarget()->GetName());

            //Local direction
            Point3 direction(0,0,-1);
            if (gameObject->GetIGameType() == IGameObject::IGAME_LIGHT)
                direction = Point3(0,0,1);
            Utilities::AddXYZElement(targetNodeXmlElement, _T("localDirection"), direction);
        }

        //User data 
        //It's very important that this appears before the node object
        //since clients of the scene loader may have object creation callbacks that
        //expect the node properties to be present
        Utilities::ExportUserData(nodeXmlElement, properties, gameNode);

        //Animations
        ExportAnimations(gameNode, gameObject, properties, nodeXmlElement);

        //Notes
        ExportNoteTracks(gameNode, properties, nodeXmlElement);
    }

    //Physics data
    TiXmlElement* physicsXmlElement = NULL;
    TiXmlElement* bindingsXmlElement = NULL;
    
    bool objectExportPhysics = false;
    TSTR objectExportPhysicsText;
    if (properties.GetValue(PropertyKeys::EXPORT_PHYSICS, objectExportPhysicsText))
        objectExportPhysics = Utilities::ToBool(objectExportPhysicsText);

    TSTR physicsShape;
    if (properties.GetValue(PropertyKeys::PHYSICS_SHAPE, physicsShape) || objectExportPhysics)
    {
        if (this->exportSettings.physicsExportType == OPAL_PHYSICS)
        {
            if (physicsSceneXmlElement == NULL)
                physicsSceneXmlElement = this->physicsExporter.GetRootNodeXmlElement();
            
            if (bindingsSceneXmlElement == NULL)
                bindingsSceneXmlElement = this->sceneBindingsXmlElement;

            //Physics to graphics binding
            if (!isPhysicsOnly)
            {
                bindingsXmlElement = (TiXmlElement*)bindingsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("physicsGraphicsBinding")));    
                bindingsXmlElement->SetAttribute(_T("physics"), name);
                bindingsXmlElement->SetAttribute(_T("graphics"), name);
            }
            
            //Solid element            
            physicsXmlElement = 
                (TiXmlElement*)physicsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("Solid")));

            //Name
            TiXmlElement* nameXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
            nameXmlElement->SetAttribute(_T("value"), name);

            //Enabled?
            TSTR physicsEnabled;
            if (properties.GetValue(PropertyKeys::PHYSICS_ENABLED, physicsEnabled))
            {
                TiXmlElement* physicsEnabledXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Enabled")));
                physicsEnabledXmlElement->SetAttribute(_T("value"), physicsEnabled);
            }

            //Export a bunch of things if the object isn't a plane
            if (!Utilities::IsSameNoCase(nodeExportType, _T("plane")))
            {
                //Is static?
                TSTR isStatic;
                if (properties.GetValue(PropertyKeys::IS_STATIC, isStatic))
                {
                    TiXmlElement* staticXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Static")));
                    staticXmlElement->SetAttribute(_T("value"), isStatic);
                }
                
                //Sleepiness
                TSTR sleepiness;
                if (properties.GetValue(PropertyKeys::SLEEPINESS, sleepiness))
                {
                    TiXmlElement* sleepinessXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Sleepiness")));
                    sleepinessXmlElement->SetAttribute(_T("value"), sleepiness);
                }

                //Linear damping
                TSTR linearDamping;
                if (properties.GetValue(PropertyKeys::LINEAR_DAMPING, linearDamping))
                {
                    TiXmlElement* linearDampingXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("LinearDamping")));
                    linearDampingXmlElement->SetAttribute(_T("value"), linearDamping);
                }

                //Angular damping
                TSTR angularDamping;
                if (properties.GetValue(PropertyKeys::ANGULAR_DAMPING, angularDamping))
                {
                    TiXmlElement* angularDampingXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("AngularDamping")));
                    angularDampingXmlElement->SetAttribute(_T("value"), angularDamping);
                }

                //Offset
                TiXmlElement* offsetXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Offset")));
                
                //Offset translation
                TiXmlElement* offsetTranslationXmlElement = (TiXmlElement*)offsetXmlElement->LinkEndChild(new TiXmlElement(_T("Transform")));
                offsetTranslationXmlElement->SetAttribute(_T("type"), _T("translate"));
                Point3 shapePosition(0,0,0);
                if (!(flags & NO_NODE_TRANSFORM))
                {
                    shapePosition = 
                        localToWorldMatrix.Translation() * this->exportSettings.scale;
                    if (!Utilities::IsPhysicsShapeMesh(properties))
                    {
                        Matrix3 rotationMatrix;
                        rotationMatrix.SetRotate(localToWorldMatrix.Rotation());
                        shapePosition += rotationMatrix * graphicsToPhysicsOffsetTranslation;
                    }
                }
                Utilities::SetXYZAttributes(offsetTranslationXmlElement, shapePosition);

                //Offset rotation
                TiXmlElement* offsetRotationXmlElement = (TiXmlElement*)offsetXmlElement->LinkEndChild(new TiXmlElement(_T("Transform")));
                offsetRotationXmlElement->SetAttribute(_T("type"), _T("rotate"));
                AngAxis shapeAngAxis(localToWorldMatrix.Rotation());

                offsetRotationXmlElement->SetAttribute(_T("angle"), Utilities::ToString(RadToDeg(-shapeAngAxis.angle)));
                offsetRotationXmlElement->SetAttribute(_T("x"), Utilities::ToString(shapeAngAxis.axis.x));
                offsetRotationXmlElement->SetAttribute(_T("y"), Utilities::ToString(shapeAngAxis.axis.y));
                offsetRotationXmlElement->SetAttribute(_T("z"), Utilities::ToString(shapeAngAxis.axis.z));

                //Make upright joint?
                TSTR makeUpright;
                if (properties.GetValue(PropertyKeys::MAKE_UPRIGHT, makeUpright) && Utilities::ToBool(makeUpright))
                {
                    //Joint element
                    TiXmlElement* physicsJointXmlElement = 
                        (TiXmlElement*)physicsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("Joint")));

                    //Name
                    TiXmlElement* jointNameXmlElement = (TiXmlElement*)physicsJointXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
                    jointNameXmlElement->SetAttribute(_T("value"), name);

                    //Type
                    TiXmlElement* jointTypeXmlElement = (TiXmlElement*)physicsJointXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
                    jointTypeXmlElement->SetAttribute(_T("value"), _T("upright"));

                    //References
                    TiXmlElement* jointReferencesXmlElement = (TiXmlElement*)physicsJointXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
                    jointReferencesXmlElement->SetAttribute(_T("solid0"), name);
                }

                //Add a volume sensor?
                TSTR hasVolumeSensor;
                if (properties.GetValue(PropertyKeys::HAS_VOLUME_SENSOR, hasVolumeSensor) && Utilities::ToBool(hasVolumeSensor))
                {
                    //Volume sensor element
                    TiXmlElement* physicsSensorXmlElement = 
                        (TiXmlElement*)physicsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("Sensor")));

                    //Name
                    TSTR sensorName;
                    if (!properties.GetValue(PropertyKeys::VOLUME_SENSOR_NAME, sensorName))
                        sensorName = name + _T("VolumeSensor");
                    TiXmlElement* sensorNameXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
                    sensorNameXmlElement->SetAttribute(_T("value"), sensorName);

                    //Type
                    TiXmlElement* sensorTypeXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
                    sensorTypeXmlElement->SetAttribute(_T("value"), _T("volume"));

                    //References
                    TiXmlElement* sensorReferencesXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
                    sensorReferencesXmlElement->SetAttribute(_T("solid"), name);

                    //Contact group
                    TSTR contactGroup;
                    if (Utilities::GetObjectContactGroupInteger(contactGroup, properties, this->sceneProperties))
                        Utilities::AddOpalContactGroupElement(physicsSensorXmlElement, contactGroup); 
                }

                //Add an acceleration sensor?
                TSTR hasAccelerationSensor;
                if (properties.GetValue(PropertyKeys::HAS_ACCELERATION_SENSOR, hasAccelerationSensor) && Utilities::ToBool(hasAccelerationSensor))
                {
                    //Acceleration sensor element
                    TiXmlElement* physicsSensorXmlElement = 
                        (TiXmlElement*)physicsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("Sensor")));

                    //Name
                    TSTR sensorName;
                    if (!properties.GetValue(PropertyKeys::ACCELERATION_SENSOR_NAME, sensorName))
                        sensorName = name + _T("AccelerationSensor");
                    TiXmlElement* sensorNameXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
                    sensorNameXmlElement->SetAttribute(_T("value"), sensorName);

                    //Type
                    TiXmlElement* sensorTypeXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
                    sensorTypeXmlElement->SetAttribute(_T("value"), _T("acceleration"));

                    //References
                    TiXmlElement* sensorReferencesXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
                    sensorReferencesXmlElement->SetAttribute(_T("solid"), name);
                    
                    //Contact group
                    TSTR contactGroup;
                    if (Utilities::GetObjectContactGroupInteger(contactGroup, properties, this->sceneProperties))
                        Utilities::AddOpalContactGroupElement(physicsSensorXmlElement, contactGroup); 
                }

                //Add an incline sensor?
                TSTR hasInclineSensor;
                if (properties.GetValue(PropertyKeys::HAS_INCLINE_SENSOR, hasInclineSensor) && Utilities::ToBool(hasInclineSensor))
                {
                    //Incline sensor element
                    TiXmlElement* physicsSensorXmlElement = 
                        (TiXmlElement*)physicsSceneXmlElement->LinkEndChild(new TiXmlElement(_T("Sensor")));

                    //Name
                    TSTR sensorName;
                    if (!properties.GetValue(PropertyKeys::INCLINE_SENSOR_NAME, sensorName))
                        sensorName = name + _T("InclineSensor");
                    TiXmlElement* sensorNameXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
                    sensorNameXmlElement->SetAttribute(_T("value"), sensorName);

                    //Type
                    TiXmlElement* sensorTypeXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
                    sensorTypeXmlElement->SetAttribute(_T("value"), _T("incline"));

                    //References
                    TiXmlElement* sensorReferencesXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
                    sensorReferencesXmlElement->SetAttribute(_T("solid"), name);

                    //Axis
                    //TODO: Add a way of specifying the axis
                    Utilities::AddXYZElement(physicsSensorXmlElement, _T("Axis"), Point3(1,0,0));

                    //Contact group
                    TSTR contactGroup;
                    if (Utilities::GetObjectContactGroupInteger(contactGroup, properties, this->sceneProperties))
                        Utilities::AddOpalContactGroupElement(physicsSensorXmlElement, contactGroup); 
                }
            }
        }
    }

    //The node object
    if (gameObject->GetIGameType() == IGameObject::IGAME_MAX_HELPER || 
        Utilities::IsSameNoCase(nodeExportType, _T("empty")))
    {
        //Do nothing
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("particleSystem")))
    {
        ExportParticleSystem(gameNode, gameObject, properties, name, id, nodeXmlElement);
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("billboardSet")))
    {
        ExportBillboardSet(gameNode, gameObject, properties, name, id, nodeXmlElement);
        ignoreChildren = true;
    }
    else if (Utilities::IsSameNoCase(nodeExportType, _T("plane")))
    {
        ExportPlane(gameNode, gameObject, properties, name, id, nodeXmlElement, physicsXmlElement);
    }
    else
    {        
        if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)            
        {
            ExportEntity
                (
                gameNode, 
                gameObject,
                properties, 
                name, 
                id, 
                nodeXmlElement, 
                physicsXmlElement, 
                bindingsXmlElement, 
                graphicsToPhysicsOffsetTranslation
                );
        }
        else if (gameObject->GetIGameType() == IGameObject::IGAME_LIGHT && nodeXmlElement)
            ExportLight(gameNode, gameObject, properties, name, id, nodeXmlElement);
        else if (gameObject->GetIGameType() == IGameObject::IGAME_CAMERA && nodeXmlElement)
            ExportCamera(gameNode, gameObject, properties, name, id, nodeXmlElement);
    }

    return nodeXmlElement;
}

void OgreMaxSceneExporter::ExportParticleSystem
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement
    )
{
    TiXmlElement* objectXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("particleSystem")));

    //Name
    objectXmlElement->SetAttribute(_T("name"), name.data());

    //ID
    objectXmlElement->SetAttribute(_T("id"), id.data());

    //File
    TSTR file;
    if (properties.GetValue(PropertyKeys::FILE, file))
        objectXmlElement->SetAttribute(_T("file"), file);

    //Render queue
    TSTR renderQueue;
    if (properties.GetValue(PropertyKeys::RENDER_QUEUE, renderQueue))
        objectXmlElement->SetAttribute(_T("renderQueue"), renderQueue);
}

void OgreMaxSceneExporter::ExportBillboardSet
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement
    )
{
    TiXmlElement* objectXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("billboardSet")));

    //Name
    objectXmlElement->SetAttribute(_T("name"), name.data());

    //ID
    objectXmlElement->SetAttribute(_T("id"), id.data());

    //Material
    IGameMaterial* gameMaterial = gameNode->GetNodeMaterial();
    ExportMaterialAttribute(objectXmlElement, gameMaterial); 

    //Width/height
    Box3 box;    
    if (Utilities::GetMeshLocalBoundingBox(gameObject, box))
    {
        Point3 size = box.Width() * this->exportSettings.scale;
        objectXmlElement->SetAttribute(_T("width"), Utilities::ToString(size.x));
        objectXmlElement->SetAttribute(_T("height"), Utilities::ToString(size.y));
    }

    //Type
    TSTR type;
    if (properties.GetValue(PropertyKeys::BILLBOARD_SET_TYPE, type))
        objectXmlElement->SetAttribute(_T("type"), type);

    //Origin
    TSTR origin;
    if (properties.GetValue(PropertyKeys::ORIGIN, origin))
        objectXmlElement->SetAttribute(_T("origin"), origin);

    //Common direction
    TSTR commonDirectionText;
    if (properties.GetValue(PropertyKeys::COMMON_DIRECTION, commonDirectionText))
    {
        Point3 commonDirection = Utilities::ToPoint3(commonDirectionText);
        Utilities::ConvertRawVertex(commonDirection, this->exportSettings.upAxis);
        Utilities::AddXYZElement(objectXmlElement, _T("commonDirection"), commonDirection);
    }

    //Common up vector
    TSTR commonUpVectorText;
    if (properties.GetValue(PropertyKeys::COMMON_UP_VECTOR, commonUpVectorText))
    {
        Point3 commonUpVector = Utilities::ToPoint3(commonUpVectorText);
        Utilities::ConvertRawVertex(commonUpVector, this->exportSettings.upAxis);
        Utilities::AddXYZElement(objectXmlElement, _T("commonUpVector"), commonUpVector);
    }

    //Rotation type
    TSTR rotationType;
    if (properties.GetValue(PropertyKeys::ROTATION_TYPE, rotationType))
        objectXmlElement->SetAttribute(_T("rotationType"), rotationType);

    //Render queue
    TSTR renderQueue;
    if (properties.GetValue(PropertyKeys::RENDER_QUEUE, renderQueue))
        objectXmlElement->SetAttribute(_T("renderQueue"), renderQueue);
    
    //Traverse child nodes, treating them as billboards
    TSTR rotationTypeLower = rotationType;
    rotationTypeLower.toLower();
    int childCount = gameNode->GetChildCount();
    for (int childIndex = 0; childIndex < childCount; childIndex++)
        ExportBillboard(gameNode->GetNodeChild(childIndex), gameNode, objectXmlElement, rotationTypeLower);
}

void OgreMaxSceneExporter::ExportBillboard
    (
    IGameNode* gameNode, 
    IGameNode* parentGameNode,
    TiXmlElement* parentXmlElement,
    const TSTR& rotationTypeLower
    )
{
    IGameObject* gameObject = gameNode->GetIGameObject();
    Utilities::InitializeData(gameObject, gameNode);

    TiXmlElement* objectXmlElement = 
        (TiXmlElement*)parentXmlElement->LinkEndChild(new TiXmlElement(_T("billboard")));

    //Parse the node user property text
    TSTR propertyText;
    gameNode->GetMaxNode()->GetUserPropBuffer(propertyText);
    KeyedValues properties(propertyText);

    //ID
    TSTR id;
    if (!properties.GetValue(PropertyKeys::ID, id))
        id = Utilities::ToString(gameNode->GetNodeID());
    objectXmlElement->SetAttribute(_T("id"), id);

    //Width/height
    Box3 box;    
    if (Utilities::GetMeshLocalBoundingBox(gameObject, box))
    {
        Point3 size = box.Width() * this->exportSettings.scale;
        objectXmlElement->SetAttribute(_T("width"), Utilities::ToString(size.x));
        objectXmlElement->SetAttribute(_T("height"), Utilities::ToString(size.y));
    }

    //Position. The position is relative to the parent billboard set
    GMatrix transform = Utilities::GetNodeTMRelativeTo(gameNode, parentGameNode, DefaultTimeValue);
    Point3 position = transform.Translation() * this->exportSettings.scale;
    Utilities::AddXYZElement(objectXmlElement, _T("position"), position);
    
    //Rotation. This is the rotation angle around the object's Z axis in world space
    transform = GetNodeTM(gameNode, false);
    Quat rotationQuat(transform.ExtractMatrix3());
    Point3 rotationAngles;
    rotationQuat.GetEuler(&rotationAngles.x, &rotationAngles.y, &rotationAngles.z);    
    float rotationAngle;
    if (this->exportSettings.upAxis == UP_AXIS_Z)
    {
        rotationAngle = rotationAngles.y;
        if (rotationTypeLower == _T("texcoord"))
            rotationAngle = -rotationAngle;
    }
    else
    {
        rotationAngle = rotationAngles.z;    
        if (Utilities::IsNull(rotationTypeLower) || rotationTypeLower == _T("vertex"))
            rotationAngle = -rotationAngle;
    }
    
    objectXmlElement->SetAttribute(_T("rotation"), Utilities::ToString(rotationAngle));
    
    //Diffuse color
    IGameMesh* gameMesh = (IGameMesh*)gameObject;
    Point3 diffuseColor(1,1,1);
    TSTR colorText;
    if (properties.GetValue(PropertyKeys::DIFFUSE_COLOR, colorText))
    {
        diffuseColor = Utilities::ToPoint3(colorText);
    }
    else if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)
    {
        //Get diffuse color from first vertex of the mesh
        if (!gameMesh->GetColorVertex(0, diffuseColor))
            diffuseColor = Point3(1,1,1);
    }
    Utilities::AddColorElement(objectXmlElement, _T("colourDiffuse"), diffuseColor);

    //Traverse child nodes, treating them as billboards
    int childCount = gameNode->GetChildCount();
    for (int childIndex = 0; childIndex < childCount; childIndex++)
        ExportBillboard(gameNode->GetNodeChild(childIndex), parentGameNode, parentXmlElement, rotationTypeLower);
}

void OgreMaxSceneExporter::ExportPlane
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement,
    TiXmlElement* physicsXmlElement
    )
{
    //Plane
    Point4 plane;
    Utilities::GetMeshPlane(gameObject, plane, false);
        
    //Export graphical plane if it has a graphical element
    if (nodeXmlElement)
    {
        TiXmlElement* objectXmlElement = 
            (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("plane")));

        //Name
        objectXmlElement->SetAttribute(_T("name"), name.data());

        //ID
        objectXmlElement->SetAttribute(_T("id"), id.data());

        //Plane
        //Plane normal
        Point3 planeNormal(plane.x, plane.y, plane.z);
        Utilities::AddXYZElement(objectXmlElement, "normal", planeNormal);
        
        //Plane distance
        float distance = plane.w;
        objectXmlElement->SetAttribute(_T("distance"), Utilities::ToString(distance));

        //Up vector
        Point3 upVector;
        if (abs(DotProd(planeNormal, this->upVector)) > .99f)
        {
            //Plane normal and world up vector point in very nearly the same
            //or opposite direction. Use the vector known to be perpendicular
            //to the world up vector
            upVector = this->upVectorPerpendicular; 
        }
        else
            upVector = this->upVector;
        Utilities::AddXYZElement(objectXmlElement, "upVector", upVector);
        
        //Width/height
        Box3 box;    
        if (Utilities::GetMeshLocalBoundingBox(gameObject, box))
        {
            Point3 size = box.Width() * this->exportSettings.scale;
            objectXmlElement->SetAttribute(_T("width"), Utilities::ToString(size.x));
            objectXmlElement->SetAttribute(_T("height"), Utilities::ToString(size.y));
        }
        
        //Number of segments
        IPoint2 segmentCount;
        if (Utilities::GetPlaneMeshSegmentCount(gameObject, segmentCount))
        {
            objectXmlElement->SetAttribute(_T("xSegments"), Utilities::ToString(segmentCount.x));
            objectXmlElement->SetAttribute(_T("ySegments"), Utilities::ToString(segmentCount.y));
        }

        //numTexCoordSets
        TSTR numTexCoordSets;
        if (properties.GetValue(PropertyKeys::NUM_TEX_COORD_SETS, numTexCoordSets))
            objectXmlElement->SetAttribute(_T("numTexCoordSets"), numTexCoordSets);

        //Tiling            
        IGameMaterial* gameMaterial = gameNode->GetNodeMaterial();
        Point2 tiling;        
        if (Utilities::GetTiling(gameMaterial, DefaultTimeValue, tiling))
        {
            objectXmlElement->SetAttribute(_T("uTile"), Utilities::ToString(tiling.x));
            objectXmlElement->SetAttribute(_T("vTile"), Utilities::ToString(tiling.y));
        }
        
        //Material
        ExportMaterialAttribute(objectXmlElement, gameMaterial); 

        //Normals
        TSTR normals;
        if (properties.GetValue(PropertyKeys::NORMALS, normals))
            objectXmlElement->SetAttribute(_T("normals"), normals);

        //Is static?
        TSTR isStatic;
        if (properties.GetValue(PropertyKeys::IS_STATIC, isStatic))
            objectXmlElement->SetAttribute(_T("static"), isStatic);

        //Render queue
        TSTR renderQueue;
        if (properties.GetValue(PropertyKeys::RENDER_QUEUE, renderQueue))
            objectXmlElement->SetAttribute(_T("renderQueue"), renderQueue);

        //Casts shadows?
        TSTR castShadows = _T("true");
        if (!gameNode->GetMaxNode()->CastShadows())
            castShadows = _T("false");
        objectXmlElement->SetAttribute(_T("castShadows"), castShadows);

        //Receives shadows?
        TSTR receiveShadows = _T("true");
        if (!gameNode->GetMaxNode()->RcvShadows())
            receiveShadows = _T("false");
        objectXmlElement->SetAttribute(_T("receiveShadows"), receiveShadows);

        //Hardware buffer usage
        Utilities::ExportHardwareBufferUsage(objectXmlElement, properties, _T("vertexBuffer"));
        Utilities::ExportHardwareBufferUsage(objectXmlElement, properties, _T("indexBuffer"));
    }

    //Export physical plane if it has a physical element
    if (physicsXmlElement)
    {
        if (this->exportSettings.physicsExportType == OPAL_PHYSICS)
        {
            //Shape
            TiXmlElement* shapeXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Shape")));
            shapeXmlElement->SetAttribute(_T("type"), _T("plane"));
            
            //Material
            Utilities::AddOpalShapeMaterialElement(shapeXmlElement, properties);

            //Shape dimensions
            TiXmlElement* dimensionsXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Dimensions")));
            dimensionsXmlElement->SetAttribute(_T("a"), Utilities::ToString(plane.x));
            dimensionsXmlElement->SetAttribute(_T("b"), Utilities::ToString(plane.y));
            dimensionsXmlElement->SetAttribute(_T("c"), Utilities::ToString(plane.z));
            dimensionsXmlElement->SetAttribute(_T("d"), Utilities::ToString(plane.w));
        }
    }
}

void OgreMaxSceneExporter::ExportEntity
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement,
    TiXmlElement* physicsXmlElement,
    TiXmlElement* bindingsXmlElement,
    Point3 graphicsToPhysicsOffsetTranslation
    )
{
    IGameMesh* gameMesh = (IGameMesh*)gameObject;
    IGameSkin* gameSkin = gameMesh->GetIGameSkin();

    //Export graphical entity if has a graphical element
    if (nodeXmlElement)
    {
        TiXmlElement* objectXmlElement = 
            (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("entity")));

        //Hold onto the node for later
        GameNodeAndObject nodeAndObject(gameNode, gameObject);
        this->entityMeshNodes.push_back(nodeAndObject);

        //Ignore bones from export process
        if (gameSkin != NULL)
        {
            int boneCount = gameSkin->GetTotalSkinBoneCount();
            for (int boneIndex = 0; boneIndex < boneCount; boneIndex++)
            {
                INode* boneNode = gameSkin->GetBone(boneIndex);
                IGameNode* gameBoneNode = this->gameInterface->GetIGameNode(boneNode);
                
                this->ignoreNodes.push_back(gameBoneNode->GetMaxNode());
            }
        }

        //Name
        objectXmlElement->SetAttribute(_T("name"), name.data());

        //ID
        objectXmlElement->SetAttribute(_T("id"), id.data());

        //Mesh file
        TSTR meshName = Utilities::GetMeshName(gameNode, properties);
        TSTR meshFile = Utilities::MakeMeshFileName(meshName, false);
        objectXmlElement->SetAttribute(_T("meshFile"), meshFile);

        //Is static?
        TSTR isStatic;
        if (properties.GetValue(PropertyKeys::IS_STATIC, isStatic))
            objectXmlElement->SetAttribute(_T("static"), isStatic);

        //Render queue
        TSTR renderQueue;
        if (properties.GetValue(PropertyKeys::RENDER_QUEUE, renderQueue))
            objectXmlElement->SetAttribute(_T("renderQueue"), renderQueue);

        //Casts shadows?
        TSTR castShadows = _T("true");
        if (!gameNode->GetMaxNode()->CastShadows())
            castShadows = _T("false");
        objectXmlElement->SetAttribute(_T("castShadows"), castShadows);

        //Receives shadows?
        TSTR receiveShadows = _T("true");
        if (!gameNode->GetMaxNode()->RcvShadows())
            receiveShadows = _T("false");
        objectXmlElement->SetAttribute(_T("receiveShadows"), receiveShadows);

        //Hardware buffer usage
        Utilities::ExportHardwareBufferUsage(objectXmlElement, properties, _T("vertexBuffer"));
        Utilities::ExportHardwareBufferUsage(objectXmlElement, properties, _T("indexBuffer"));
    }

    //Export physical entity if it has a physical element
    if (physicsXmlElement)
    {
        if (this->exportSettings.physicsExportType == OPAL_PHYSICS)
        {
            TSTR physicsShape;
            properties.GetValue(PropertyKeys::PHYSICS_SHAPE, physicsShape);
            if (!Utilities::IsSameNoCase(physicsShape, _T("none")))
            {
                //Shape
                TiXmlElement* shapeXmlElement = (TiXmlElement*)physicsXmlElement->LinkEndChild(new TiXmlElement(_T("Shape")));
                
                //Material             
                Utilities::AddOpalShapeMaterialElement(shapeXmlElement, properties);

                //Contact group
                TSTR contactGroup;
                if (Utilities::GetObjectContactGroupInteger(contactGroup, properties, this->sceneProperties))
                    Utilities::AddOpalContactGroupElement(shapeXmlElement, contactGroup);
                
                //Shape dimensions
                if (Utilities::IsSameNoCase(physicsShape, _T("box")))
                {
                    //Box shape
                    shapeXmlElement->SetAttribute(_T("type"), _T("box"));
                
                    //Dimensions
                    TiXmlElement* dimensionsXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Dimensions")));

                    Box3 box = GetBoundingBox(gameMesh);
                    Point3 boxSize = box.Width() * this->exportSettings.scale;
                    Utilities::SetXYZAttributes(dimensionsXmlElement, boxSize);
                }
                else if (Utilities::IsSameNoCase(physicsShape, _T("sphere")))
                {
                    //Sphere shape
                    shapeXmlElement->SetAttribute(_T("type"), _T("sphere"));
                
                    //Dimensions
                    TiXmlElement* dimensionsXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Dimensions")));

                    float radius = Utilities::GetBoundingSphereRadius(gameMesh) * this->exportSettings.scale;
                    dimensionsXmlElement->SetAttribute(_T("radius"), Utilities::ToString(radius));
                }
                else if (Utilities::IsSameNoCase(physicsShape, _T("capsule")) || 
                    Utilities::IsSameNoCase(physicsShape, _T("fullCapsule")))
                {
                    //Capsule shape
                    shapeXmlElement->SetAttribute(_T("type"), _T("capsule"));

                    //Add extra shape rotation if necessary
                    if (this->exportSettings.upAxis == UP_AXIS_Y)
                    {
                        TiXmlElement* offsetXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Offset")));
                        TiXmlElement* transformXmlElement = (TiXmlElement*)offsetXmlElement->LinkEndChild(new TiXmlElement(_T("Transform")));
                        transformXmlElement->SetAttribute(_T("type"), _T("euler"));
                        transformXmlElement->SetAttribute(_T("x"), _T("90"));
                        transformXmlElement->SetAttribute(_T("y"), _T("0"));
                        transformXmlElement->SetAttribute(_T("z"), _T("0"));
                    }
                
                    //Dimensions
                    TiXmlElement* dimensionsXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Dimensions")));

                    float radius = Utilities::GetBoundingSphereRadius(gameMesh, Point3(1,1,0)) * this->exportSettings.scale;
                    dimensionsXmlElement->SetAttribute(_T("radius"), Utilities::ToString(radius));

                    float length = Utilities::GetMeshLength(gameMesh) * this->exportSettings.scale;
                    if (Utilities::IsSameNoCase(physicsShape, _T("capsule")))
                        length -= radius * 2;
                    dimensionsXmlElement->SetAttribute(_T("length"), Utilities::ToString(length));
                }
                else
                {
                    //Mesh shape
                    shapeXmlElement->SetAttribute(_T("type"), _T("mesh"));
                    
                    //Dimensions
                    TiXmlElement* dimensionsXmlElement = (TiXmlElement*)shapeXmlElement->LinkEndChild(new TiXmlElement(_T("Dimensions")));

                    TiXmlElement* facesXmlElement = (TiXmlElement*)dimensionsXmlElement->LinkEndChild(new TiXmlElement(_T("Triangles")));                
                    facesXmlElement->SetAttribute(_T("count"), Utilities::ToString(gameMesh->GetNumberOfFaces()));
                    for (int faceIndex = 0; faceIndex < gameMesh->GetNumberOfFaces(); faceIndex++)
                    {
                        FaceEx* face = gameMesh->GetFace(faceIndex);
                        Utilities::AddFaceElement(facesXmlElement, _T("Tri"), face);
                    }

                    TiXmlElement* verticesXmlElement = (TiXmlElement*)dimensionsXmlElement->LinkEndChild(new TiXmlElement(_T("Vertices")));
                    verticesXmlElement->SetAttribute(_T("count"), Utilities::ToString(gameMesh->GetNumberOfVerts()));
                    for (int vertexIndex = 0; vertexIndex < gameMesh->GetNumberOfVerts(); vertexIndex++)
                    {
                        Point3 vertex = gameMesh->GetVertex(vertexIndex, true) * this->exportSettings.scale;
                        Utilities::AddXYZElement(verticesXmlElement, _T("Vertex"), vertex);
                    }
                }
            }
        }

        //Physics to graphics binding offset
        //Not needed if a physics shape is a mesh since the vertices fully define the
        //position of the volume
        if (!Utilities::IsPhysicsShapeMesh(properties))
        {
            Point3 physicsToGraphicsOffsetTranslation = -graphicsToPhysicsOffsetTranslation;
            TiXmlElement* offsetXmlElement = (TiXmlElement*)bindingsXmlElement->LinkEndChild(new TiXmlElement(_T("offset")));    
            Utilities::AddXYZElement(offsetXmlElement, _T("translation"), physicsToGraphicsOffsetTranslation);
            //Utilities::AddRotationElement(offsetXmlElement, _T("rotation"), physicsToGraphicsOffset.Rotation());
        }
    }
}

void OgreMaxSceneExporter::ExportLight
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement
    )
{
    ObjectState objectState = gameNode->GetMaxNode()->EvalWorldState(DefaultTimeValue);
    GenLight* maxLight = (GenLight*)objectState.obj;
    LightState lightState;
    maxLight->EvalLightState(DefaultTimeValue, FOREVER, &lightState);

    TiXmlElement* objectXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("light")));

    //Name
    objectXmlElement->SetAttribute(_T("name"), name.data());

    //ID
    objectXmlElement->SetAttribute(_T("id"), id.data());

    //Light type
    TSTR lightType;
    bool isSpotLight = false;
    switch (lightState.type)
    {
        case FSPOT_LIGHT:
        case TSPOT_LIGHT: isSpotLight = true; lightType = _T("spot"); break;
        case DIR_LIGHT:
        case TDIR_LIGHT: lightType = _T("directional"); break;
        default: lightType = _T("point"); break;
    }
    objectXmlElement->SetAttribute(_T("type"), lightType);

    //Is visible?
    if (gameNode->IsNodeHidden())
        objectXmlElement->SetAttribute(_T("visible"), _T("false"));
    
    //Casts shadows?
    if (!lightState.shadow)
        objectXmlElement->SetAttribute(_T("castShadows"), _T("false"));    

    //Diffuse color
    Utilities::AddColorElement(objectXmlElement, _T("colourDiffuse"), lightState.color);

    //Attenuation
    TiXmlElement* attenuationXmlElement = 
        (TiXmlElement*)objectXmlElement->LinkEndChild(new TiXmlElement(_T("lightAttenuation")));    
    float range = lightState.attenEnd * this->exportSettings.scale;
    attenuationXmlElement->SetAttribute(_T("range"), Utilities::ToString(range));

    //Spotlight range
    if (isSpotLight)
    {
        TiXmlElement* rangeXmlElement = 
            (TiXmlElement*)objectXmlElement->LinkEndChild(new TiXmlElement(_T("lightRange")));        
        rangeXmlElement->SetAttribute(_T("inner"), Utilities::ToString(DegToRad(lightState.hotsize)));        
        rangeXmlElement->SetAttribute(_T("outer"), Utilities::ToString(DegToRad(lightState.fallsize)));
    }
}

void OgreMaxSceneExporter::ExportCamera
    (
    IGameNode* gameNode, 
    IGameObject* gameObject,
    const KeyedValues& properties, 
    const TSTR& name, 
    const TSTR& id, 
    TiXmlElement* nodeXmlElement
    )
{
    IGameCamera* gameCamera = (IGameCamera*)gameObject;

    TiXmlElement* objectXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("camera")));

    //Name
    objectXmlElement->SetAttribute(_T("name"), name.data());

    //ID
    objectXmlElement->SetAttribute(_T("id"), id.data());

    CameraState cameraState;
    CameraObject* cameraObject = (CameraObject*)gameCamera->GetMaxObject();
    cameraObject->EvalCameraState(DefaultTimeValue, FOREVER, &cameraState);            

    //Field of view
    objectXmlElement->SetAttribute(_T("fov"), Utilities::ToString(cameraState.fov));

    //Projection type
    if (cameraState.isOrtho)
        objectXmlElement->SetAttribute(_T("projectionType"), _T("orthographic"));

    //Clipping
    float nearClip = cameraState.hither * this->exportSettings.scale;
    float farClip = cameraState.yon * this->exportSettings.scale;
    Utilities::AddClippingElement(objectXmlElement, nearClip, farClip);
}

void OgreMaxSceneExporter::ExportAnimations(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties, TiXmlElement* nodeXmlElement)
{
    IGameControl* gameControl = gameNode->GetIGameControl();
    std::vector<TimeValue> allKeyTimes;    
    Utilities::GetPRSKeyTimes(gameNode, allKeyTimes);
    if (!allKeyTimes.empty() && allKeyTimes[0] != 0)
    {   
        //There's no key time for time 0. Add it

        //Since there's no vector::push_front, we have to achieve the same effect with this code
        std::vector<TimeValue> tempKeyTimes;
        tempKeyTimes.resize(allKeyTimes.size() + 1);
        tempKeyTimes[0] = 0;
        std::copy(allKeyTimes.begin(), allKeyTimes.end(), &tempKeyTimes[1]);
        allKeyTimes.swap(tempKeyTimes);
    }

    //Exit early if there are no animations
    if (allKeyTimes.size() <= 1)
        return;
    
    //Create "animations" and "animation" XML elements
    TiXmlElement* animationsXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("animations")));

    TiXmlElement* oneAnimationXmlElement = 
        (TiXmlElement*)animationsXmlElement->LinkEndChild(new TiXmlElement(_T("animation")));

    //Set animation name
    TSTR animationName;
    if (!properties.GetValue(PropertyKeys::ANIMATION_NAME, animationName))
        animationName = gameNode->GetName();
    oneAnimationXmlElement->SetAttribute(_T("name"), animationName);

    //Set enabled
    TSTR enabled;
    if (!properties.GetValue(PropertyKeys::ANIMATION_ENABLED, enabled))
        enabled = _T("true");
    oneAnimationXmlElement->SetAttribute(_T("enable"), enabled);

    //Set looping
    TSTR looping;
    if (!properties.GetValue(PropertyKeys::ANIMATION_LOOPED, looping))
        looping = _T("true");
    oneAnimationXmlElement->SetAttribute(_T("loop"), looping);

    //Set interpolation mode
    Control* positionController = gameControl ? gameControl->GetMaxControl(IGAME_POS) : NULL;    
    if (positionController != NULL)
    {
        TSTR interpolationMode;
        if (positionController->ClassID() == Class_ID(LININTERP_POSITION_CLASS_ID, 0))
            interpolationMode = _T("linear");
        else
            interpolationMode = _T("spline");
        oneAnimationXmlElement->SetAttribute(_T("interpolationMode"), interpolationMode);
    }
    
    //Set rotation interpolation mode
    Control* rotationController = gameControl ? gameControl->GetMaxControl(IGAME_ROT) : NULL;        
    if (rotationController != NULL)
    {
        TSTR rotationInterpolationMode;
        if (rotationController->ClassID() == Class_ID(LININTERP_ROTATION_CLASS_ID, 0))
            rotationInterpolationMode = _T("linear");
        else
            rotationInterpolationMode = _T("spherical");
        oneAnimationXmlElement->SetAttribute(_T("rotationInterpolationMode"), rotationInterpolationMode);
    }

    //Export all the keys
    for (size_t i = 0; i < allKeyTimes.size(); i++)
    {   
        TimeValue keyTime = allKeyTimes[i];

        //Create the key XML element
        TiXmlElement* keyXmlElement = 
            (TiXmlElement*)oneAnimationXmlElement->LinkEndChild(new TiXmlElement(_T("keyframe")));

        //Set current key time
        keyXmlElement->SetAttribute(_T("time"), Utilities::ToString(TicksToSec(keyTime)));

        //Get transform
        GMatrix transform = Utilities::GetNodeTM(gameNode, keyTime, false);

        //Export the keys at the current time        
        Point3 positionKey = transform.Translation() * this->exportSettings.scale;
        Utilities::AddXYZElement(keyXmlElement, _T("translation"), positionKey);
        
        Quat rotationKey = transform.Rotation();
        Utilities::AddRotationElement(keyXmlElement, _T("rotation"), rotationKey);
        
        Point3 scaleKey = transform.Scaling();
        Utilities::AddXYZElement(keyXmlElement, _T("scale"), scaleKey);
    }

    //Animation length, in seconds
    TimeValue animationLength = allKeyTimes.back();
    oneAnimationXmlElement->SetAttribute(_T("length"), Utilities::ToString(TicksToSec(animationLength)));
}

void OgreMaxSceneExporter::ExportNoteTracks(IGameNode* gameNode, const KeyedValues& properties, TiXmlElement* nodeXmlElement)
{
    INode* maxNode = gameNode->GetMaxNode();
    if (!maxNode->HasNoteTracks())
        return;

    //Create "noteTracks" element
    TiXmlElement* noteTracksXmlElement = 
        (TiXmlElement*)nodeXmlElement->LinkEndChild(new TiXmlElement(_T("noteTracks")));

    //Export note tracks
    for (int trackIndex = 0; trackIndex < maxNode->NumNoteTracks(); trackIndex++)
    {
        //Get the note track
        DefNoteTrack* noteTrack = (DefNoteTrack*)maxNode->GetNoteTrack(trackIndex);

        //Create "noteTrack" element
        TiXmlElement* oneNoteTrackXmlElement = 
            (TiXmlElement*)noteTracksXmlElement->LinkEndChild(new TiXmlElement(_T("noteTrack")));

        //Set the note track name
        TSTR noteTrackNameKey;
        noteTrackNameKey.printf(PropertyKeys::NOTE_TRACK_NAME_FORMAT.data(), trackIndex + 1);
        
        TSTR noteTrackName;
        if (properties.GetValue(noteTrackNameKey, noteTrackName))
            oneNoteTrackXmlElement->SetAttribute(_T("name"), noteTrackName);

        //Export notes for the track
        for (int keyIndex = 0; keyIndex < noteTrack->keys.Count(); keyIndex++)
        {
            //Create the note XML element
            TiXmlElement* noteXmlElement = 
                (TiXmlElement*)oneNoteTrackXmlElement->LinkEndChild(new TiXmlElement(_T("note")));

            //Set the time
            noteXmlElement->SetAttribute(_T("time"), Utilities::ToString(TicksToSec(noteTrack->keys[keyIndex]->time)));
            
            //Add the note text
            TiXmlText* textNode = (TiXmlText*)noteXmlElement->LinkEndChild(new TiXmlText(noteTrack->keys[keyIndex]->note.data()));
            textNode->SetCDATA(true);
        }
    }
}

void OgreMaxSceneExporter::ExportExternal(IGameNode* gameNode, const KeyedValues& properties)
{
    //Ensure 'externals' XML element exists
    if (!this->externalsXmlElement)
    {
        this->externalsXmlElement = 
            (TiXmlElement*)this->sceneXmlElement->LinkEndChild(new TiXmlElement(_T("externals")));
    }

    TiXmlElement* externalItemXmlElement = 
        (TiXmlElement*)this->externalsXmlElement->LinkEndChild(new TiXmlElement(_T("item")));

    //Name
    TSTR name;
    if (!properties.GetValue(PropertyKeys::NAME, name))
        name = gameNode->GetName();
    externalItemXmlElement->SetAttribute(_T("name"), name);

    //Item type
    TSTR itemType;
    if (properties.GetValue(PropertyKeys::ITEM_TYPE, itemType))
        externalItemXmlElement->SetAttribute(_T("type"), itemType);

    //File
    TSTR file;
    if (properties.GetValue(PropertyKeys::FILE, file))
    {
        TiXmlElement* fileXmlElement = 
            (TiXmlElement*)externalItemXmlElement->LinkEndChild(new TiXmlElement(_T("file")));

        fileXmlElement->SetAttribute(_T("name"), file);
    }

    //Position/rotation/scale
    GMatrix transform = GetNodeTM(gameNode, false);
    Point3 position = transform.Translation() * this->exportSettings.scale;
    Utilities::AddXYZElement(externalItemXmlElement, _T("position"), position);
    Utilities::AddRotationElement(externalItemXmlElement, _T("rotation"), transform.Rotation());
    Utilities::AddXYZElement(externalItemXmlElement, _T("scale"), transform.Scaling());
    
    //User data
    Utilities::ExportUserData(externalItemXmlElement, properties, gameNode);
}

void OgreMaxSceneExporter::ExportTerrain(IGameNode* gameNode, const KeyedValues& properties)
{
    TSTR file;
    if (!this->terrainXmlElement && properties.GetValue(PropertyKeys::FILE, file))
    {
        this->terrainXmlElement = 
            (TiXmlElement*)this->sceneXmlElement->LinkEndChild(new TiXmlElement(_T("terrain")));

        //Data file
        this->terrainXmlElement->SetAttribute(_T("dataFile"), file);

        //User data
        Utilities::ExportUserData(this->terrainXmlElement, properties, gameNode);
    }
}

void OgreMaxSceneExporter::ExportSkyBox(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (!this->skyBoxXmlElement)
    {
        this->skyBoxXmlElement = 
            (TiXmlElement*)this->environmentXmlElement->LinkEndChild(new TiXmlElement(_T("skyBox")));

        //Rotation
        Quat rotation = GetSkyRotation(gameNode);
        Utilities::AddRotationElement(this->skyBoxXmlElement, _T("rotation"), rotation);

        //Enable?
        TSTR enable;
        if (properties.GetValue(PropertyKeys::ENABLE, enable))            
            this->skyBoxXmlElement->SetAttribute(_T("enable"), enable);

        //Material
        IGameMaterial* gameMaterial = gameNode->GetNodeMaterial();
        ExportMaterialAttribute(this->skyBoxXmlElement, gameMaterial);
        
        //Distance
        Box3 box;    
        if (Utilities::GetMeshLocalBoundingBox(gameObject, box))
        {
            float distance = Utilities::GetBoxHalfSize(box) * this->exportSettings.scale;
            this->skyBoxXmlElement->SetAttribute(_T("distance"), Utilities::ToString(distance));
        }

        //Draw first (true or false)
        TSTR drawFirst;
        if (properties.GetValue(PropertyKeys::DRAW_FIRST, drawFirst))            
            this->skyBoxXmlElement->SetAttribute(_T("drawFirst"), drawFirst);

        //User data
        Utilities::ExportUserData(this->skyBoxXmlElement, properties, gameNode);
    }
}

void OgreMaxSceneExporter::ExportSkyDome(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (!this->skyDomeXmlElement)
    {
        this->skyDomeXmlElement = 
            (TiXmlElement*)this->environmentXmlElement->LinkEndChild(new TiXmlElement(_T("skyDome")));

        //Rotation
        Quat rotation = GetSkyRotation(gameNode);
        Utilities::AddRotationElement(this->skyDomeXmlElement, _T("rotation"), rotation);

        //Enable?
        TSTR enable;
        if (properties.GetValue(PropertyKeys::ENABLE, enable))            
            this->skyDomeXmlElement->SetAttribute(_T("enable"), enable);

        //Material
        IGameMaterial* gameMaterial = gameNode->GetNodeMaterial();
        ExportMaterialAttribute(this->skyDomeXmlElement, gameMaterial);
        
        //Curvature
        TSTR curvature;
        if (properties.GetValue(PropertyKeys::CURVATURE, curvature))            
            this->skyDomeXmlElement->SetAttribute(_T("curvature"), curvature);

        //Tiling
        float tiling;
        if (Utilities::GetTiling(gameMaterial, DefaultTimeValue, tiling))
            this->skyDomeXmlElement->SetAttribute(_T("tiling"), Utilities::ToString(tiling));
        
        //Distance
        Box3 box;    
        if (Utilities::GetMeshLocalBoundingBox(gameObject, box))
        {
            float distance = Utilities::GetBoxHalfSize(box) * this->exportSettings.scale;
            this->skyDomeXmlElement->SetAttribute(_T("distance"), Utilities::ToString(distance));
        }
        
        //Draw first (true or false)
        TSTR drawFirst;
        if (properties.GetValue(PropertyKeys::DRAW_FIRST, drawFirst))            
            this->skyDomeXmlElement->SetAttribute(_T("drawFirst"), drawFirst);

        //Segment counts
        IPoint2 segmentCount;
        if (Utilities::GetSphereSegmentCount(gameObject, properties, DefaultTimeValue, segmentCount))
        {
            this->skyDomeXmlElement->SetAttribute(_T("xSegments"), Utilities::ToString(segmentCount.x));
            this->skyDomeXmlElement->SetAttribute(_T("ySegments"), Utilities::ToString(segmentCount.y));
        }

        //User data
        Utilities::ExportUserData(this->skyDomeXmlElement, properties, gameNode);
    }
}

void OgreMaxSceneExporter::ExportSkyPlane(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (!this->skyPlaneXmlElement)
    {
        this->skyPlaneXmlElement = 
            (TiXmlElement*)this->environmentXmlElement->LinkEndChild(new TiXmlElement(_T("skyPlane")));

        //Enable?
        TSTR enable;
        if (properties.GetValue(PropertyKeys::ENABLE, enable))            
            this->skyPlaneXmlElement->SetAttribute(_T("enable"), enable);

        //Material
        IGameMaterial* gameMaterial = gameNode->GetNodeMaterial();
        ExportMaterialAttribute(this->skyPlaneXmlElement, gameMaterial);

        //Plane
        Point4 plane;
        if (Utilities::GetMeshPlane(gameObject, plane, true))
            Utilities::SetPlaneAttributes(this->skyPlaneXmlElement, plane);
        
        //Scale
        TSTR scale;
        if (properties.GetValue(PropertyKeys::SCALE, scale))            
            this->skyPlaneXmlElement->SetAttribute(_T("scale"), scale);

        //Bow
        TSTR bow;
        if (properties.GetValue(PropertyKeys::BOW, bow))            
            this->skyPlaneXmlElement->SetAttribute(_T("bow"), bow);

        //Tiling
        float tiling;
        if (Utilities::GetTiling(gameMaterial, DefaultTimeValue, tiling))
            this->skyPlaneXmlElement->SetAttribute(_T("tiling"), Utilities::ToString(tiling));

        //Draw first (true or false)
        TSTR drawFirst;
        if (properties.GetValue(PropertyKeys::DRAW_FIRST, drawFirst))            
            this->skyPlaneXmlElement->SetAttribute(_T("drawFirst"), drawFirst);

        //Number of segments
        IPoint2 segmentCount;
        if (Utilities::GetPlaneMeshSegmentCount(gameObject, segmentCount))
        {
            this->skyPlaneXmlElement->SetAttribute(_T("xSegments"), Utilities::ToString(segmentCount.x));
            this->skyPlaneXmlElement->SetAttribute(_T("ySegments"), Utilities::ToString(segmentCount.y));
        }

        //User data
        Utilities::ExportUserData(this->skyPlaneXmlElement, properties, gameNode);
    }
}

TiXmlElement* OgreMaxSceneExporter::ExportModel
    (
    IGameNode* gameNode, 
    IGameObject* gameObject, 
    const KeyedValues& properties, 
    const TSTR& nodeExportType, 
    bool& ignoreChildren
    )
{
    //Model name
    TSTR modelName;
    if (!properties.GetValue(PropertyKeys::MODEL_NAME, modelName))
        modelName = gameNode->GetName();
    
    ModelExporter* modelExporter = new ModelExporter(modelName, gameNode, this->exportSettings.physicsExportType);
    this->modelExporters.push_back(modelExporter);

    //Export nodes, using the ModelExporter's root element, so that all the 
    //exported nodes go into the model. 
    //Also, disable the root transform and node targets from being exported
    ExportNode
        (
        gameNode,
        gameObject,
        properties, 
        nodeExportType,
        ignoreChildren,
        modelExporter->GetRootNodeXmlElement(),
        modelExporter->GetRootPhysicsXmlElement(),
        modelExporter->GetRootBindingsXmlElement(),
        NO_NODE_TRANSFORM | NO_NODE_TARGETS
        );

    return modelExporter->GetRootNodeXmlElement();
}

TiXmlElement* OgreMaxSceneExporter::ExportModelInstance(IGameNode* gameNode, const KeyedValues& properties)
{   
    TiXmlElement* modelInstanceXmlElement = 
        (TiXmlElement*)GetCurrentNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("modelInstance")));

    //Name
    TSTR name;
    if (!properties.GetValue(PropertyKeys::NAME.data(), name))
        name = gameNode->GetName();
    modelInstanceXmlElement->SetAttribute(_T("name"), name);

    //Model name
    TSTR modelName;
    properties.GetValue(PropertyKeys::MODEL_NAME, modelName);
    modelName = Utilities::MakeModelName(modelName);
    modelInstanceXmlElement->SetAttribute(_T("modelName"), modelName);

    //Position/rotation/scale    
    GMatrix transform = GetNodeTM(gameNode, true);
    Point3 position = transform.Translation() * this->exportSettings.scale;
    Point3 scaling = transform.Scaling();
    Quat rotation = transform.Rotation();

    Utilities::AddXYZElement(modelInstanceXmlElement, _T("scale"), scaling);
    Utilities::AddRotationElement(modelInstanceXmlElement, _T("rotation"), rotation);
    Utilities::AddXYZElement(modelInstanceXmlElement, _T("position"), position);    
        
    return modelInstanceXmlElement;
}

void OgreMaxSceneExporter::ExportHinge(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (this->exportSettings.physicsExportType == NO_PHYSICS)
        return;

    //Get joint type property
    TSTR jointType;
    if (!properties.GetValue(PropertyKeys::TYPE, jointType))
        jointType = _T("hinge");

    IPropertyContainer* propertyContainer = gameObject->GetIPropertyContainer();
    
    //Joint element
    TiXmlElement* jointXmlElement = 
        (TiXmlElement*)this->physicsExporter.GetRootNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("Joint")));

    //Name
    TiXmlElement* nameXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
    nameXmlElement->SetAttribute(_T("value"), gameNode->GetName());

    //Type
    TiXmlElement* typeXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
    typeXmlElement->SetAttribute(_T("value"), jointType);

    //References
    TSTR referenceName = _T("solid0");
    TSTR nextReferenceName = _T("solid1");
    
    TiXmlElement* referencesXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
    
    //Parent reference
    int hasParentBody = 0;
    INode* parentBodyNode = NULL;
    IGameProperty* hasParentBodyGameProperty = propertyContainer->QueryProperty(_T("hasParentBody"));
    hasParentBodyGameProperty->GetPropertyValue(hasParentBody, DefaultTimeValue);
    if (hasParentBody)
    {
        IGameProperty* parentBodyGameProperty = propertyContainer->QueryProperty(_T("parentBody"));
        parentBodyNode = Utilities::GetNodeValue(parentBodyGameProperty);
        if (parentBodyNode)
        {
            referencesXmlElement->SetAttribute(referenceName, parentBodyNode->GetName());
            referenceName = nextReferenceName;
        }
    }

    //Child reference
    IGameProperty* childBodyGameProperty = propertyContainer->QueryProperty(_T("childBody"));
    INode* childBodyNode = Utilities::GetNodeValue(childBodyGameProperty);
    if (childBodyNode && childBodyNode != parentBodyNode)
        referencesXmlElement->SetAttribute(referenceName, childBodyNode->GetName());
    
    //Is enabled?
    TSTR enabled;
    if (!properties.GetValue(PropertyKeys::ENABLE, enabled))
        enabled = _T("true");
    TiXmlElement* enabledXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Enabled")));
    enabledXmlElement->SetAttribute(_T("value"), enabled);

    //Break mode
    int isBreakable = 0;
    IGameProperty* isBreakableGameProperty = propertyContainer->QueryProperty(_T("isBreakable"));
    isBreakableGameProperty->GetPropertyValue(isBreakable, DefaultTimeValue);
    if (isBreakable)
    {
        //Break mode
        TSTR breakMode;
        if (!properties.GetValue(PropertyKeys::BREAK_MODE, breakMode))
            breakMode = _T("threshold");
        TiXmlElement* breakModeXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("BreakMode")));        
        breakModeXmlElement->SetAttribute(_T("value"), breakMode);

        //Accumulated strength
        float linearBreakingStrength;
        IGameProperty* linearBreakingStrengthGameProperty = propertyContainer->QueryProperty(_T("linearBreakingStrength"));
        linearBreakingStrengthGameProperty->GetPropertyValue(linearBreakingStrength, DefaultTimeValue);
        TiXmlElement* accumThresholdXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("AccumThresh")));        
        accumThresholdXmlElement->SetAttribute(_T("value"), Utilities::ToString(linearBreakingStrength));
        
        //Threshold
        float breakingThreshold;
        IGameProperty* breakingThresholdGameProperty = propertyContainer->QueryProperty(_T("breakingThreshold"));
        breakingThresholdGameProperty->GetPropertyValue(breakingThreshold, DefaultTimeValue);
        TiXmlElement* breakThresholdXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("BreakThresh")));        
        breakThresholdXmlElement->SetAttribute(_T("value"), Utilities::ToString(breakingThreshold));
    }

    if (!Utilities::IsSameNoCase(jointType, _T("fixed")))
    {
        //Not a fixed joint. Export other parameters

        //Anchor point
        GMatrix transform = GetNodeTM(gameNode, false);
        Point3 position = transform.Translation() * this->exportSettings.scale;
        Utilities::AddXYZElement(jointXmlElement, _T("Anchor"), position);
        
        //Contacts enabled?
        TSTR contactsEnabled;
        if (!properties.GetValue(PropertyKeys::CONTACTS_ENABLED, contactsEnabled))
            contactsEnabled = _T("false");
        TiXmlElement* contactsEnabledXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("ContactsEnabled")));
        contactsEnabledXmlElement->SetAttribute(_T("value"), contactsEnabled);

        //Axes
        if (Utilities::IsSameNoCase(jointType, _T("ball")))
        {
            //Ball joint
            Utilities::AddOpalAxisElement(jointXmlElement, 0, transform, propertyContainer, properties, DefaultTimeValue);
            Utilities::AddOpalAxisElement(jointXmlElement, 1, transform, propertyContainer, properties, DefaultTimeValue);
        }
        else
        {
            //Hinge joint
            Utilities::AddOpalAxisElement(jointXmlElement, 0, transform, propertyContainer, properties, DefaultTimeValue);
        }
    }
}

void OgreMaxSceneExporter::ExportPointPoint(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (this->exportSettings.physicsExportType == NO_PHYSICS)
        return;

    IPropertyContainer* propertyContainer = gameObject->GetIPropertyContainer();
    
    //Motor element
    TiXmlElement* jointXmlElement = 
        (TiXmlElement*)this->physicsExporter.GetRootNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("Motor")));

    //Name
    TiXmlElement* nameXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
    nameXmlElement->SetAttribute(_T("value"), gameNode->GetName());

    //Type
    TiXmlElement* typeXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
    typeXmlElement->SetAttribute(_T("value"), _T("attractor"));

    //References
    TSTR referenceName = _T("solid0");
    TSTR nextReferenceName = _T("solid1");
    
    TiXmlElement* referencesXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
    
    //Parent reference
    int hasParentBody = 0;
    INode* parentBodyNode = NULL;
    IGameProperty* hasParentBodyGameProperty = propertyContainer->QueryProperty(_T("hasParent"));
    hasParentBodyGameProperty->GetPropertyValue(hasParentBody, DefaultTimeValue);
    if (hasParentBody)
    {
        IGameProperty* parentBodyGameProperty = propertyContainer->QueryProperty(_T("parentBody"));
        parentBodyNode = Utilities::GetNodeValue(parentBodyGameProperty);
        if (parentBodyNode)
        {
            referencesXmlElement->SetAttribute(referenceName, parentBodyNode->GetName());
            referenceName = nextReferenceName;
        }
    }

    //Child reference
    IGameProperty* childBodyGameProperty = propertyContainer->QueryProperty(_T("childBody"));
    INode* childBodyNode = Utilities::GetNodeValue(childBodyGameProperty);
    if (childBodyNode && childBodyNode != parentBodyNode)
        referencesXmlElement->SetAttribute(referenceName, childBodyNode->GetName());
    
    //Is enabled?
    TSTR enabled;
    if (!properties.GetValue(PropertyKeys::ENABLE, enabled))
        enabled = _T("true");
    TiXmlElement* enabledXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Enabled")));
    enabledXmlElement->SetAttribute(_T("value"), enabled);

    //Strength
    float strength;
    IGameProperty* strengthGameProperty = propertyContainer->QueryProperty(_T("strength"));
    strengthGameProperty->GetPropertyValue(strength, DefaultTimeValue);
    TiXmlElement* strengthXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Strength")));        
    strengthXmlElement->SetAttribute(_T("value"), Utilities::ToString(strength));
    
    //Exponent
    float exponent;
    IGameProperty* exponentGameProperty = propertyContainer->QueryProperty(_T("tau"));
    exponentGameProperty->GetPropertyValue(exponent, DefaultTimeValue);
    TiXmlElement* exponentXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Exponent")));        
    exponentXmlElement->SetAttribute(_T("value"), Utilities::ToString(exponent));
}

void OgreMaxSceneExporter::ExportSpring(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties)
{
    if (this->exportSettings.physicsExportType == NO_PHYSICS)
        return;

    //Motor element
    TiXmlElement* motorXmlElement = 
        (TiXmlElement*)this->physicsExporter.GetRootNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("Motor")));

    //Name
    TiXmlElement* nameXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
    nameXmlElement->SetAttribute(_T("value"), gameNode->GetName());

    //Type
    TiXmlElement* typeXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
    typeXmlElement->SetAttribute(_T("value"), _T("spring"));

    //References
    TSTR referenceBody;
    if (properties.GetValue(PropertyKeys::BODY, referenceBody))
    {
        TiXmlElement* referencesXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
        referencesXmlElement->SetAttribute(_T("solid"), referenceBody);
    }

    //Is enabled?
    TSTR enabled;
    if (!properties.GetValue(PropertyKeys::ENABLE, enabled))
        enabled = _T("true");
    TiXmlElement* enabledXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("Enabled")));
    enabledXmlElement->SetAttribute(_T("value"), enabled);

    //Mode
    TSTR mode;
    if (!properties.GetValue(PropertyKeys::MODE, mode))
        mode = _T("linear");
    TiXmlElement* modeXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("Mode")));
    modeXmlElement->SetAttribute(_T("value"), mode);

    bool isLinear = Utilities::IsSameNoCase(mode, _T("linear")) || Utilities::IsSameNoCase(mode, _T("linearAndAngular"));
    bool isAngular = Utilities::IsSameNoCase(mode, _T("angular")) || Utilities::IsSameNoCase(mode, _T("linearAndAngular"));

    //Desired position
    GMatrix transform = GetNodeTM(gameNode, false);
    if (isLinear)
    {
        Point3 position = transform.Translation() * this->exportSettings.scale;
        Utilities::AddXYZElement(motorXmlElement, _T("DesiredPos"), position);

        TSTR linearKd;
        if (properties.GetValue(PropertyKeys::LINEAR_KD, linearKd))
        {
            TiXmlElement* linearKdXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("LinearKd")));
            linearKdXmlElement->SetAttribute(_T("value"), linearKd);
        }

        TSTR linearKs;
        if (properties.GetValue(PropertyKeys::LINEAR_KS, linearKs))
        {
            TiXmlElement* linearKsXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("LinearKs")));
            linearKsXmlElement->SetAttribute(_T("value"), linearKs);
        }
    }

    //Desired orientation
    if (isAngular)
    {
        Utilities::AddXYZElement(motorXmlElement, _T("DesiredRight"), transform.GetRow(0));
        Utilities::AddXYZElement(motorXmlElement, _T("DesiredUp"), transform.GetRow(1));
        Utilities::AddXYZElement(motorXmlElement, _T("DesiredForward"), transform.GetRow(2));

        TSTR angularKd;
        if (properties.GetValue(PropertyKeys::ANGULAR_KD, angularKd))
        {
            TiXmlElement* angularKdXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("angularKd")));
            angularKdXmlElement->SetAttribute(_T("value"), angularKd);
        }

        TSTR angularKs;
        if (properties.GetValue(PropertyKeys::ANGULAR_KS, angularKs))
        {
            TiXmlElement* angularKsXmlElement = (TiXmlElement*)motorXmlElement->LinkEndChild(new TiXmlElement(_T("angularKs")));
            angularKsXmlElement->SetAttribute(_T("value"), angularKs);
        }
    }
}

void OgreMaxSceneExporter::ExportGravity(IGameNode* gameNode, const KeyedValues& properties)
{
    INode* targetMaxNode = gameNode->GetMaxNode()->GetTarget();

    if (this->exportSettings.physicsExportType == NO_PHYSICS || targetMaxNode == 0)
        return;

    IGameNode* targetGameNode = this->gameInterface->GetIGameNode(targetMaxNode);

    GMatrix nodeTransform = GetNodeTM(gameNode, false);
    GMatrix targetTransform = GetNodeTM(targetGameNode, false);
    Point3 gravity = targetTransform.Translation() - nodeTransform.Translation();
    gravity *= this->exportSettings.scale;
    
    if (this->exportSettings.physicsExportType == OPAL_PHYSICS)
    {
        TiXmlElement* simulatorInfoXmlElement = this->physicsExporter.GetOpalSimulatorInfoElement();
        Utilities::AddXYZElement(simulatorInfoXmlElement, _T("Gravity"), gravity);
    }
}

void OgreMaxSceneExporter::ExportRaySensor(IGameNode* gameNode, const KeyedValues& properties)
{
    INode* targetMaxNode = gameNode->GetMaxNode()->GetTarget();

    if (this->exportSettings.physicsExportType == NO_PHYSICS || targetMaxNode == 0)
        return;

    //Name
    TSTR name;
    if (!properties.GetValue(PropertyKeys::NAME, name))
        name = gameNode->GetName();

    //Solid reference
    TSTR solidName;
    properties.GetValue(PropertyKeys::SOLID_REFERENCE, solidName);

    //Target
    IGameNode* targetGameNode = this->gameInterface->GetIGameNode(targetMaxNode);

    //Ray position
    GMatrix nodeTransform = GetNodeTM(gameNode, false);
    Point3 position = nodeTransform.Translation();

    //Direction (with distance)
    GMatrix targetTransform = GetNodeTM(targetGameNode, false);
    Point3 offset = targetTransform.Translation() - position;
    offset *= this->exportSettings.scale;    
    
    if (this->exportSettings.physicsExportType == OPAL_PHYSICS)
    {
        //Ray sensor element
        TiXmlElement* physicsSensorXmlElement = 
            (TiXmlElement*)this->physicsExporter.GetRootNodeXmlElement()->LinkEndChild(new TiXmlElement(_T("Sensor")));

        //Name
        TiXmlElement* sensorNameXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Name")));
        sensorNameXmlElement->SetAttribute(_T("value"), name);

        //Type
        TiXmlElement* sensorTypeXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("Type")));
        sensorTypeXmlElement->SetAttribute(_T("value"), _T("raycast"));

        //References
        if (!solidName.isNull())
        {
            TiXmlElement* sensorReferencesXmlElement = (TiXmlElement*)physicsSensorXmlElement->LinkEndChild(new TiXmlElement(_T("References")));
            sensorReferencesXmlElement->SetAttribute(_T("solid"), solidName);
        }

        //Ray
        Utilities::AddRayElement(physicsSensorXmlElement, position, offset);

        //Contact group
        TSTR contactGroup;
        if (Utilities::GetObjectContactGroupInteger(contactGroup, properties, this->sceneProperties))
            Utilities::AddOpalContactGroupElement(physicsSensorXmlElement, contactGroup); 
    }
}

Point3 OgreMaxSceneExporter::GetAxisVector(int axis)
{
    switch (axis)
    {
        case 0: return this->rightVector;
        case 1: return this->forwardVector;
        case 2: return this->upVector;
        default : return Point3(0, 0, 0);
    }
}

IGameNode* OgreMaxSceneExporter::GetControllerLookAtNode(IGameNode* gameNode)
{
    IGameNode* lookAtNode = 0;

    Control* transformController = gameNode->GetMaxNode()->GetTMController();
    Control* rotationController = transformController->GetRotationController();
    if (rotationController)
    {
        if (rotationController->ClassID() == Class_ID(LOOKAT_CONSTRAINT_CLASS_ID,0))
        {
            ILookAtConstRotation* lookatControl = (ILookAtConstRotation*)rotationController;            
            if (lookatControl->GetNumTargets() > 0)
            {                
                INode* maxNode = lookatControl->GetNode(0);
                lookAtNode = this->gameInterface->GetIGameNode(maxNode);
            }            
        }
    }

    return lookAtNode;
}

GMatrix OgreMaxSceneExporter::GetNodeTM(IGameNode* gameNode, bool relativeToParent)
{
    GMatrix nodeTransform = Utilities::GetNodeTM(gameNode, DefaultTimeValue, relativeToParent);    
    return nodeTransform;
}

Quat OgreMaxSceneExporter::GetSkyRotation(IGameNode* gameNode)
{
    GMatrix transform = GetNodeTM(gameNode, false);
    Quat rotation = transform.Rotation();
    if (this->exportSettings.upAxis == UP_AXIS_Z)    
    {
        AngAxis rotationAngAxis(1, 0, 0, DegToRad(-90));
        rotation = rotation * Quat(rotationAngAxis);
    }

    return rotation;
}

Box3 OgreMaxSceneExporter::GetBoundingBox(IGameMesh* gameMesh)
{
    Matrix3 boxMatrix(TRUE);
    if (this->exportSettings.upAxis == UP_AXIS_Y)
        boxMatrix = RotateXMatrix(DegToRad(90));
    return gameMesh->GetMaxMesh()->getBoundingBox(&boxMatrix);                
}

void OgreMaxSceneExporter::ExportLookAtDirection
    (
    IGameNode* gameNode, 
    const TSTR& elementName, 
    TiXmlElement* parentXmlElement
    )
{       
    Control* transformController = gameNode->GetMaxNode()->GetTMController();
    Control* rotationController = transformController->GetRotationController();
    if (rotationController)
    {
        Point3 localDirection;
        bool foundLocalDirection = false;

        if (rotationController->ClassID() == Class_ID(LOOKAT_CONTROL_CLASS_ID, 0))
        {
            //Old style look at controller
            
            ILookatControl* lookatControl = (ILookatControl*)rotationController;
            localDirection = GetAxisVector(lookatControl->GetAxis());
            if (lookatControl->GetFlip())
                localDirection = -localDirection;
            
            foundLocalDirection = true;
        }
        else if (rotationController->ClassID() == Class_ID(LOOKAT_CONSTRAINT_CLASS_ID, 0))
        {
            //The more recent look at controller
            
            ILookAtConstRotation* lookatControl = (ILookAtConstRotation*)rotationController;            
            localDirection = GetAxisVector(lookatControl->GetTargetAxis());
            if (lookatControl->GetTargetAxisFlip())
                localDirection = -localDirection;
            
            foundLocalDirection = true;
        }

        if (foundLocalDirection)
            Utilities::AddXYZElement(parentXmlElement, elementName.data(), localDirection);
    }
}

void OgreMaxSceneExporter::ExportMaterialAttribute(TiXmlElement* element, IGameMaterial* gameMaterial, const TSTR& attributeName)
{
    //Set the attribute in the XML element
    TSTR materialName = Utilities::MakeMaterialName(gameMaterial, this->exportSettings.noMaterialName);
    element->SetAttribute(attributeName.data(), materialName);

    //Add material to export table if it hasn't been explicitly disabled from export 
    //and hasn't already been added to the table
    if (Utilities::CanExportMaterial(materialName, this->sceneProperties) && 
        this->materials.find(materialName.data()) == this->materials.end())
    {
        this->materials[materialName.data()] = gameMaterial;
    }
}

bool OgreMaxSceneExporter::IgnoreNode(INode* maxNode, const TSTR& nodeExportType)
{
    //Ignore nodes with an "ignore" type, and nodes in the ignore list
    if (Utilities::IsSameNoCase(nodeExportType, _T("ignore")) ||
        std::find(this->ignoreNodes.begin(), this->ignoreNodes.end(), maxNode) != this->ignoreNodes.end())
    {
        return true;
    }

    //Skip particle view objects
    Class_ID objectClassID = maxNode->GetObjectRef()->ClassID();
    if (objectClassID == ParticleView_Class_ID ||
        objectClassID == SKELOBJ_CLASS_ID ||
        objectClassID == FOOTPRINT_CLASS_ID ||
        objectClassID == BIPED_CLASS_ID)
    {
        return true;
    }

    //Skip objects with physique-related controllers
    Control* c = maxNode->GetTMController();
    if ((c->ClassID() == BIPSLAVE_CONTROL_CLASS_ID) ||
        (c->ClassID() == BIPBODY_CONTROL_CLASS_ID) ||
        (c->ClassID() == FOOTPRINT_CLASS_ID))
    {
        return true;
    }
    
    return false;
}

Point3 Utilities::ParseXYZMask(const TSTR& text, float foundValue, float notFoundValue)
{
    Point3 mask;

    if (FindCharNoCase(text, _T('x')) != -1)
        mask.x = foundValue;
    else
        mask.x = notFoundValue;

    if (FindCharNoCase(text, _T('y')) != -1)
        mask.y = foundValue;
    else
        mask.y = notFoundValue;

    if (FindCharNoCase(text, _T('z')) != -1)
        mask.z = foundValue;
    else
        mask.z = notFoundValue;

    return mask;
}