#ifndef OgreMax_OgreMaxSceneExporter_INCLUDED
#define OgreMax_OgreMaxSceneExporter_INCLUDED


//Includes---------------------------------------------------------------------
#include "KeyedValues.h"
#include "ExportSettings.h"
#include "MaterialExporter.h"
#include "MeshExporter.h"
#include "PhysicsExporter.h"
#include "ModelExporter.h"


//Classes----------------------------------------------------------------------
namespace OgreMax
{
    class OgreMaxSceneExporter : public SceneExport
    {
    public:
	    OgreMaxSceneExporter();
	    ~OgreMaxSceneExporter();

	    /** Number of extensions supported  */
	    int ExtCount();

        /** Gets extension number 'n' */
	    const TCHAR* Ext(int n);

        /** Long ASCII description (i.e. "Ascii Export"). */
	    const TCHAR* LongDesc();

        /** Short ASCII description (i.e. "Ascii"). */
	    const TCHAR* ShortDesc();

        /** ASCII Author name. */
	    const TCHAR* AuthorName();

        /** ASCII Copyright message  */
	    const TCHAR* CopyrightMessage();

        /** Other message #1.*/
	    const TCHAR* OtherMessage1(); 

        /** Other message #2 */
	    const TCHAR* OtherMessage2(); 

        /** Version number * 100 (i.e. v3.01 = 301) */
	    unsigned int Version();

        /** Show DLL's "About..." box. */
	    void ShowAbout(HWND hWnd);

	    int DoExport
            (
            const TCHAR* fileName, 
            ExpInterface* exportInterface, 
            Interface* maxInterface, 
            BOOL suppressPrompts = FALSE, 
            DWORD options = 0
            );
	    
        BOOL SupportsOptions(int ext, DWORD options);

        static ClassDesc* GetClassDesc();

    private:
        
        /** Shows the specified error message in a message box. */
        void ShowError(const TSTR& errorMessage);

        TiXmlElement* GetCurrentNodeXmlElement();

        bool Initialize();
        void Shutdown();

        void InitializeSceneProperties();
        void InitializeExportSettings();
        bool InitializeGameInterface();
        void InitializeMeshAndMaterialExporters();

        void ExportScene();
        void ExportMeshesAndMaterials();
        void ExportPhysics();
        void ExportModels();
        void ExportSceneProperties();
        
        void ExportEnvironment();
        void ExportEnvironmentRange();
        void ExportEnvironmentFog();
        void ExportEnvironmentShadows(const TSTR& shadowTechnique);

        void ExportNodes();

        bool GetInitialNodeStates(INode* gameNode);

        /**
         * Traverses the node tree, calling ActOnNode() as necessary
         */
        void TraverseNode(IGameNode* gameNode);

        /**
         * Performs an export action on the node
         * @param gameNode [in] - The node to act on
         * @param properties [in] - Parsed Key/value properties to be used in exporting node
         * @param ignoreChildren [out] - Indicates whether the caller should ignore the node's children
         * @return The XML element that contains the exported node data
         */
        TiXmlElement* ActOnNode(IGameNode* gameNode, const KeyedValues& properties, bool& ignoreChildren);

        /** Flags that can be passed to ExportNode() */
        enum
        {
            NO_NODE_TRANSFORM = 0x1,
            NO_NODE_TARGETS = 0x2
        };

        /**
         * Exports a 3DS Max node that ends up as an Ogre::Node
         * @param gameNode [in] - The 3DS Max node to node to export
         * @param properties [in] - Parsed Key/value properties to be used in exporting node
         * @param nodeExportType [in] - The predetermined export type for the node object. This may be empty,
         * in which case the type is determined by the 3DS Max node object's type
         * @param ignoreChildren [out] - Indicates whether the caller should ignore the node's children
         * @param nodeXmlElement [in/out] - The XML element that the node and its children are exported into. If
         * null, an element will be created. If not null, this same element will be returned
         * @param physicsSceneXmlElement [in/out] - The XML element that will contain the exported 
         * physics objects. If null, the default physics XML document will be used
         * @param bindingsSceneXmlElement [in/out] - The XML element that will contain the exported 
         * physics/graphics bindings. If null, the default bindings XML document will be used
         * @return The XML element that contains the exported node data
         */
        TiXmlElement* ExportNode
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& nodeExportType, 
            bool& ignoreChildren,
            TiXmlElement* nodeXmlElement = 0,
            TiXmlElement* physicsSceneXmlElement = 0,
            TiXmlElement* bindingsSceneXmlElement = 0,
            size_t flags = 0
            );
        
        TiXmlElement* ExportModel
            (
            IGameNode* gameNode, 
            IGameObject* gameObject, 
            const KeyedValues& properties, 
            const TSTR& nodeExportType, 
            bool& ignoreChildren
            );
        
        void ExportParticleSystem
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement
            );
        void ExportBillboardSet
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement
            );
        void ExportBillboard
            (
            IGameNode* gameNode, 
            IGameNode* parentGameNode,
            TiXmlElement* parentXmlElement,
            const TSTR& rotationTypeLower
            );
        void ExportPlane
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement,
            TiXmlElement* physicsXmlElement
            );
        void ExportEntity
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement,
            TiXmlElement* physicsXmlElement,
            TiXmlElement* bindingsXmlElement,
            Point3 graphicsToPhysicsOffsetTranslation
            );
        void ExportLight
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement
            );
        void ExportCamera
            (
            IGameNode* gameNode, 
            IGameObject* gameObject,
            const KeyedValues& properties, 
            const TSTR& name, 
            const TSTR& id, 
            TiXmlElement* nodeXmlElement
            );
        
        
        void ExportAnimations(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties, TiXmlElement* nodeXmlElement);
        void ExportNoteTracks(IGameNode* gameNode, const KeyedValues& properties, TiXmlElement* nodeXmlElement);

        void ExportExternal(IGameNode* gameNode, const KeyedValues& properties);
        void ExportTerrain(IGameNode* gameNode, const KeyedValues& properties);
        void ExportSkyBox(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);
        void ExportSkyDome(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);
        void ExportSkyPlane(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);

        TiXmlElement* ExportModelInstance(IGameNode* gameNode, const KeyedValues& properties);
        
        void ExportHinge(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);
        void ExportPointPoint(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);
        void ExportSpring(IGameNode* gameNode, IGameObject* gameObject, const KeyedValues& properties);
        void ExportGravity(IGameNode* gameNode, const KeyedValues& properties);
        void ExportRaySensor(IGameNode* gameNode, const KeyedValues& properties);

        Point3 GetAxisVector(int axis);
        IGameNode* GetControllerLookAtNode(IGameNode* gameNode);

        GMatrix GetNodeTM(IGameNode* gameNode, bool relativeToParent);
        Quat GetSkyRotation(IGameNode* gameNode);
        Box3 GetBoundingBox(IGameMesh* gameMesh);
        
        void ExportLookAtDirection
            (
            IGameNode* gameNode, 
            const TSTR& elementName, 
            TiXmlElement* parentXmlElement
            );

        void ExportMaterialAttribute(TiXmlElement* element, IGameMaterial* gameMaterial, const TSTR& attributeName = _T("material"));

        bool IgnoreNode(INode* maxNode, const TSTR& nodeExportType);
        
    private:
        ExportSettings exportSettings;
        Point3 rightVector;
        Point3 forwardVector;
        Point3 upVector;
        Point3 upVectorPerpendicular; //Used to orient exported plane entities

        InitialNodeStates initialNodeStates;
        std::map<TSTR, InitialNodeState*> initialNodeStatesByName;
        
        IGameScene* gameInterface;

        TSTR exportFileName;
        TSTR exportPath;
	    Interface* maxInterface;
        bool suppressPrompts;
        DWORD options;

        TSTR sceneSummary;
        KeyedValues sceneProperties;
        
        TiXmlDocument sceneXmlDocument;
	    TiXmlElement* sceneXmlElement;        
            TiXmlElement* nodesXmlElement;
            TiXmlElement* externalsXmlElement;
            TiXmlElement* environmentXmlElement;
                TiXmlElement* skyBoxXmlElement;
                TiXmlElement* skyDomeXmlElement;
                TiXmlElement* skyPlaneXmlElement;
            TiXmlElement* terrainXmlElement;
        
        std::vector<TiXmlElement*> nodeXmlElementStack;

        TiXmlDocument sceneBindingsXmlDocument;        
	    TiXmlElement* sceneBindingsXmlElement;        

        std::list<GameNodeAndObject> entityMeshNodes;

        std::list<ModelExporter*> modelExporters;

        std::list<INode*> ignoreNodes;

        MaterialMap materials;
        
        PhysicsExporter physicsExporter;

    public:
        //Default time value and frame number at which objects are evaluated
        //These two values should coincide
        static const TimeValue DefaultTimeValue;
        static const int DefaultFrameNumber;
    };
}


#endif
