
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "TabControl.h"
#include "PluginGlobals.h"

using namespace OgreMax;


//Implementation---------------------------------------------------------------
TabControl::TabControl()
{    
    this->lastSelectedPage = -1;
}

TabControl::~TabControl()
{
    Destroy();
}

void TabControl::Create(HWND hTab, bool stretch)
{
    Create(hTab, NULL, NULL, NULL, 0, stretch);
}

void TabControl::Create
    (
    HWND hTab, 
    const int* tabNameStringIDs, 
    const int* tabDlgResourceIDs, 
    const bool* tabVisibleStates,
    int tabPageCount,
    bool stretch
    )
{
    if (this->hTab != NULL)
        Destroy();

    this->hTab = hTab;    
    this->stretchPages = stretch;
        
    //Link struct 'this' pointer to hTab
	SetWindowLongPtr(this->hTab, GWLP_USERDATA, (LONG_PTR)this);

	if (tabPageCount > 0)
    {
        //Create array of pages
	    this->tabPages.resize(tabPageCount);

	    //Add a tab for each name
        TCITEM tabItem;
	    tabItem.mask = TCIF_TEXT;	
	    for (size_t tabPageIndex = 0; tabPageIndex < this->tabPages.size(); tabPageIndex++)
	    {
            //Add tab
            if (tabVisibleStates == NULL || tabVisibleStates[tabPageIndex])
            {
                tabItem.pszText = PluginGlobals::GetString(tabNameStringIDs[tabPageIndex]);
		        TabCtrl_InsertItem(this->hTab, tabPageIndex, &tabItem);
            }
            else
                this->tabPages[tabPageIndex].hidden = true;

		    //Add page
            this->tabPages[tabPageIndex].labelStringID = tabNameStringIDs[tabPageIndex];
            this->tabPages[tabPageIndex].dialogResourceID = tabDlgResourceIDs[tabPageIndex];        
            this->tabPages[tabPageIndex].hWnd = CreateDialogParam
                (
                PluginGlobals::GetInstanceHandle(),
                MAKEINTRESOURCE(tabDlgResourceIDs[tabPageIndex]),
			    GetParent(this->hTab),
			    (DLGPROC)TabPage_DlgProc,
                (LPARAM)this
                );
            
		    //Set initial page position
		    if (this->stretchPages)
			    StretchPage((int)tabPageIndex);
		    else
			    RepositionPage((int)tabPageIndex);
	    }

        //Show first visible page
        int firstPage = GetNearestVisiblePage(0, true);
        if (firstPage != -1)
	        SelectPage(firstPage);
    }
}

void TabControl::Destroy()
{
    if (!this->tabPages.empty())
    {
        //Destroy all tab page windows
        for (size_t tabPageIndex = 0; tabPageIndex < this->tabPages.size(); tabPageIndex++)
		    DestroyWindow(this->tabPages[tabPageIndex].hWnd);
        this->tabPages.clear();

        //Delete all tabs
        TabCtrl_DeleteAllItems(this->hTab);

	    this->lastSelectedPage = -1;
    }
}

int TabControl::GetTabCount()
{
    return TabCtrl_GetItemCount(this->hTab);
}

int TabControl::GetPageCount()
{
    return (int)this->tabPages.size();
}

HWND TabControl::GetPageHWND(int page)
{
    return this->tabPages[page].hWnd;
}

void TabControl::SelectPage(int page)
{
    if (!this->tabPages[page].hidden)
    {
        //Hide current page
        int selectedPage = GetSelectedPage();
        if (selectedPage != -1)
            ShowWindow(this->tabPages[selectedPage].hWnd, SW_HIDE);

        //Select new tab
        TabCtrl_SetCurSel(this->hTab, PageToTab(page));
	    NMHDR nmh = { GetParent(this->hTab), GetDlgCtrlID(this->hTab), TCN_SELCHANGE };
	    SendMessage(nmh.hwndFrom, WM_NOTIFY, (WPARAM)nmh.idFrom, (LPARAM)&nmh);

        //Show new page
        ShowWindow(this->tabPages[page].hWnd, SW_SHOW);    
    }
}

HWND TabControl::AddPage(int labelStringID, int dialogResourceID, int insertAt, bool visible)
{
    if (insertAt == -1)
        insertAt = (int)this->tabPages.size();

    TabPage newTab;

    //Create tab
    if (visible)
    {
        TCITEM tabItem;
	    tabItem.mask = TCIF_TEXT;	
	    tabItem.pszText = PluginGlobals::GetString(labelStringID);
	    TabCtrl_InsertItem(this->hTab, insertAt, &tabItem);
    }
    else
        newTab.hidden = true;

	//Create page    
    newTab.labelStringID = labelStringID;
    newTab.dialogResourceID = dialogResourceID;
    newTab.hWnd = CreateDialogParam
        (
        PluginGlobals::GetInstanceHandle(),
        MAKEINTRESOURCE(dialogResourceID),
		GetParent(this->hTab),
		(DLGPROC)TabPage_DlgProc,
        (LPARAM)this
        );
    
    //Add the page to the collection of pages
    this->tabPages.push_back(newTab); //Make room for the new page
    for (int i = (int)this->tabPages.size() - 1; i > insertAt; i--)
        this->tabPages[i] = this->tabPages[i - 1];
    this->tabPages[insertAt] = newTab;
        
    //Set initial page position
    if (this->stretchPages)
	    StretchPage(insertAt);
    else
	    RepositionPage(insertAt);

    //Show page if we just added the first visible tab
    if (visible && GetTabCount() == 1)
        ShowWindow(this->tabPages[insertAt].hWnd, SW_SHOW);
    
    return this->tabPages[insertAt].hWnd;
}

void TabControl::RemovePage(int page)
{
    //Destroy page
    DestroyWindow(this->tabPages[page].hWnd);

    //Destroy tab
    TabCtrl_DeleteItem(this->hTab, page);

    //Remove page from collection of pages
    for (int i = page; i < (int)this->tabPages.size() - 1; i++)
        this->tabPages[i] = this->tabPages[i + 1];
    this->tabPages.pop_back();
}

void TabControl::ShowPageTab(int page)
{
    if (this->tabPages[page].hidden)
    {
        //Create tab
        TCITEM tabItem;
	    tabItem.mask = TCIF_TEXT;	
	    tabItem.pszText = PluginGlobals::GetString(this->tabPages[page].labelStringID);
	    TabCtrl_InsertItem(this->hTab, page, &tabItem);

        this->tabPages[page].hidden = false;
    }    
}

void TabControl::HidePageTab(int page)
{
    if (!this->tabPages[page].hidden)
    {        
        //Determine if the page being hidden is the current one
        bool hidingCurrentPage = page == GetSelectedPage();
        
        //Destroy tab
        TabCtrl_DeleteItem(this->hTab, PageToTab(page));
        this->tabPages[page].hidden = true;

        //If the current page is being hidden...
        if (hidingCurrentPage)
        {
            //Hide the current page
            ShowWindow(this->tabPages[page].hWnd, SW_HIDE);

            //Show the nearest visible page
            int newPage = GetNearestVisiblePage(page);
            if (newPage != -1)
                SelectPage(newPage);
        }
    }
}

void TabControl::ShowPageTabRange(int startPage, int endPage)
{
    if (endPage == -1)
        endPage = (int)this->tabPages.size() - 1;

    for (int page = startPage; page <= endPage; page++)
        ShowPageTab(page);
}

void TabControl::HidePageTabRange(int startPage, int endPage)
{
    if (endPage == -1)
        endPage = (int)this->tabPages.size() - 1;

    for (int page = startPage; page <= endPage; page++)
        HidePageTab(page);
}

void TabControl::TogglePageTab(int page)
{
    if (this->tabPages[page].hidden)
        ShowPageTab(page);
    else
        HidePageTab(page);
}

bool TabControl::IsPageVisible(int page)
{
    return !this->tabPages[page].hidden;
}

bool TabControl::RepositionPage(int page)
{
    RECT tabRect;
    GetClientRect(this->hTab, tabRect);

    //Move the tab
    RECT pageRect;
	GetClientRect(this->tabPages[page].hWnd, pageRect);
	LONG width = pageRect.right - pageRect.left;
	LONG height = pageRect.bottom - pageRect.top;
	pageRect.left = tabRect.left;
	pageRect.top = tabRect.top;
    pageRect.right = pageRect.left + width;
	pageRect.bottom = pageRect.top + height;
    
    //Move the child and put it on top
	return SetWindowPos
        (
        this->tabPages[page].hWnd, 
        HWND_TOP,
		pageRect.left, pageRect.top, pageRect.right, pageRect.bottom,
		0) == TRUE;
}

bool TabControl::StretchPage(int page)
{
	RECT rect;
	GetClientRect(this->hTab, rect);

	//Move the child and put it on top
	return SetWindowPos
        (
        this->tabPages[page].hWnd, 
        HWND_TOP,
		rect.left, rect.top, rect.right, rect.bottom,
		0
        ) == TRUE;
}

int TabControl::GetSelectedPage()
{
    return TabToPage(TabCtrl_GetCurSel(this->hTab));
}

BOOL TabControl::Notify(NMHDR* pnm)
{
	switch (pnm->code)
	{
        case TCN_SELCHANGING: return OnSelectionChanging();
		case TCN_SELCHANGE: return OnSelectionChanged();
	}
	return FALSE;
}

void TabControl::GetClientRect(HWND hwnd, RECT& rect)
{
    RECT* prc = &rect;

    RECT rtab_0;
	LONG lStyle = GetWindowLong(hwnd, GWL_STYLE); 

	//Calculate the tab control's display area
	GetWindowRect(hwnd, prc);
	ScreenToClient(GetParent(hwnd), (POINT*)&prc->left);
	ScreenToClient(hwnd, (POINT*)&prc->right);
	TabCtrl_GetItemRect(hwnd,0,&rtab_0); //The tab itself

	if ((lStyle & TCS_BOTTOM) && (lStyle & TCS_VERTICAL)) //Tabs to Right
	{
		prc->top = prc->top + 6; //x coord
		prc->left = prc->left + 4; //y coord
		prc->bottom = prc->bottom - 12; //height
		prc->right = prc->right - (12 + rtab_0.right-rtab_0.left); //width
	}
	else if (lStyle & TCS_VERTICAL) //Tabs to Left
	{
		prc->top = prc->top + 6; //x coord
		prc->left = prc->left + (4 + rtab_0.right-rtab_0.left); //y coord
		prc->bottom = prc->bottom - 12; //height
		prc->right = prc->right - (12 + rtab_0.right-rtab_0.left); //width
	}
	else if (lStyle & TCS_BOTTOM) //Tabs on Bottom
	{
		prc->top = prc->top + 6; //x coord
		prc->left = prc->left + 4; //y coord
		prc->bottom = prc->bottom - (16 + rtab_0.bottom-rtab_0.top); //height
		prc->right = prc->right - 12; //width
	}
	else //Tabs on top
	{
		prc->top = prc->top + (6 + rtab_0.bottom-rtab_0.top); //x coord
		prc->left = prc->left + 4; //y coord
		prc->bottom = prc->bottom - (16 + rtab_0.bottom-rtab_0.top); //height
		prc->right = prc->right - 12; //width
	}
}

int TabControl::TabToPage(int tab)
{
    //Tabs are visible, so iterate over the tab pages until the visible counter
    //is the same as the tab index

    int visibleCount = 0;

    for (int i = 0; i < (int)this->tabPages.size(); i++)
    {
        if (!this->tabPages[i].hidden)
        {
            if (visibleCount == tab)
                return i;
            
            visibleCount++;
        }
    }

    return -1;
}

int TabControl::PageToTab(int page)
{
    //We iterate over all the pages until the iteration counter equals the page index, 
    //then return the number of visible tabs it took to get there

    int visibleCount = 0;

    for (int i = 0; i < (int)this->tabPages.size(); i++)
    {
        if (i == page)
            return visibleCount;

        if (!this->tabPages[i].hidden)
            visibleCount++;
    }

    return -1;
}

int TabControl::GetNearestVisiblePage(int page, bool includeStartPage)
{
    int newPage = -1;

    if (includeStartPage && !this->tabPages[page].hidden)
        return page;

    //Go forward
    for (int i = page + 1; i < (int)this->tabPages.size(); i++)
    {
        if (!this->tabPages[i].hidden)
        {
            newPage = i;
            break;
        }
    }
    
    if (newPage == -1)
    {
        //Go backward
        for (int i = page - 1; i >= 0; i--)
        {
            if (!this->tabPages[i].hidden)
            {
                newPage = i;
                break;
            }
        }
    }   

    return newPage;
}

BOOL TabControl::OnSelectionChanging()
{
    this->lastSelectedPage = GetSelectedPage();
    return TRUE;
}

BOOL TabControl::OnSelectionChanged()
{
    //Hide the current child dialog box, if any.
    if (this->lastSelectedPage != -1)	
	    ShowWindow(this->tabPages[this->lastSelectedPage].hWnd, FALSE);

    //Show the new child dialog box.
    int visiblePageIndex = GetSelectedPage();
    if (visiblePageIndex != -1)	
        ShowWindow(this->tabPages[visiblePageIndex].hWnd, TRUE);

    this->lastSelectedPage = -1;

	return TRUE;
}

BOOL CALLBACK TabControl::TabPage_DlgProc(HWND hTabPageDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    TabControl* tabControl;

    if (message == WM_INITDIALOG)
    {
        tabControl = (TabControl*)lParam;        
        SetWindowLongPtr(hTabPageDlg, GWLP_USERDATA, lParam);
    }
    else
    {
        tabControl = (TabControl*)GetWindowLongPtr(hTabPageDlg, GWLP_USERDATA);
        if (tabControl == NULL)
            return FALSE;
    }

    switch (message)
    {
        case WM_INITDIALOG:
        {
            tabControl->TabPage_OnInitDialog(hTabPageDlg, (HWND)wParam, lParam);
            return 0;
        }
        case WM_SIZE:
        {
            tabControl->TabPage_OnSize(hTabPageDlg, (UINT)wParam, (int)LOWORD(lParam), (int)HIWORD(lParam));
            return 0;
        }
        case WM_CLOSE:
        {
            //Prevents the tab control window from being closed if the user hits 
            //escape in a multi-line text box
            return 0;
        }
    }

    WNDPROC parentWndProc = (WNDPROC)GetWindowLongPtr(tabControl->hTab, GWLP_WNDPROC);    
    return parentWndProc(hTabPageDlg, message, wParam, lParam);
}

void TabControl::TabPage_OnSize(HWND hTabPageDlg, UINT state, int cx, int cy)
{	
}

BOOL TabControl::TabPage_OnInitDialog(HWND hTabPageDlg, HWND hwndFocus, LPARAM lParam)
{
	//We handle this message so that it is not sent to the main dlg proc
	//each time a tab page is initialized.
	return (BOOL)DefWindowProc(hTabPageDlg, WM_INITDIALOG, (WPARAM)hwndFocus, lParam);
}