#ifndef OgreMax_TabControl_INCLUDED
#define OgreMax_TabControl_INCLUDED


//Includes---------------------------------------------------------------------
#include <vector>


//Classes----------------------------------------------------------------------
namespace OgreMax
{    
    /** 
     * Provides a convenient programming interface for a tab control
     * A key feature this class provides is the ability hide tab pages.
     * Regardless of whether a page is visible or hidden, all public methods
     * take an index to the page
     */
    class TabControl 
    {
    public:
        TabControl();
        virtual ~TabControl();

        /** Initializes the tab control with no initial tab pages */
        void Create(HWND hTab, bool stretch = false);

        /** Initializes the tab control with an initial collection of tab pages */
        void Create
            (
            HWND hTab, 
            const int* tabNameStringIDs, 
            const int* tabDlgResourceIDs, 
            const bool* tabVisibleStates,
            int tabPageCount,
            bool stretch = false
            );

        /** Destroys all of the tab pages in the control. The control itself is not destroyed */
        void Destroy();

        /** 
         * Gets the number of visible tabs (not pages, just the tabs)
         * Visible tabs are tabs that have not been hidden
         */
        int GetTabCount();

        /** Gets the total number of pages that have been added */
        int GetPageCount();

        /** Gets the window handle for the specified page */
        HWND GetPageHWND(int page);

        /** Displays the specified page and selects its corresponding tab */
        void SelectPage(int page);

        /** Adds a new page to the tab control */
        HWND AddPage(int labelStringID, int dialogResourceID, int insertAt = -1, bool visible = true);
    
        /** Deletes the specified page and its corresponding tab */
        void RemovePage(int page);

        /** Unhides the specified page and its corresponding tab */
        void ShowPageTab(int page);

        /** Hides the specified page and its corresponding tab */
        void HidePageTab(int page);

        /** Hides the specified range of tab pages */
        void HidePageTabRange(int startPage, int endPage = -1);

        /** Shows the specified range of tab pages */
        void ShowPageTabRange(int startPage, int endPage = -1);

        /** Shows or hides the specified page and its corresponding tab */
        void TogglePageTab(int page);

        /** Determines if the specified page is visible */
        bool IsPageVisible(int page);

        /**
         * Stretch the tab page to fit the tab control's display area.
         * @param page [in] - Index of tab page
         * @return Returns true if successful, false otherwise
         */
	    bool StretchPage(int page);

        /** Gets the currently selected page, or -1 if no page is selected */
        int GetSelectedPage();

        /** To be called by the control owner when a WM_NOTIFY is received for this control */
        BOOL Notify(NMHDR* pnm);

        /** 
         * Gets the index of the page closest to the specified page, or -1 if there are no visible pages 
         * @param page [in] - The start page from which the searching occurs
         * @param includeStartPage - Determines if the start page is considered in the searching
         */
        int GetNearestVisiblePage(int page, bool includeStartPage = false);

    private:
        static void GetClientRect(HWND hwnd, RECT& rect);

        /** Converts a tab index to a page index */
        int TabToPage(int tab);

        /** Converts a page index to a tab index */
        int PageToTab(int page);

        /**
         * Repositions the tab page in the tab control's client area.
         * @param page [in] - Index of tab page
         * @return Returns true if successful, false otherwise
         */
	    bool RepositionPage(int page);

        /** Called just before the current tab selection changes */
        BOOL OnSelectionChanging();

        /** Called just after the current tab selection changes */
        BOOL OnSelectionChanged();

        /**
         * Window procedure for the tab pages.
         * @param hTabPageDlg [in] - The tab page dialog window handle
         * @param message [in] - Message ID
         * @param wparam [in] - message parameter
         * @param wparam [in] - message parameter
         * @return Depends on message
         */
        static BOOL CALLBACK TabPage_DlgProc(HWND hTabPageDlg, UINT message, WPARAM wParam, LPARAM lParam);

        /**
         * Tab page handler for the WM_SIZE message
         * @param hTabPageDlg [in] - The tab page dialog window handle
         * @param state - Resizing flag
         * @param cx - Width
         * @param cy - Height
         */
        void TabPage_OnSize(HWND hTabPageDlg, UINT state, INT cx, INT cy);

        /**
         * Tab page handler for the WM_INITDIALOG message
         * @param hTabPageDlg [in] - The tab page dialog window handle
         * @param hwndFocus - handle of control to receive focus
         * @param lParam - initialization parameter
         */
        BOOL TabPage_OnInitDialog(HWND hTabPageDlg, HWND hwndFocus, LPARAM lParam);

    protected:
        /** The tab control window handle */
        HWND hTab;        

        struct TabPage
        {
            TabPage()
            {
                this->hidden = false;
            }

            int labelStringID;
            int dialogResourceID;
            HWND hWnd;
            bool hidden;
        };
        
        /** All the tab pages */
        std::vector<TabPage> tabPages;	    

        /** Whether or not to stretch new pages */
        bool stretchPages;

        /** Used during OnSelectionChanging()/OnSelectionChanged */
        int lastSelectedPage;
    };
}

#endif