
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "TypeDetectors.h"
#include "PropertyKeys.h"
#include "PropertyValues.h"
#include "Utilities.h"

using namespace OgreMax;


//Implementation---------------------------------------------------------------
ObjectTypeDetector::ObjectTypeDetector()
{
    this->type = DEFAULT;
}

ObjectTypeDetector::ObjectTypeDetector(INode* node, TSTR* typeTextPtr, bool useTypeText)
{
    Detect(node, typeTextPtr, useTypeText);
}

void ObjectTypeDetector::Detect(INode* node, TSTR* typeTextPtr, bool useTypeText)
{
    this->type = DEFAULT;
    this->allowedTypes.clear();

    this->allowedTypes.reserve(10);
    this->allowedTypes.push_back(IGNORE_OBJECT);
    this->allowedTypes.push_back(EMPTY);
    this->allowedTypes.push_back(MODEL_INSTANCE);
    this->allowedTypes.push_back(EXTERNAL);
    this->allowedTypes.push_back(PARTICLE_SYSTEM);
    this->allowedTypes.push_back(BILLBOARD_SET);

    //Get node type text
    if (useTypeText)
    {
        if (typeTextPtr != NULL)
            this->typeText = *typeTextPtr;
        else
        {
            KeyedValues properties;
            Utilities::GetNodeUserProperties(node, properties);
            if (!properties.GetValue(PropertyKeys::TYPE, this->typeText))
                properties.GetValue(PropertyKeys::GRAPHICS_TYPE, this->typeText);
        }
    }

    //Get node object superclass ID
    Object* obj = node->EvalWorldState(0).obj;
    SClass_ID superClassID = obj ? obj->SuperClassID() : 0;

    if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::IGNORE_OBJECT))
        this->type = IGNORE_OBJECT;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::EMPTY) || superClassID == HELPER_CLASS_ID)
        this->type = EMPTY;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::MODEL_INSTANCE))
        this->type = MODEL_INSTANCE;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::ENTITY))
        this->type = ENTITY;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::LIGHT) || superClassID == LIGHT_CLASS_ID)
        this->type = LIGHT;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::CAMERA) || superClassID == CAMERA_CLASS_ID)
        this->type = CAMERA;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::EXTERNAL))
        this->type = EXTERNAL;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::TERRAIN))
        this->type = TERRAIN;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::SKY_BOX))
        this->type = SKY_BOX;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::SKY_DOME))
        this->type = SKY_DOME;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::SKY_PLANE))
        this->type = SKY_PLANE;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::PARTICLE_SYSTEM))
        this->type = PARTICLE_SYSTEM;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::BILLBOARD_SET))
        this->type = BILLBOARD_SET;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::BILLBOARD))
        this->type = BILLBOARD;
    else if (Utilities::IsSameNoCase(this->typeText, PropertyValues::ObjectType::PLANE))
        this->type = PLANE;
    
    switch (superClassID)
    {
        case GEOMOBJECT_CLASS_ID: 
        {
            if (this->type == DEFAULT)
                this->type = ENTITY;

            PushFront(this->allowedTypes, ENTITY);
            this->allowedTypes.push_back(TERRAIN);
            this->allowedTypes.push_back(SKY_BOX);
            this->allowedTypes.push_back(SKY_DOME);
            this->allowedTypes.push_back(SKY_PLANE);
            this->allowedTypes.push_back(BILLBOARD);
            this->allowedTypes.push_back(PLANE);

            break;
        }
    }

    //Add the primary type to the beginning of the allowed types
    if (this->type != DEFAULT && 
        std::find(this->allowedTypes.begin(), this->allowedTypes.end(), this->type) == this->allowedTypes.end())
    {
        PushFront(this->allowedTypes, this->type);
    }
}

bool ObjectTypeDetector::IsAllowed(Type type) const
{
    if (this->type == type)
        return true;

    return std::find(this->allowedTypes.begin(), this->allowedTypes.end(), type) != this->allowedTypes.end();
}

MeshAnimationTypeDetector::MeshAnimationTypeDetector()
{
    this->type = DEFAULT;
}

MeshAnimationTypeDetector::MeshAnimationTypeDetector(INode* node, TSTR* typeTextPtr, bool useTypeText)
{
    Detect(node, typeTextPtr, useTypeText);
}

void MeshAnimationTypeDetector::Detect(INode* node, TSTR* typeTextPtr, bool useTypeText)
{
    this->type = DEFAULT;
    this->allowedTypes.clear();

    this->allowedTypes.reserve(5);
    
    //Determine if the object has skinning and/or morpher modifiers
    this->hasSkeleton = Utilities::FindSkinModifier(node) != NULL || Utilities::FindPhysiqueModifier(node) != NULL;
    this->hasMorpher = Utilities::FindMorpherModifier(node) != NULL;

    //Get vertex animation type text
    TSTR typeText;
    if (useTypeText)
    {        
        if (typeTextPtr != NULL)
            typeText = *typeTextPtr;
        else
        {
            KeyedValues properties;
            Utilities::GetNodeUserProperties(node, properties);
            TSTR typeText;
            properties.GetValue(PropertyKeys::VERTEX_ANIMATION_TYPE, typeText);
        }
    }
    
    //Determine the type
    if (this->hasSkeleton && this->hasMorpher)
    {
        this->allowedTypes.push_back(SKELETON_AND_MORPH);
        this->allowedTypes.push_back(SKELETON_AND_POSE);
        this->allowedTypes.push_back(MORPH_ONLY);

        if (Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH_ONLY))
            this->type = MORPH_ONLY;            
        else if (Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH))
            this->type = SKELETON_AND_MORPH;            
        else
            this->type = SKELETON_AND_POSE;
    }
    else if (this->hasSkeleton)
    {
        this->allowedTypes.push_back(SKELETON_ONLY);
        this->allowedTypes.push_back(MORPH_ONLY);

        if (Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH) ||
            Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH_ONLY))
        {
            this->type = MORPH_ONLY;            
        }
        else
            this->type = SKELETON_ONLY;
    }
    else if (this->hasMorpher)
    {
        this->allowedTypes.push_back(POSE_ONLY);
        this->allowedTypes.push_back(MORPH_ONLY);

        if (Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH) ||
            Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH_ONLY))
        {
            this->type = MORPH_ONLY;            
        }
        else
            this->type = POSE_ONLY;
    }
    else
    {
        if (Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH) ||
            Utilities::IsSameNoCase(typeText, PropertyValues::MeshAnimationType::MORPH_ONLY))
        {
            this->type = MORPH_ONLY;            
        }

        this->allowedTypes.push_back(MORPH_ONLY);
    }

    if (this->type != DEFAULT)
    {
        //Add the primary type to the beginning of the allowed types
        if (std::find(this->allowedTypes.begin(), this->allowedTypes.end(), this->type) == this->allowedTypes.end())
            PushFront(this->allowedTypes, this->type);
    }
}
    