
//Includes---------------------------------------------------------------------
#include "stdafx.h"
#include "Utilities.h"
#include "UniqueCoordinateTab.h"
#include "StringSplitter.h"
#include "PluginGlobals.h"
#include "resource.h"
#include "ExportSettings.h"
#include "PropertyKeys.h"
#include <shlobj.h>

using namespace OgreMax;


//Macros-----------------------------------------------------------------------
#ifdef UNICODE
    #define tsplitpath_s _wsplitpath_s
    #define tfullpath _wfullpath    
    #define tmakepath _wmakepath
#else
    #define tsplitpath_s _splitpath_s
    #define tfullpath _fullpath
    #define tmakepath _makepath
#endif

#define MORPHER_MODIFIER_CLASS_ID Class_ID(0x17bb6854, 0xa5cba2a3)
#define MAYBE_OLD_SKIN_CLASS_ID Class_ID(0x68477bb4, 0x28cf6b86) //Old skin class ID? Commented out in Max 9 SDK iskin.h
#define CURRENT_SKIN_CLASS_ID Class_ID(9815843, 87654) //Defined value for SKIN_CLASSID in Max 9 SDK iskin.h        


//Implementation---------------------------------------------------------------
bool Utilities::IsNull(const TSTR& text)
{
    //Max 8's TSTR::isNull() is not const
    //Max 9's TSTR::isNull() is const
    
    //This will work regardless of which Max is being targeted
    return const_cast<TSTR&>(text).isNull() ? true : false;
}

bool Utilities::IsSeparator(TCHAR c)
{
    return c == _T('\\') || c == _T('/');
}

void Utilities::EnsureTrailingPathSeparator(TSTR& path)
{
    if (path.length() > 0)
    {
        TCHAR lastChar = path[path.length() - 1];
        if (!IsSeparator(lastChar))
            path += _T("\\");
    }        
    else
        path = _T("\\");
}

TSTR Utilities::JoinPath(const TSTR& path1, const TSTR& path2)
{
    //If path2 is not empty and it's an absolute path, it supercedes path1
    if (path2.Length() > 1 && path2[1] == _T(':'))
        return path2;

    //Path2 is not absolute
    TSTR joined(path1);
    EnsureTrailingPathSeparator(joined);
    
    //Skip leading separators in path2
    int charIndex = 0;
    while (charIndex < path2.Length() && IsSeparator(path2[charIndex]))
        charIndex++;

    //If not at the end of path2, append it
    if (charIndex < path2.Length())
        joined += path2.Substr(charIndex, path2.Length() - charIndex);

    //Remove relative components
    joined = MakeFullPath(joined);
    
    return joined;
}

TSTR Utilities::MakeFullPath(const TSTR& path)
{
    TCHAR absPath[MAX_PATH + 1];
    tfullpath(absPath, path.data(), MAX_PATH);
    return absPath;
}

TSTR Utilities::GetFileName(const TSTR& path)
{
    TCHAR drive[MAX_PATH + 1];
    TCHAR dir[MAX_PATH + 1];
    TCHAR fileName[MAX_PATH + 1];
    TCHAR ext[MAX_PATH + 1];
    TCHAR result[MAX_PATH + 1];

    //Parse the path
    tsplitpath_s(path.data(), drive, MAX_PATH, dir, MAX_PATH, fileName, MAX_PATH, ext, MAX_PATH);

    //Generate the result
    tmakepath(result, NULL, NULL, fileName, ext);

    return result;
}

int Utilities::GetFileNameOffset(const TSTR& path)
{
    TSTR fileName = GetFileName(path);
    return path.Length() - fileName.Length();
}

TSTR Utilities::GetFileNameNoExtension(const TSTR& path)
{
    TCHAR drive[MAX_PATH + 1];
    TCHAR dir[MAX_PATH + 1];
    TCHAR fileName[MAX_PATH + 1];
    TCHAR ext[MAX_PATH + 1];

    tsplitpath_s(path.data(), drive, MAX_PATH, dir, MAX_PATH, fileName, MAX_PATH, ext, MAX_PATH);

    return fileName;
}

TSTR Utilities::GetExtension(const TSTR& path, bool includeDot)
{
    TCHAR drive[MAX_PATH + 1];
    TCHAR dir[MAX_PATH + 1];
    TCHAR fileName[MAX_PATH + 1];
    TCHAR ext[MAX_PATH + 1];

    tsplitpath_s(path.data(), drive, MAX_PATH, dir, MAX_PATH, fileName, MAX_PATH, ext, MAX_PATH);

    return (includeDot || !ext[0]) ? ext : &ext[1];
}

TSTR Utilities::RemoveExtension(const TSTR& path)
{
    TCHAR drive[MAX_PATH + 1];
    TCHAR dir[MAX_PATH + 1];
    TCHAR fileName[MAX_PATH + 1];
    TCHAR ext[MAX_PATH + 1];
    TCHAR result[MAX_PATH + 1];

    //Parse the path
    tsplitpath_s(path.data(), drive, MAX_PATH, dir, MAX_PATH, fileName, MAX_PATH, ext, MAX_PATH);

    //Generate the result
    tmakepath(result, drive, dir, fileName, NULL);

    return result;
}

void Utilities::MakeExtensionLowercase(TSTR& path)
{
    TSTR extLowercase = GetExtension(path);
    extLowercase.toLower();
    path = RemoveExtension(path) + extLowercase;
}

TSTR Utilities::GetDirectoryPath(const TSTR& fullFilePath)
{
    TCHAR drive[MAX_PATH + 1];
    TCHAR dir[MAX_PATH + 1];
    TCHAR fileName[MAX_PATH + 1];
    TCHAR ext[MAX_PATH + 1];
    TCHAR result[MAX_PATH + 1];

    //Parse the path
    tsplitpath_s(fullFilePath.data(), drive, MAX_PATH, dir, MAX_PATH, fileName, MAX_PATH, ext, MAX_PATH);

    //Generate the result
    tmakepath(result, drive, dir, NULL, NULL);

    return result;
}

void Utilities::CreateDirectoryRecursive(const TSTR& path)
{
    SHCreateDirectoryEx(NULL, path.data(), NULL);
}

void Utilities::CreateFileDirectoryRecursive(const TSTR& path)
{
    TSTR directory = GetDirectoryPath(path);
    CreateDirectoryRecursive(directory);
}

TSTR Utilities::CreateExportDirectory(const TSTR& defaultExportPath, const TSTR& subdirectory)
{
    TSTR outputPath;

    if (IsNull(subdirectory))
        outputPath = defaultExportPath;
    else
    {
        outputPath = JoinPath(defaultExportPath, subdirectory);
        CreateDirectoryRecursive(outputPath);
    }

    return outputPath;
}

bool Utilities::IsSameNoCase(const TSTR& s1, const TCHAR* s2)
{
    return _tcsicmp(s1.data(), s2) == 0;
}

bool Utilities::IsSameNoCase(const TSTR& s1, const TSTR& s2)
{
    return _tcsicmp(s1.data(), s2.data()) == 0;
}

bool Utilities::StartsWithNoCase(const TSTR& s1, const TSTR& s2)
{
    return _tcsnicmp(s1.data(), s2.data(), s2.Length()) == 0;
}

int Utilities::FindCharNoCase(const TSTR& s, TCHAR c)
{
    c = _totlower(c);

    for (int i = 0; i < s.Length(); i++)
    {
        if (_totlower(s[i]) == c)
            return i;
    }
    return -1;
}

int Utilities::FindDivide(const TSTR& s)
{
    int divideIndex = s.first('/');
    if (divideIndex == -1)
        divideIndex = s.first('\\');
    return divideIndex;
}

void Utilities::CopySubstring(TSTR& result, const TCHAR* start, int length)
{
    //It's not clear from the 3DS Max documentation whether Resize() allocates
    //the extra character for the terminator. It probably does, but add an extra
    //character anyway
    result.Resize(length + 1); 
    _tcsncpy(result.data(), start, length);
    result[length] = 0;
}

TSTR Utilities::ToString(int value)
{
    TSTR str;
    str.printf("%d", value);
    return str;
}

TSTR Utilities::ToString(DWORD value)
{
    TSTR str;
    str.printf("%u", value);
    return str;
}

TSTR Utilities::ToString(float value)
{
    stdstringstream stream;
    stream << value;    
    std::string valueText = stream.str();    
    return valueText.c_str();
}

TSTR Utilities::ToString(double value)
{
    stdstringstream stream;
    stream << value;    
    std::string valueText = stream.str();    
    return valueText.c_str();
}

TSTR Utilities::ToString(Color value)
{
    stdstringstream stream;
    stream << value.r << ',' << value.g << ',' << value.b;
    std::string valueText = stream.str();    
    return valueText.c_str();
}

TSTR Utilities::ToString(Point3 value)
{
    stdstringstream stream;
    stream << value.x << ',' << value.y << ',' << value.z;
    std::string valueText = stream.str();    
    return valueText.c_str();
}

bool Utilities::ToBool(const TSTR& value)
{
    if (IsNull(value) || IsSameNoCase(value, _T("false")) || IsSameNoCase(value, _T("no")) || IsSameNoCase(value, _T("0")))
        return false;
    else
        return true;
}

float Utilities::ToFloat(const TSTR& value)
{
    return (float)_tstof(value.data());
}

unsigned int Utilities::ToUInt(const TSTR& value)
{
    return (unsigned int)ToInt(value.data());
}

int Utilities::ToInt(const TSTR& value)
{
    return _tstoi(value.data());
}

Point3 Utilities::ToPoint3(const TSTR& text)
{
    TSTR trimmedText = text;
    TrimWhitespace(trimmedText);
    
    int firstComma = trimmedText.first(',');
    int lastComma = trimmedText.last(',');

    Point3 color(0,0,0);

    if (firstComma != -1 && lastComma != -1 && firstComma != lastComma)
    {
        TSTR red = trimmedText.Substr(0, firstComma);
        TrimWhitespace(red);

        TSTR green = trimmedText.Substr(firstComma + 1, lastComma - firstComma - 1);
        TrimWhitespace(green);

        TSTR blue = trimmedText.Substr(lastComma + 1, trimmedText.length() - lastComma - 1);
        TrimWhitespace(blue);

        color.x = _tstof(red);
        color.y = _tstof(green);
        color.z = _tstof(blue);
    }

    return color;
}

float Utilities::Average(Point3 xyz)
{
    return (xyz.x + xyz.y + xyz.z) / 3;
}

float Utilities::GetBoxHalfSize(const Box3& box)
{
    return Average(box.Width()) / 2;
}

TSTR Utilities::GetPropertyString(int propertySet, PROPID propid)
{
    TSTR propString;

    PROPSPEC propSpec;
	propSpec.ulKind = PRSPEC_PROPID;
	propSpec.propid = propid;
	int index = GetCOREInterface()->FindProperty(propertySet, &propSpec);	
    if (index != -1) 
    {
	    const PROPVARIANT* prop = GetCOREInterface()->GetPropertyVariant(propertySet, index);
        if (prop && prop->vt == VT_LPSTR)
            propString = prop->pszVal;
    }

    return propString;
}

void Utilities::SetPropertyString(int propertySet, PROPID propid, const TSTR& text)
{
    PROPSPEC propSpec;
	propSpec.ulKind = PRSPEC_PROPID;
	propSpec.propid = propid;
	
    PROPVARIANT prop;
    prop.vt = VT_LPSTR;
    prop.pszVal = (TCHAR*)text.data();

    GetCOREInterface()->AddProperty(propertySet, &propSpec, &prop);
}

TSTR Utilities::GetSceneSummaryInfo()
{
    //The scene summary information appears in two places in 3DS Max:
    // 1)File->Summary Info...->Description
    // 2)File->File Properties->Summary->Comments
    
    return Utilities::GetPropertyString(PROPSET_SUMMARYINFO, PIDSI_COMMENTS);
}

void Utilities::SetSceneSummaryInfo(const TSTR& text)
{
    Utilities::SetPropertyString(PROPSET_SUMMARYINFO, PIDSI_COMMENTS, text);
}

INode* Utilities::GetNodeValue(IGameProperty* gameProperty)
{
    INode* node = NULL;
    
    if (gameProperty->IsPBlock2())
    {
        IParamBlock2* pblock2 = gameProperty->GetMaxParamBlock2();
        ParamID id = pblock2->IndextoID(gameProperty->GetParamBlockIndex());		
        node = pblock2->GetINode(id);
    }
    
    return node;
}

void Utilities::TrimWhitespace(TSTR& text)
{
    int start = 0;
    int length = text.length();
    while (start < length && _istspace(text[start]))
        start++;

    if (start == length)
    {
        //Reached end, the entire string is whitespace
        text = _T("");
    }
    else
    {
        //There is at least one non-whitespace character

        int end = length - 1;
        while (end > start && _istspace(text[end]))
            end--;

        TSTR trimmed = text.Substr(start, end - start + 1);
        text = trimmed;
    }
}

bool Utilities::IsNewline(TCHAR c)
{
    return c == _T('\r') || c == _T('\n');
}

int Utilities::FindNextLineStart(const TSTR& text, int start, bool skipExtraNewlines)
{
    int index = start;
    int length = text.length();
    
    //Skip past all non-newline characters
    while (index < length && !IsNewline(text[index]))
        index++;

    //Skip all the remaining newlines
    while (index < length && IsNewline(text[index]))
    {
        index++;

        if (!skipExtraNewlines && text[index - 1] == _T('\n'))
        {
            //Caller wants to quit after the first newline is found
            break;
        }
    }

    return index;
}

bool Utilities::ParseKeyValueLine(const TSTR& text, int start, int& nextStart, TSTR& key, TSTR& value, bool skipExtraNewlines)
{    
    int index = start;
    int length = text.length();

    //Exit early if the end has already been reached, the leading text is whitespace,
    //or the leading text is a # (comment character)
    if (index >= length || _istspace(text[start]) || text[start] == _T('#'))
    {
        nextStart = FindNextLineStart(text, index, skipExtraNewlines);
        return false;
    }

    //Find the key. This is the text to the left of the '=' character
    while (index < length && !IsNewline(text[index]))
    {
        if (text[index] == _T('='))
            break;
        index++;
    }

    //A key has been found if the current index has moved past the start index,
    //the index didn't reach the end of the text, and the current character is the
    //'=' character
    if (index == start || index == length || text[index] != _T('='))
    {
        //No key found
        nextStart = FindNextLineStart(text, index, skipExtraNewlines);
        return false;
    }
    else
    {
        //A key was found
        key = text.Substr(start, index - start);
        TrimWhitespace(key);
    }

    //If we're this far, we have the index of the '=' character. Everything to the right is the value
    index++; //Skip the '=' character
    int valueStart = index;
    while (index < length && !IsNewline(text[index]))
        index++;

    //Get the value
    value = text.Substr(valueStart, index - valueStart + 1);
    TrimWhitespace(value);

    //Get next start location
    nextStart = FindNextLineStart(text, index, skipExtraNewlines);

    return true;
}

bool Utilities::ParseKeyValueLine(const TSTR& text, TSTR& key, TSTR& value)
{
    int nextIndex;
    return ParseKeyValueLine(text, 0, nextIndex, key, value);
}

bool Utilities::ParsePrefixedKeyValueLine(const TSTR& text, TSTR& key, TSTR& value, TSTR* prefix)
{
    int index = 0;
    int length = text.Length();

    //Move to first space
    while (index < length && !_istspace(text[index]))
        index++;
    if (index >= length)
        return false;

    if (prefix != NULL)
        *prefix = text.Substr(0, index);

    //Move to first non-space
    while (index < length && _istspace(text[index]))
        index++;
    if (index >= length)
        return false;

    int nextIndex;
    return ParseKeyValueLine(text, index, nextIndex, key, value);
}

bool Utilities::ParseUserData(const TSTR& text, const TSTR& userDataMarker, TSTR& userData)
{
    //There must be 3 occurrances of the marker:
    // 1)The first one is an indicator of the marker itself in the configuration data,
    //   for example: userDataMarker=--MyUserData--
    //   The first instance is skipped
    // 2)The second one is the beginning of the user data
    // 3)The third one is the end of the user data

    int userDataMarkerLength = userDataMarker.length();

    //Find the first instance
    const TCHAR* firstInstance = _tcsstr(text.data(), userDataMarker.data());
    if (!firstInstance)
        return false;

    //Find the second instance
    const TCHAR* secondInstance = _tcsstr(firstInstance + userDataMarkerLength, userDataMarker.data());
    if (!secondInstance)
        return false;
    
    //Find the third instance
    const TCHAR* userDataStart = secondInstance + userDataMarkerLength;
    const TCHAR* userDataEnd = _tcsstr(userDataStart, userDataMarker.data());
    if (!userDataEnd)
        return false;
    
    //Extract the user data substring
    int userDataStartIndex = userDataStart - text.data();
    int userDataLength = userDataEnd - userDataStart;
    userData = text.Substr(userDataStartIndex, userDataLength);
    TrimWhitespace(userData);

    return true;
}

int Utilities::ShowMessage(const TSTR& message, const TSTR& title, UINT type)
{
    return MessageBox(GetCOREInterface()->GetMAXHWnd(), message.data(), title.data(), type);
}

int Utilities::ShowMessage(int messageID, int titleID, UINT type)
{
    return MessageBox
        (
        GetCOREInterface()->GetMAXHWnd(), 
        PluginGlobals::GetString(messageID), 
        PluginGlobals::GetString2(titleID), 
        type
        );
}

void Utilities::ShowMeshInfo(IGameObject* gameObject, IGameNode* gameNode, const TCHAR* header, UINT type)
{
    if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)
    {
        IGameMesh* gameMesh = (IGameMesh*)gameObject;
        TSTR msg;
        msg.printf("%s: %d vertices, %d faces", header, gameMesh->GetNumberOfVerts(), gameMesh->GetNumberOfFaces());
        Utilities::ShowMessage(msg, gameNode->GetName(), type);
    }
}

void Utilities::ShowClassID(Class_ID cid, const TSTR& title, UINT type)
{
    TSTR cidText;
    cidText.printf("ClassID: 0x%x, 0x%x", cid.PartA(), cid.PartB());
    ShowMessage(cidText, title, type);
}

void Utilities::ShowAbout()
{
    TSTR newline = _T("\r\n\r\n");

    //Reformat the version string from "a, b, c, d" to "a.b.c"
    TSTR version = PluginGlobals::versionInfo.GetProductVersion();
    ReformatVersion(version);

    //Generate the "Version: a.b.c" string
    TSTR versionFormat = PluginGlobals::GetString(IDS_VERSION_FORMAT);
    TSTR versionText;
    versionText.printf(versionFormat, version.data());
    
    //Create the message
    TSTR message = 
        PluginGlobals::versionInfo.GetFileDescription() + newline + 
        PluginGlobals::versionInfo.GetLegalCopyright() + newline + 
        versionText;

    //Show the message
    ShowMessage(message, PluginGlobals::GetString(IDS_ABOUT_CAPTION), MB_ICONINFORMATION);
}

void Utilities::ShowNodeErrorMessage(IGameNode* gameNode, int stringResourceID)
{
    TSTR formatString = PluginGlobals::GetString(stringResourceID);
    TSTR message;
    message.printf(formatString, gameNode->GetName());
    
    TSTR title = PluginGlobals::GetString(IDS_EXPORT_ERROR_CAPTION);

    ShowMessage(message, title);
}

TiXmlElement* Utilities::AddColorElement(TiXmlNode* parentNode, const TCHAR* elementName, Point3 color)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    element->SetAttribute(_T("r"), ToString(color.x));
    element->SetAttribute(_T("g"), ToString(color.y));
    element->SetAttribute(_T("b"), ToString(color.z));
    return element;
}

TiXmlElement* Utilities::AddColorElement(TiXmlNode* parentNode, const TCHAR* elementName, Color color)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    element->SetAttribute(_T("r"), ToString(color.r));
    element->SetAttribute(_T("g"), ToString(color.g));
    element->SetAttribute(_T("b"), ToString(color.b));
    return element;
}

TiXmlElement* Utilities::AddXYZElement(TiXmlNode* parentNode, const TCHAR* elementName, Point3 xyz)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    SetXYZAttributes(element, xyz);
    return element;
}

TiXmlElement* Utilities::AddXYZElement(TiXmlNode* parentNode, const TCHAR* elementName, Point4 xyz)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    SetXYZAttributes(element, xyz);
    return element;
}

void Utilities::SetXYZAttributes(TiXmlElement* element, Point3 xyz)
{
    element->SetAttribute(_T("x"), ToString(xyz.x));
    element->SetAttribute(_T("y"), ToString(xyz.y));
    element->SetAttribute(_T("z"), ToString(xyz.z));
}

void Utilities::SetXYZAttributes(TiXmlElement* element, Point4 xyz)
{
    element->SetAttribute(_T("x"), ToString(xyz.x));
    element->SetAttribute(_T("y"), ToString(xyz.y));
    element->SetAttribute(_T("z"), ToString(xyz.z));
}

TiXmlElement* Utilities::AddFaceElement(TiXmlElement* parentNode, const TCHAR* elementName, FaceEx* face)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    element->SetAttribute(_T("v0"), ToString(face->vert[0]));
    element->SetAttribute(_T("v1"), ToString(face->vert[1]));
    element->SetAttribute(_T("v2"), ToString(face->vert[2]));
    return element;
}

TiXmlElement* Utilities::AddRotationElement
    (
    TiXmlNode* parentNode, 
    const TCHAR* elementName, 
    Quat q, 
    RotationType rotationType
    )
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));

    switch (rotationType)
    {
        case ROTATION_QUATERNION:
        {
            q.Invert();
            element->SetAttribute(_T("qx"), ToString(q.x));
            element->SetAttribute(_T("qy"), ToString(q.y));
            element->SetAttribute(_T("qz"), ToString(q.z));
            element->SetAttribute(_T("qw"), ToString(q.w));
            break;
        }
        case ROTATION_ANGLE_AXIS:
        {
            AngAxis angAxis(q);
            element->SetAttribute(_T("axisX"), ToString(angAxis.axis.x));
            element->SetAttribute(_T("axisY"), ToString(angAxis.axis.y));
            element->SetAttribute(_T("axisZ"), ToString(angAxis.axis.z));
            element->SetAttribute(_T("angle"), ToString(-angAxis.angle));
            break;
        }
        case ROTATION_EULER_ANGLE:
        {
            float x, y, z;
            q.GetEuler(&x, &y, &z);
            element->SetAttribute(_T("angleX"), ToString(x));
            element->SetAttribute(_T("angleY"), ToString(y));
            element->SetAttribute(_T("angleZ"), ToString(z));
        }
    }
    
    return element;
}

TiXmlElement* Utilities::AddRayElement(TiXmlNode* parentNode, Point3 position, Point3 direction)
{
    TiXmlElement* element = (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(_T("Ray")));
    element->SetAttribute(_T("originx"), position.x);
    element->SetAttribute(_T("originy"), position.y);
    element->SetAttribute(_T("originz"), position.z);
    element->SetAttribute(_T("directionx"), direction.x);
    element->SetAttribute(_T("directiony"), direction.y);
    element->SetAttribute(_T("directionz"), direction.z);    
    return element;
}

TiXmlElement* Utilities::AddTextElement(TiXmlNode* parentNode, const TCHAR* elementName, const TSTR& text, bool isCDATA)
{
    TiXmlElement* element = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(elementName));
    TiXmlText* textNode = (TiXmlText*)element->LinkEndChild(new TiXmlText(text.data()));
    textNode->SetCDATA(isCDATA);
    return element;
}

TiXmlElement* Utilities::AddClippingElement
    (
    TiXmlNode* parentNode,
    float nearClip,
    float farClip
    )
{
    TiXmlElement* clippingXmlElement = 
        (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(_T("clipping")));

    clippingXmlElement->SetAttribute(_T("near"), ToString(nearClip));        
    clippingXmlElement->SetAttribute(_T("far"), ToString(farClip));

    return clippingXmlElement;
}

TSTR Utilities::LookupContactGroupInteger(const TSTR& contactGroup, const KeyedValues& sceneProperties)
{
    TSTR result = contactGroup;

    //Translate value to integer if the retrieved value isn't an integer
    if (!Utilities::IsInteger(contactGroup))
    {
        TSTR contactGroupKey = PropertyKeys::CONTACT_GROUP;
        contactGroupKey += _T(" ");
        contactGroupKey += contactGroup;                    
        TSTR contactGroupValue;
        if (sceneProperties.GetValue(contactGroupKey, contactGroupValue))
            result = contactGroupValue;        
    }

    return result;
}

bool Utilities::GetObjectContactGroupInteger(TSTR& contactGroup, const KeyedValues& objectProperties, const KeyedValues& sceneProperties)
{
    contactGroup = _T("");

    if (objectProperties.GetValue(PropertyKeys::CONTACT_GROUP, contactGroup))
        contactGroup = LookupContactGroupInteger(contactGroup, sceneProperties);
    
    return !contactGroup.isNull();
}

TiXmlElement* Utilities::AddOpalContactGroupSetupElement(TiXmlNode* parentNode, const TSTR& contactGroupSetup, const KeyedValues& sceneProperties)
{
    TiXmlElement* contactGroupXmlElement = (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(_T("ContactGroup")));

    StringSplitter split(contactGroupSetup.data(), _T(','));

    TSTR makeContacts = _T("enabled");
    TSTR group0;
    TSTR group1;

    if (split.size() == 1)
    {
        //One group and enabled
        group0 = LookupContactGroupInteger(split[0], sceneProperties);
    }
    else if (split.size() == 2)
    {
        if (IsSameNoCase(split[1], _T("enabled")) || IsSameNoCase(split[1], _T("disabled")))
        {
            //Either one group enabled/disabled
            group0 = LookupContactGroupInteger(split[0], sceneProperties);
            makeContacts = split[1];
        }
        else
        {
            //Two groups enabled
            group0 = LookupContactGroupInteger(split[0], sceneProperties);
            group1 = LookupContactGroupInteger(split[1], sceneProperties);
        }
    }
    else if (split.size() == 3)
    {
        //Two groups enabled/disabled
        group0 = LookupContactGroupInteger(split[0], sceneProperties);
        group1 = LookupContactGroupInteger(split[1], sceneProperties);
        makeContacts = split[2];
    }

    //Export 'makeContacts', turning it from enabled/disabled to true/false
    if (IsSameNoCase(makeContacts, _T("enabled")))
        makeContacts = _T("true");
    else
        makeContacts = _T("false");
    contactGroupXmlElement->SetAttribute(_T("makeContacts"), makeContacts);

    //Export groups
    if (group1.isNull())
    {
        //One group        
        contactGroupXmlElement->SetAttribute(_T("group"), group0);
    }
    else
    {
        //Two groups
        contactGroupXmlElement->SetAttribute(_T("group0"), group0);
        contactGroupXmlElement->SetAttribute(_T("group1"), group1);
    }

    return contactGroupXmlElement;
}

TiXmlElement* Utilities::AddOpalContactGroupElement(TiXmlNode* parentNode, const TSTR& contactGroup)
{
    TiXmlElement* contactGroupXmlElement = (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(_T("ContactGroup")));
    contactGroupXmlElement->SetAttribute(_T("value"), contactGroup.data());
    return contactGroupXmlElement;
}

TiXmlElement* Utilities::AddOpalShapeMaterialElement(TiXmlNode* parentNode, const KeyedValues& properties)
{
    TiXmlElement* materialXmlElement = NULL;

    TSTR type, hardness, friction, bounciness, density;
    bool hasType = properties.GetValue(PropertyKeys::MATERIAL_TYPE, type);
    bool hasHardness = properties.GetValue(PropertyKeys::MATERIAL_HARDNESS, hardness);
    bool hasFriction = properties.GetValue(PropertyKeys::MATERIAL_FRICTION, friction);
    bool hasBounciness = properties.GetValue(PropertyKeys::MATERIAL_BOUNCINESS, bounciness);
    bool hasDensity = properties.GetValue(PropertyKeys::MATERIAL_DENSITY, density);

    if (hasType)
    {
        bool isValidType = true;

        if (IsSameNoCase(type, _T("metal")))
        {
            if (!hasHardness) hardness = _T("1.0");
		    if (!hasFriction) friction = _T("0.7");
		    if (!hasBounciness) bounciness = _T("0.05");
		    if (!hasDensity) density = _T("1.0");
        }
        else if (IsSameNoCase(type, _T("wood")))
        {
		    if (!hasHardness) hardness = _T("0.8");
		    if (!hasFriction) friction = _T("1.0");
		    if (!hasBounciness) bounciness = _T("0.15");
		    if (!hasDensity) density = _T("0.2");
        }
        else if (IsSameNoCase(type, _T("rubber")))
        {
		    if (!hasHardness) hardness = _T("0");
		    if (!hasFriction) friction = _T("1.0");
		    if (!hasBounciness) bounciness = _T("1.0");
		    if (!hasDensity) density = _T("0.4");
        }
        else if (IsSameNoCase(type, _T("ice")))
        {
		    if (!hasHardness) hardness = _T("1.0");
		    if (!hasFriction) friction = _T("0");
		    if (!hasBounciness) bounciness = _T("0.05");
		    if (!hasDensity) density = _T("0.25");
        }
        else
            isValidType = false;
            
        if (isValidType)
            hasHardness = hasFriction = hasBounciness = hasDensity = true;
    }
    
    if (hasHardness || hasFriction || hasBounciness || hasDensity)
    {
        materialXmlElement = 
            (TiXmlElement*)parentNode->LinkEndChild(new TiXmlElement(_T("Material")));

        if (hasType)
            materialXmlElement->SetAttribute(_T("type"), type);        
        if (hasHardness)
            materialXmlElement->SetAttribute(_T("hardness"), hardness);        
        if (hasFriction)
            materialXmlElement->SetAttribute(_T("friction"), friction);        
        if (hasBounciness)
            materialXmlElement->SetAttribute(_T("bounciness"), bounciness);        
        if (hasDensity)
            materialXmlElement->SetAttribute(_T("density"), density);        
    }
    
    return materialXmlElement;
}

TiXmlElement* Utilities::AddOpalAxisElement
    (
    TiXmlElement* jointXmlElement, 
    int axisNumber, 
    const GMatrix& transform, 
    IPropertyContainer* propertyContainer,
    const KeyedValues& properties,
    TimeValue timeValue
    )
{
    //Axis
    TiXmlElement* axisXmlElement = (TiXmlElement*)jointXmlElement->LinkEndChild(new TiXmlElement(_T("Axis")));

    //Axis number
    TiXmlElement* axisNumberXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("AxisNum")));
    axisNumberXmlElement->SetAttribute(_T("value"), ToString(axisNumber));

    //Axis direction
    TSTR jointAxisName;
    jointAxisName.printf(PropertyKeys::JOINT_AXIS_FORMAT.data(), axisNumber + 1);
    int rowIndex = axisNumber;
    TSTR axisText;
    if (properties.GetValue(jointAxisName, axisText) || 
        (axisNumber == 0 && properties.GetValue(PropertyKeys::JOINT_AXIS, axisText)))
    {
        if (Utilities::IsSameNoCase(axisText, _T("x")))
            rowIndex = 0;
        else if (Utilities::IsSameNoCase(axisText, _T("y")))
            rowIndex = 1;
        else if (Utilities::IsSameNoCase(axisText, _T("z")))
            rowIndex = 2;
    }
    Point4 axisRow = transform.GetRow(rowIndex);
    Point3 axis(axisRow.x, axisRow.y, axisRow.z);
    AddXYZElement(axisXmlElement, _T("Direction"), axis);

    //Limits
    int isLimited = 0;
    IGameProperty* isLimitedGameProperty = propertyContainer->QueryProperty(_T("isLimited"));
    isLimitedGameProperty->GetPropertyValue(isLimited, timeValue);
    if (isLimited)
    {
        //Limits enabled?
        TiXmlElement* limitsEnabledXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("LimitsEnabled")));
        limitsEnabledXmlElement->SetAttribute(_T("value"), _T("true"));

        //Low limit
        float minLimitAngle;
        IGameProperty* minLimitAngleGameProperty = propertyContainer->QueryProperty(_T("minLimitAngle"));
        minLimitAngleGameProperty->GetPropertyValue(minLimitAngle, timeValue);
        TiXmlElement* lowLimitXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("LimitLow")));
        lowLimitXmlElement->SetAttribute(_T("value"), ToString(minLimitAngle));

        //High limit
        float maxLimitAngle;
        IGameProperty* maxLimitAngleGameProperty = propertyContainer->QueryProperty(_T("maxLimitAngle"));
        maxLimitAngleGameProperty->GetPropertyValue(maxLimitAngle, timeValue);
        TiXmlElement* highLimitXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("LimitHigh")));
        highLimitXmlElement->SetAttribute(_T("value"), ToString(maxLimitAngle));
        
        float hardness;
        IGameProperty* hardnessGameProperty = propertyContainer->QueryProperty(_T("strength"));
        hardnessGameProperty->GetPropertyValue(hardness, timeValue);
        hardness = Utilities::Limit(hardness / 100, 0, 1);
        TiXmlElement* hardnessXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("LimitHardness")));
        hardnessXmlElement->SetAttribute(_T("value"), ToString(hardness));
        
        float bounciness;
        IGameProperty* bouncinessGameProperty = propertyContainer->QueryProperty(_T("tau"));
        bouncinessGameProperty->GetPropertyValue(bounciness, timeValue);
        bounciness = Utilities::Limit(bounciness / 100, 0, 1);
        TiXmlElement* bouncinessXmlElement = (TiXmlElement*)axisXmlElement->LinkEndChild(new TiXmlElement(_T("LimitBounciness")));
        bouncinessXmlElement->SetAttribute(_T("value"), ToString(bounciness));
    }

    return axisXmlElement;
}

void Utilities::SetPlaneAttributes(TiXmlElement* element, Point4 plane)
{
    element->SetAttribute(_T("planeX"), ToString(plane.x));
    element->SetAttribute(_T("planeY"), ToString(plane.y));
    element->SetAttribute(_T("planeZ"), ToString(plane.z));
    element->SetAttribute(_T("planeD"), ToString(plane.w));
}

bool Utilities::GetTiling(IGameMaterial* gameMaterial, TimeValue timeValue, Point2& tiling)
{
    if (gameMaterial && gameMaterial->GetNumberOfTextureMaps() > 0)
    {
        IGameTextureMap* gameTexture = gameMaterial->GetIGameTextureMap(0);
        if (gameTexture)
        {
            IGameUVGen* gameUVGen = gameTexture->GetIGameUVGen();
            
            IGameProperty* uTiling = gameUVGen->GetUTilingData();
            uTiling->GetPropertyValue(tiling.x, timeValue);

            IGameProperty* vTiling = gameUVGen->GetVTilingData();
            vTiling->GetPropertyValue(tiling.y, timeValue);

            return true;
        }
    }

    return false;
}

bool Utilities::GetTiling(IGameMaterial* gameMaterial, TimeValue timeValue, float& tiling)
{    
    Point2 tiling2;
    if (GetTiling(gameMaterial, timeValue, tiling2))
    {
        tiling = tiling2.x;
        return true;
    }
    return false;
}

bool Utilities::GetSphereSegmentCount
    (
    IGameObject* gameObject, 
    const KeyedValues& properties, 
    TimeValue timeValue, 
    IPoint2& segmentCount
    )
{
    //Reset segment counts
    segmentCount.x = 0;
    segmentCount.y = 0;

    Object* maxObject = gameObject->GetMaxObject();
    if (maxObject->ClassID() == Class_ID(SPHERE_CLASS_ID,0))
    {
        //It's a sphere object. Get the segment count directly from the object

        const int PB_SEGS = 1; //Defined in maxsdk\samples\objects\sphere.cpp
        IParamArray* params = maxObject->GetParamBlock();
        int segments;
        params->GetValue(PB_SEGS, timeValue, segments, FOREVER);

        segmentCount.x = segments;
        segmentCount.y = segments / 2 - 1;
    }
    
    return segmentCount.x > 0 && segmentCount.y > 0;
}

bool Utilities::GetPlaneMeshSegmentCount(IGameObject* gameObject, IPoint2& segmentCount)
{
    if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)
    {
        //Figure out the number of x and y segments by examining the mesh
        
        //Get the 3DS Max mesh
        IGameMesh* gameMesh = (IGameMesh*)gameObject;
        Mesh* maxMesh = gameMesh->GetMaxMesh();

        //Add the vertex x/y coordinates to the lookup tables
        UniqueCoordinateTab xCoordinates;
        UniqueCoordinateTab yCoordinates;
        for (int index = 0; index < maxMesh->numVerts; index++)
        {
            xCoordinates.AddCoordinate(maxMesh->verts[index].x);
            yCoordinates.AddCoordinate(maxMesh->verts[index].y);
        }

        //The number of segments is one less than the number of unique coordinates
        //on a particular dimension
        segmentCount.x = xCoordinates.Count() - 1;
        segmentCount.y = yCoordinates.Count() - 1;
    }

    return segmentCount.x > 0 && segmentCount.y > 0;
}

Point4 Utilities::CreatePlane(Point3 p0, Point3 p1, Point3 p2)
{
    Point3 normal = CreateNormal(p0, p1, p2);

    return Point4(normal.x, normal.y, normal.z, -DotProd(normal, p0));
}

Point3 Utilities::CreateNormal(Point3 p0, Point3 p1, Point3 p2)
{
    Point3 v0 = p1 - p0;
    Point3 v1 = p2 - p1;

    Point3 result = CrossProd(v0, v1);
    result = Normalize(result);
    return result;
}

bool Utilities::IsMirrored(IGameNode* gameNode)
{
    GMatrix transform = gameNode->GetWorldTM(0);
    return transform.Determinant() < 0;
}

void Utilities::FixMirroring(GMatrix& matrix)
{
    static const GMatrix MirrorMatrix(ScaleMatrix(Point3(-1, 1, 1)));

    if (matrix.Determinant() < 0)
        matrix = MirrorMatrix * matrix;
}

GMatrix Utilities::GetNodeTM(IGameNode* gameNode, TimeValue timeValue, bool relativeToParent)
{
    GMatrix transform;

    if (relativeToParent && gameNode->GetNodeParent() && !gameNode->GetNodeParent()->GetMaxNode()->IsRootNode())
        transform = GetNodeTMRelativeTo(gameNode, gameNode->GetNodeParent(), timeValue);        
    else
    {
        transform = gameNode->GetWorldTM(timeValue);
        FixMirroring(transform);
    }

    return transform;
}

GMatrix Utilities::GetNodeTMRelativeTo(IGameNode* gameNode, IGameNode* parentNode, TimeValue timeValue)
{
    GMatrix transform = gameNode->GetWorldTM(timeValue) * parentNode->GetWorldTM(timeValue).Inverse();
    FixMirroring(transform);
    return transform;
}

GMatrix Utilities::GetNodeTMRelativeToSelf(IGameNode* gameNode, TimeValue referenceTime, TimeValue currentTime)
{
    GMatrix transform = gameNode->GetWorldTM(currentTime) * gameNode->GetWorldTM(referenceTime).Inverse();
    return transform;
}

TSTR Utilities::MakeModelName(const TSTR& name)
{
    return name + _T(".model");
}

TSTR Utilities::MakeMaterialName(IGameMaterial* gameMaterial, const TSTR& defaultName)
{
    if (gameMaterial != NULL)
        return MakeMaterialName(gameMaterial->GetMaterialName());
    else
        return defaultName;
}

TSTR Utilities::MakeMaterialName(const TSTR& name)
{
	std::string test=name.data();
	for (size_t i=0;i<test.length();++i) {
		if (false==(test[i]<='z'&&test[i]>='a')&&
			false==(test[i]<='Z'&&test[i]>='A')&&
			false==(test[i]<='9'&&test[i]>='0')) {
			test[i]='_';
		}
	}
	TSTR retval(test.c_str());
    return retval;
}

TSTR Utilities::MakeMaterialFileName(const TSTR& name)
{
    return name + _T(".os");
}

TSTR Utilities::MakeMeshFileName(const TSTR& name, bool isXml)
{
    if (isXml)
        return name + _T(".mesh.xml");
    else
        return name + _T(".mesh");
}

TSTR Utilities::MakeSkeletonFileName(const TSTR& name, bool isXml)
{
    if (isXml)
        return name + _T(".skeleton.xml");
    else
        return name + _T(".skeleton");
}

bool Utilities::CanExportMaterial(const TSTR& materialName, const KeyedValues& sceneProperties)
{
    //Create the key to look up
    TSTR key = PropertyKeys::EXPORT_MATERIAL;
    key += _T(" ");
    key += materialName;

    //Look up the key value
    bool exportMaterial = true;
    TSTR exportMaterialText;
    if (sceneProperties.GetValue(key, exportMaterialText))
        exportMaterial = ToBool(exportMaterialText);

    return exportMaterial;
}

bool Utilities::IsValidMeshNameCharacter(TCHAR c)
{
    return
        isalnum(c) ||
        c == ' ' ||
        c == '.' ||
        c == '-' ||
        c == '_';
}

void Utilities::FixMeshName(TSTR& meshName)
{
    for (int charIndex = 0; charIndex < meshName.length(); charIndex++)
    {
        if (!IsValidMeshNameCharacter(meshName[charIndex]))
            meshName[charIndex] = _T('_');
    }
}

bool Utilities::CanExportMesh(const TSTR& meshName, const KeyedValues& sceneProperties)
{
    //Create the key to look up
    TSTR key = PropertyKeys::EXPORT_MESH;
    key += _T(" ");
    key += meshName;

    //Look up the key value
    bool exportMesh = true;
    TSTR exportMeshText;
    if (sceneProperties.GetValue(key, exportMeshText))
        exportMesh = ToBool(exportMeshText);

    return exportMesh;
}

TSTR Utilities::GetRenamedBitmapFileName(const TSTR& fileName, const KeyedValues& sceneProperties)
{
    TSTR remappedFileName;
    
    //Create the key to look up the full file name
    TSTR key = PropertyKeys::RENAME_BITMAP;
    key += _T(" ");
    key += fileName;

    //Look up the remapped file name
    if (!sceneProperties.GetValue(key, remappedFileName))
    {
        //There was no remapped file name. Look for its file extension
        TSTR extension = GetExtension(fileName, false);

        //Create the key to look up the file extension        
        key = PropertyKeys::RENAME_BITMAP_EXTENSION;
        key += _T(" ");
        key += extension;

        TSTR remappedExtension;
        if (sceneProperties.GetValue(key, remappedExtension))
        {
            //There was a remapped file extension
            remappedFileName = GetFileNameNoExtension(fileName);
            remappedFileName += _T(".");
            remappedFileName += remappedExtension;
        }        
    }
    
    return !remappedFileName.isNull() ? remappedFileName : fileName;
}

void Utilities::GetNodeUserProperties(INode* maxNode, KeyedValues& properties)
{
    TSTR propertyText;
    maxNode->GetUserPropBuffer(propertyText);
    properties.Parse(propertyText);
}

void Utilities::GetNodeUserProperties(IGameNode* gameNode, KeyedValues& properties)
{
    GetNodeUserProperties(gameNode->GetMaxNode(), properties);
}

TSTR Utilities::GetMeshName(IGameNode* gameNode, const KeyedValues& properties)
{
    //Get mesh name
    TSTR meshName;
    if (!properties.GetValue(PropertyKeys::MESH_NAME, meshName))
        meshName = gameNode->GetName();

    //Remove invalid characters
    FixMeshName(meshName);

    return meshName;
}

bool Utilities::GetMeshPlane(IGameObject* gameObject, Point4& plane, bool worldSpace)
{
    if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)
    {
        IGameMesh* gameMesh = (IGameMesh*)gameObject;
        if (gameMesh->GetNumberOfFaces() > 0)
        {
            bool objectSpace = !worldSpace;

            FaceEx* face = gameMesh->GetFace(0);
            Point3 p0 = gameMesh->GetVertex(face->vert[0], objectSpace);
            Point3 p1 = gameMesh->GetVertex(face->vert[1], objectSpace);
            Point3 p2 = gameMesh->GetVertex(face->vert[2], objectSpace);

            plane = Utilities::CreatePlane(p0, p1, p2);
            
            return true;
        }
    }

    return false;
}

bool Utilities::GetMeshLocalBoundingBox(IGameObject* gameObject, Box3& box)
{
    if (gameObject->GetIGameType() == IGameObject::IGAME_MESH)
    {
        IGameMesh* gameMesh = (IGameMesh*)gameObject;
        box = gameMesh->GetMaxMesh()->getBoundingBox();

        return true;
    }

    return false;
}

float Utilities::GetBoundingSphereRadius(IGameMesh* gameMesh, Point3 axes)
{
    float radius = 0;
    Mesh* maxMesh = gameMesh->GetMaxMesh();

    if (maxMesh->getNumVerts() > 0)
    {
        //Get the squared radius for each vertex (faster than getting the radius)
        float radiusSquared = 0;
        for (int index = 0; index < maxMesh->getNumVerts(); index++)
        {
            Point3 vertex = maxMesh->getVert(index) * axes;
            radiusSquared = std::max(radiusSquared, vertex.LengthSquared());
        }

        //Get the radius
        radius = sqrtf(radiusSquared);
    }

    return radius;
}

float Utilities::GetMeshLength(IGameMesh* gameMesh)
{
    Box3 box = gameMesh->GetMaxMesh()->getBoundingBox();
    Point3 boxSize = box.Width();
    return boxSize.z;
}

void Utilities::Abs(Point3& p)
{
    p.x = fabs(p.x);
    p.y = fabs(p.y);
    p.z = fabs(p.z);
}

float Utilities::Limit(float value, float minValue, float maxValue)
{
    if (value < minValue)
        value = minValue;
    else if (value > maxValue)
        value = maxValue;
    return value;
}

Point3 Utilities::Lerp(Point3 p0, Point3 p1, float t)
{
    return p0 + (p1 - p0) * t;
}

Point3 Utilities::GetSampledPoint3Key(const IGameKeyTab& positionKeys, TimeValue keyTime, Point3 defaultValue)
{
    Point3 key = defaultValue;
    
    if (positionKeys.Count() > 0)
    {
        if (keyTime >= positionKeys[positionKeys.Count() - 1].t)
        {
            //The key time goes beyond the specified keys. Take the last key
            key = positionKeys[positionKeys.Count() - 1].sampleKey.pval;
        }
        else
        {
            //The key time is within range
            for (int keyIndex = 0; keyIndex < positionKeys.Count(); keyIndex++)
            {
                if (keyTime >= positionKeys[keyIndex].t && keyTime < positionKeys[keyIndex + 1].t)
                {
                    TimeValue offset = keyTime - positionKeys[keyIndex].t;
                    TimeValue distance = positionKeys[keyIndex + 1].t - positionKeys[keyIndex].t;
                    float percentage = float(offset) / float(distance);
                    key = Lerp(positionKeys[keyIndex].sampleKey.pval, positionKeys[keyIndex + 1].sampleKey.pval, percentage);
                    break;
                }
            }
        }
    }

    return key;
}

Quat Utilities::GetSampledRotationKey(const IGameKeyTab& rotationKeys, TimeValue keyTime)
{
    Quat key;
    
    if (rotationKeys.Count() > 0)
    {
        if (keyTime >= rotationKeys[rotationKeys.Count() - 1].t)
        {
            //The key time goes beyond the specified keys. Take the last key
            key = rotationKeys[rotationKeys.Count() - 1].sampleKey.qval;
        }
        else
        {
            //The key time is within range
            for (int rotationKeyIndex = 0; rotationKeyIndex < rotationKeys.Count(); rotationKeyIndex++)
            {
                if (keyTime >= rotationKeys[rotationKeyIndex].t && keyTime < rotationKeys[rotationKeyIndex + 1].t)
                {
                    TimeValue offset = keyTime - rotationKeys[rotationKeyIndex].t;
                    TimeValue distance = rotationKeys[rotationKeyIndex + 1].t - rotationKeys[rotationKeyIndex].t;
                    float percentage = float(offset) / float(distance);
                    key = Slerp(rotationKeys[rotationKeyIndex].sampleKey.qval, rotationKeys[rotationKeyIndex + 1].sampleKey.qval, percentage);
                    break;
                }
            }
        }
    }

    return key;
}

void Utilities::ExportUserData
    (
    TiXmlElement* parentXmlElement, 
    const KeyedValues& properties, 
    IGameNode* gameNode
    )
{
    TSTR propertyBuffer;
    gameNode->GetMaxNode()->GetUserPropBuffer(propertyBuffer);
    ExportUserData(parentXmlElement, properties, propertyBuffer);
}

void Utilities::ExportUserData
    (
    TiXmlElement* parentXmlElement,
    const KeyedValues& properties, 
    const TSTR& propertyBuffer
    )
{
    //Create userDataReference element if necessary
    TSTR userDataReference;
    if (properties.GetValue(PropertyKeys::USER_DATA_REFERENCE, userDataReference))
    {
        TiXmlElement* userDataReferenceElement = 
            (TiXmlElement*)parentXmlElement->LinkEndChild(new TiXmlElement(_T("userDataReference")));
        userDataReferenceElement->SetAttribute(_T("id"), userDataReference);
    }

    //Create userData element if necessary
    TSTR userDataMarker;
    if (properties.GetValue(PropertyKeys::USER_DATA_MARKER, userDataMarker))
    {
        TSTR userData;
        if (Utilities::ParseUserData(propertyBuffer, userDataMarker, userData))
        {
            //Get the property that determines if CDATA should be used to wrap user data
            bool isCDATA = true;
            TSTR isCDATAText;
            if (properties.GetValue(PropertyKeys::IS_USER_DATA_CDATA, isCDATAText))
                isCDATA = Utilities::ToBool(isCDATAText);

            //Add the user data text
            Utilities::AddTextElement(parentXmlElement, _T("userData"), userData, isCDATA);
        }
    }
}

void Utilities::ExportHardwareBufferUsage
    (
    TiXmlElement* parentXmlElement, 
    const KeyedValues& properties, 
    const TSTR& resourceName
    )
{
    TSTR usage;
    bool hasUsage = properties.GetValue(resourceName + PropertyKeys::USAGE, usage);

    TSTR useShadow;
    bool hasShadow = properties.GetValue(resourceName + PropertyKeys::USE_SHADOW, useShadow);

    if (hasUsage || hasShadow)
    {
        TiXmlElement* vertexIndexBufferXmlElement = 
            (TiXmlElement*)parentXmlElement->LinkEndChild(new TiXmlElement(resourceName.data()));

        if (hasUsage)
            vertexIndexBufferXmlElement->SetAttribute(_T("usage"), usage);
        if (hasShadow)
            vertexIndexBufferXmlElement->SetAttribute(_T("useShadow"), useShadow);
    }
}

bool Utilities::IsLightOrCamera(IGameObject* gameObject)
{
    return 
        gameObject->GetIGameType() == IGameObject::IGAME_LIGHT || 
        gameObject->GetIGameType() == IGameObject::IGAME_CAMERA;
}

GMatrix Utilities::Transpose(const GMatrix& matrix)
{
    GMatrix transpose;
    Point4 row;

    for (int rowIndex = 0; rowIndex < 4; rowIndex++)
    {
        row = matrix.GetRow(rowIndex); 
        transpose.SetColumn(rowIndex, row);
    }
        
    return transpose;
}

GMatrix Utilities::InverseTranspose(const GMatrix& worldMatrix)
{
    GMatrix inverse = const_cast<GMatrix&>(worldMatrix).Inverse();
    return Transpose(inverse);
}

bool Utilities::IsInteger(const TSTR& text)
{
    if (IsNull(text))
        return false;
    else
    {
        for (int charIndex = 0; charIndex < text.Length(); charIndex++)
        {
            if (!isdigit(text[charIndex]))
                return false;
        }

        return true;
    }
}

bool Utilities::IsPhysicsShapeMesh(const KeyedValues& properties)
{
    TSTR physicsShape;
    return
        !properties.GetValue(PropertyKeys::PHYSICS_SHAPE, physicsShape) ||
        IsSameNoCase(physicsShape, _T("mesh"));
}

NoteTrack* Utilities::GetAnimationNoteTrack(Animatable* anim)
{
    NoteTrack* noteTrack = NULL;
    if (anim != NULL)
    {
        int noteTrackCount = anim->NumNoteTracks();    
        if (noteTrackCount > 0)
        {
            for (int trackIndex = 0; trackIndex < noteTrackCount; trackIndex++)
            {
                NoteTrack* theNoteTrack = anim->GetNoteTrack(trackIndex);
                if (HasAnimation(theNoteTrack))
                {
                    noteTrack = theNoteTrack;
                    break;
                }
            }
        }
    }
    return noteTrack;
}

bool Utilities::HasAnimation(NoteTrack* noteTrack)
{
    bool result = false;
    DefNoteTrack* defNoteTrack = (DefNoteTrack*)noteTrack;
    int keyCount = defNoteTrack->NumKeys();
    if (keyCount > 1 && keyCount % 2 == 0)
    {
        result = true;

        for (int keyIndex = 0; keyIndex < keyCount; keyIndex += 2)
        {
            KeyedValues properties(defNoteTrack->keys[keyIndex]->note);
            if (!properties.HasValue(_T("animationName")))
            {
                result = false;
                break;
            }
        }
    }
    return result;
}

void Utilities::KeepKeysInInterval(IGameKeyTab& keys, Interval interval)
{
    for (int keyIndex = 0; keyIndex < keys.Count();)
    {
        if (keys[keyIndex].t < interval.Start() || keys[keyIndex].t > interval.End())
            keys.Delete(keyIndex, 1);
        else
            keyIndex++;
    }
}

void Utilities::GetTimesInInterval(std::vector<TimeValue>& outTimes, const std::vector<TimeValue>& inTimes, Interval interval)
{
    for (size_t keyIndex = 0; keyIndex < inTimes.size(); keyIndex++)
    {
        if (inTimes[keyIndex] >= interval.Start())
        {
            for (size_t j = keyIndex; j < inTimes.size() && inTimes[j] <= interval.End(); j++)
                outTimes.push_back(inTimes[j]);

            break;
        }
    }
}

template <class T>
void AddKeyTimes(IKeyControl* ikeys, std::vector<TimeValue>& keyTimes)
{
    T key;
    int keyCount = ikeys->GetNumKeys();
    for (int keyIndex = 0; keyIndex < keyCount; keyIndex++)
    {        
        ikeys->GetKey(keyIndex, &key);
        if (std::find(keyTimes.begin(), keyTimes.end(), key.time) == keyTimes.end())
            keyTimes.push_back(key.time);
    }
}

void Utilities::GetFloatKeyTimes(Control* control, std::vector<TimeValue>& keyTimes)
{
    if (control != NULL)
    {
        IKeyControl* ikeys = GetKeyControlInterface(control);
        if (ikeys != NULL)
        {
            if (control->ClassID() == Class_ID(TCBINTERP_FLOAT_CLASS_ID, 0))
                AddKeyTimes<ITCBFloatKey>(ikeys, keyTimes);
            else if (control->ClassID() == Class_ID(HYBRIDINTERP_FLOAT_CLASS_ID, 0))
                AddKeyTimes<IBezFloatKey>(ikeys, keyTimes);
            else if (control->ClassID() == Class_ID(LININTERP_FLOAT_CLASS_ID, 0))
                AddKeyTimes<ILinFloatKey>(ikeys, keyTimes);
        }
    }
}

static void SortAndRemoveDuplicates(std::vector<TimeValue>& keyTimes)
{
    std::vector<TimeValue> uniqueKeyTimes;
    uniqueKeyTimes.reserve(keyTimes.size());

    //Sort the times
    std::sort(keyTimes.begin(), keyTimes.end());

    //Remove duplicates
    for (size_t i = 0; i < keyTimes.size();)
    {
        uniqueKeyTimes.push_back(keyTimes[i]);

        size_t currentIndex = i++;        
        while (i < keyTimes.size() && keyTimes[i] == keyTimes[currentIndex])
            i++;        
    }

    //Return the unique times
    keyTimes.swap(uniqueKeyTimes);
}

void Utilities::GetPRSKeyTimes(IGameNode* gameNode, std::vector<TimeValue>& keyTimes, size_t flags)
{
    IGameControl* gameControl = gameNode->GetIGameControl();
    Control* maxControl = gameNode->GetMaxNode()->GetTMController();

    IGameKeyTab keys;
    
    if (gameControl != NULL)
    {
        if (flags & POSITION_KEY)
        {
            keys.ZeroCount();
            gameControl->GetQuickSampledKeys(keys, IGAME_POS);    
            for (int i = 0; i < keys.Count(); i++)
                keyTimes.push_back(keys[i].t);
        }

        if (flags & ROTATION_KEY)
        {
            keys.ZeroCount();
            gameControl->GetQuickSampledKeys(keys, IGAME_ROT);    
            for (int i = 0; i < keys.Count(); i++)
                keyTimes.push_back(keys[i].t);
        }

        if (flags & SCALE_KEY)
        {
            keys.ZeroCount();
            gameControl->GetQuickSampledKeys(keys, IGAME_SCALE);    
            for (int i = 0; i < keys.Count(); i++)
                keyTimes.push_back(keys[i].t);
        }
    }
    else if (maxControl != NULL && maxControl->ClassID() == Class_ID(PRS_CONTROL_CLASS_ID, 0))
    {
        IKeyControl* ikeys;
        int keyCount;

        //Position
        if (flags & POSITION_KEY)
        {
            Control* positionControl = maxControl->GetPositionController();        
            ikeys = GetKeyControlInterface(positionControl);
            keyCount = ikeys ? ikeys->GetNumKeys() : 0;
            if (keyCount > 1)
            {
                if (positionControl->ClassID() == Class_ID(TCBINTERP_POSITION_CLASS_ID, 0))
                    AddKeyTimes<ITCBPoint3Key>(ikeys, keyTimes);
                else if (positionControl->ClassID() == Class_ID(HYBRIDINTERP_POSITION_CLASS_ID, 0))
                    AddKeyTimes<IBezPoint3Key>(ikeys, keyTimes);
                else if (positionControl->ClassID() == Class_ID(LININTERP_POSITION_CLASS_ID, 0))
                    AddKeyTimes<ILinPoint3Key>(ikeys, keyTimes);
            }
        }

        //Rotation
        if (flags & ROTATION_KEY)
        {
            Control* rotationControl = maxControl->GetRotationController();
            ikeys = GetKeyControlInterface(rotationControl);
            keyCount = ikeys ? ikeys->GetNumKeys() : 0;
            if (keyCount > 1)
            {
                if (rotationControl->ClassID() == Class_ID(TCBINTERP_ROTATION_CLASS_ID, 0))
                    AddKeyTimes<ITCBRotKey>(ikeys, keyTimes);
                else if (rotationControl->ClassID() == Class_ID(HYBRIDINTERP_ROTATION_CLASS_ID, 0))
                    AddKeyTimes<IBezQuatKey>(ikeys, keyTimes);
                else if (rotationControl->ClassID() == Class_ID(LININTERP_ROTATION_CLASS_ID, 0))
                    AddKeyTimes<ILinRotKey>(ikeys, keyTimes);
            }
            else if (rotationControl->ClassID() == Class_ID(EULER_CONTROL_CLASS_ID, 0))
            {
                GetFloatKeyTimes(rotationControl->GetXController(), keyTimes);
                GetFloatKeyTimes(rotationControl->GetYController(), keyTimes);
                GetFloatKeyTimes(rotationControl->GetYController(), keyTimes);
            }
        }

        //Scale
        if (flags & SCALE_KEY)
        {
            Control* scaleControl = maxControl->GetScaleController();
            ikeys = GetKeyControlInterface(scaleControl);        
            keyCount = ikeys ? ikeys->GetNumKeys() : 0;
            if (keyCount > 1)
            {
                if (scaleControl->ClassID() == Class_ID(TCBINTERP_SCALE_CLASS_ID, 0))
                    AddKeyTimes<ITCBScaleKey>(ikeys, keyTimes);
                else if (scaleControl->ClassID() == Class_ID(HYBRIDINTERP_SCALE_CLASS_ID, 0))
                    AddKeyTimes<IBezScaleKey>(ikeys, keyTimes);
                else if (scaleControl->ClassID() == Class_ID(LININTERP_SCALE_CLASS_ID, 0))
                    AddKeyTimes<ILinScaleKey>(ikeys, keyTimes);
            }
        }
    }

    SortAndRemoveDuplicates(keyTimes);
}

IGameMorpher* Utilities::GetMorpher(IGameObject* gameObject)
{
	int modifierCount = gameObject->GetNumModifiers();
	for (int modifierIndex = modifierCount - 1; modifierIndex >= 0; --modifierIndex)
	{
		IGameModifier* modifier = gameObject->GetIGameModifier(modifierIndex);
		if (modifier != NULL)
		{
			if (modifier->GetModifierType() == IGameModifier::IGAME_MORPHER)
			{
				IGameMorpher* morpher = (IGameMorpher*)modifier;
				return morpher;
			}
		}
	}

    return NULL;
}

bool Utilities::IsZero(const Point3& v)
{
    return v.x == 0 && v.y == 0 && v.z == 0;
}

#define EPSILON 1.1929093e-5

bool Utilities::AlmostZero(float v)
{
    return fabs(v) < EPSILON;
}

bool Utilities::AlmostZero(const Point3& v)
{
    return AlmostZero(v.x) && AlmostZero(v.y) && AlmostZero(v.z);
}

void Utilities::GetSampledKeyTimes(std::vector<TimeValue>& keyTimes, float sampleInterval, float animationLength, TimeValue baseTime)
{
    //Calculate the number of sample intervals contained in the animation
    int sampleIntervalCount = (int)ceilf(animationLength / sampleInterval);
    
    //Recalculate the sample interval so that the samples are evenly spaced in time
    sampleInterval = animationLength / sampleIntervalCount;

    //Calculate the key times
    int sampleCount = sampleIntervalCount + 1;
    float keyTime = TicksToSec(baseTime);
    keyTimes.resize(sampleCount);
    for (int keyIndex = 0; keyIndex < sampleCount; keyIndex++)
    {
        keyTimes[keyIndex] = SecToTicks(keyTime);
        keyTime += sampleInterval;
    }
}

int Utilities::TicksToFrames(int ticks)
{
    return ticks / GetTicksPerFrame();
}

int Utilities::FramesToTicks(int frames)
{
    return frames * GetTicksPerFrame();
}

int Utilities::GetClosestVertexIndex(IGameMesh* gameMesh, const Point3& vertex, float* closestDistance)
{
    bool objectSpace = true;
    int closestIndex = -1;
    float closestDistanceSquared = FLT_MAX;
    float newDistanceSquared;

    for (int index = 0; index < gameMesh->GetNumberOfVerts(); index++)
    {
        newDistanceSquared = (gameMesh->GetVertex(index, objectSpace) - vertex).LengthSquared();
        if (newDistanceSquared < closestDistanceSquared)
        {
            closestDistanceSquared = newDistanceSquared;
            closestIndex = index;
        }
    }

    if (closestDistance != NULL)
        *closestDistance = sqrtf(closestDistanceSquared);

    return closestIndex;
}

void Utilities::ConvertRawVertex(Point3& p, UpAxis upAxis)
{
    if (upAxis == UP_AXIS_Y)
    {
        std::swap(p.y, p.z);
        p.z *= -1;
    }
}

void Utilities::GetVertices(Object* object, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals)
{
    if (object != NULL && object->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID, 0)))
    {
        TriObject* triObject = (TriObject*)object->ConvertToType(time, Class_ID(TRIOBJ_CLASS_ID, 0));
        Mesh& maxMesh = triObject->GetMesh();
        
        verts.resize(maxMesh.numVerts);
        if (upAxis == UP_AXIS_Y)
        {
            for (int vertexIndex = 0; vertexIndex < maxMesh.numVerts; vertexIndex++)
            {
                verts[vertexIndex] = maxMesh.verts[vertexIndex];

                std::swap(verts[vertexIndex].y, verts[vertexIndex].z);
                verts[vertexIndex].z *= -1;
            }

            if (normals != NULL)
            {
                normals->Create(maxMesh);

                for (int i = 0; i < normals->vertexNormals.Count(); i++)
                {
                    for (VNormal* vnormal = &normals->vertexNormals[i]; vnormal != NULL; vnormal = vnormal->next)
                    {
                        std::swap(vnormal->norm.y, vnormal->norm.z);
                        vnormal->norm.z *= -1;
                    }
                }

                for (int i = 0; i < normals->faceNormals.Count(); i++)
                {
                    std::swap(normals->faceNormals[i].y, normals->faceNormals[i].z);
                    normals->faceNormals[i].z *= -1;
                }
            }
        }
        else
        {
            for (int vertexIndex = 0; vertexIndex < maxMesh.numVerts; vertexIndex++)
                verts[vertexIndex] = maxMesh.verts[vertexIndex];

            if (normals != NULL)
                normals->Create(maxMesh);
        }

        if (triObject != object)
            triObject->DeleteThis();
    }
}

void Utilities::EvalMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals)
{
    const ObjectState& objectState = maxNode->EvalWorldState(time);    
    GetVertices(objectState.obj, time, upAxis, verts, normals);
}

void Utilities::EvalUnskinnedMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals)
{
    static const Class_ID morphClassIDs[] =
        {
        MAYBE_OLD_SKIN_CLASS_ID,
        CURRENT_SKIN_CLASS_ID, 
        Class_ID(PHYSIQUE_CLASS_ID_A, PHYSIQUE_CLASS_ID_B)
        };    
    static const int morphClassIDCount = COUNT_OF(morphClassIDs);

    Modifier* physique = FindPhysiqueModifier(maxNode);
    if (physique != NULL)
        SetFigureMode(physique, true);
    
    Object* maxObject = maxNode->GetObjectRef();
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        //Disable the morph modifiers
        std::vector<bool> oldStates;
        SetModifiersState(maxObject, morphClassIDs, morphClassIDCount, false, &oldStates);

        //Get verts
        EvalMeshVertices(maxNode, time, upAxis, verts, normals);

        //Restore skin enabled states
        SetModifiersStates(maxObject, morphClassIDs, morphClassIDCount, oldStates);
    }
    else
        EvalMeshVertices(maxNode, time, upAxis, verts, normals);
    
    if (physique != NULL)
        SetFigureMode(physique, false);
}

void Utilities::EvalUnmorphedMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals)
{
    static const Class_ID morphClassIDs[] =
        {
        MORPHER_MODIFIER_CLASS_ID,
        MAYBE_OLD_SKIN_CLASS_ID,
        CURRENT_SKIN_CLASS_ID,
        Class_ID(PHYSIQUE_CLASS_ID_A, PHYSIQUE_CLASS_ID_B)
        };    
    static const int morphClassIDCount = COUNT_OF(morphClassIDs);

    Modifier* physique = FindPhysiqueModifier(maxNode);

    if (physique != NULL)
        SetFigureMode(physique, true);
    
    Object* maxObject = maxNode->GetObjectRef();
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        //Disable the morph modifiers
        std::vector<bool> oldStates;
        SetModifiersState(maxObject, morphClassIDs, morphClassIDCount, false, &oldStates);

        //Get verts
        EvalMeshVertices(maxNode, time, upAxis, verts, normals);

        //Restore skin enabled states
        SetModifiersStates(maxObject, morphClassIDs, morphClassIDCount, oldStates);
    }
    else
        EvalMeshVertices(maxNode, time, upAxis, verts, normals);
    
    if (physique != NULL)
        SetFigureMode(physique, false);
}

void Utilities::SetModifiersState(Object* maxObject, const Class_ID* classIDs, int classIDCount, bool enabled, std::vector<bool>* oldStates)
{
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        IDerivedObject* maxDerivedObject = (IDerivedObject*)maxObject;    
    
        int modifierCount = maxDerivedObject->NumModifiers();
        for (int index = 0; index < modifierCount; index++)
        {
            Modifier* modifier = maxDerivedObject->GetModifier(index);
            if (ContainsClassID(classIDs, classIDCount, modifier->ClassID()))
            {
                if (oldStates != NULL)
                    oldStates->push_back(modifier->IsEnabled() ? true : false);

                if (enabled)
                    modifier->EnableMod();
                else
                    modifier->DisableMod();
            }
        }
    }
}

void Utilities::SetModifiersStates(Object* maxObject, const Class_ID* classIDs, int classIDCount, const std::vector<bool>& states)
{
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        IDerivedObject* maxDerivedObject = (IDerivedObject*)maxObject;    
        size_t stateIndex = 0;
    
        int modifierCount = maxDerivedObject->NumModifiers();
        for (int index = 0; index < modifierCount && stateIndex < states.size(); index++)
        {
            Modifier* modifier = maxDerivedObject->GetModifier(index);
            if (ContainsClassID(classIDs, classIDCount, modifier->ClassID()))
            {
                if (states[stateIndex++])
                    modifier->EnableMod();
                else
                    modifier->DisableMod();
            }
        }
    }
}

void Utilities::SetModifierState(Object* maxObject, Class_ID classID, bool enabled, std::vector<bool>* oldStates)
{
    SetModifiersState(maxObject, &classID, 1, enabled, oldStates);    
}

void Utilities::GetModifierStackState(Object* maxObject, std::vector<bool>& state)
{
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        IDerivedObject* maxDerivedObject = (IDerivedObject*)maxObject;    

        int modifierCount = maxDerivedObject->NumModifiers();
        state.resize(modifierCount);
        for (int index = 0; index < modifierCount; index++)
            state[index] = maxDerivedObject->GetModifier(index)->IsEnabled() ? true : false;
    }
}

void Utilities::SetModifierStackState(Object* maxObject, const std::vector<bool>& state)
{
    if (maxObject->SuperClassID() == GEN_DERIVOB_CLASS_ID)
    {
        IDerivedObject* maxDerivedObject = (IDerivedObject*)maxObject;    

        int modifierCount = maxDerivedObject->NumModifiers();
        modifierCount = std::min(modifierCount, (int)state.size());
        for (int index = 0; index < modifierCount; index++)
        {
            Modifier* modifier = maxDerivedObject->GetModifier(index);
            if (state[index])
                modifier->EnableMod();
            else
                modifier->DisableMod();
        }
    }
}

bool Utilities::ContainsClassID(const Class_ID* classIDs, int classIDCount, Class_ID id)
{
    for (int classIDIndex = 0; classIDIndex < classIDCount; classIDIndex++)
    {
        if (classIDs[classIDIndex] == id)
            return true;
    }

    return false;
}

bool Utilities::IsBone(INode* maxNode, IGameSkin* gameSkin)
{
    int boneCount = gameSkin->GetTotalSkinBoneCount();
    for (int boneIndex = 0; boneIndex < boneCount; boneIndex++)
    {
        if (gameSkin->GetBone(boneIndex) == maxNode)
            return true;
    }

    return false;
}

bool Utilities::IsBone(IGameNode* gameNode, IGameSkin* gameSkin)
{
    return IsBone(gameNode->GetMaxNode(), gameSkin);
}

Modifier* Utilities::FindModifier(INode* nodePtr, const Class_ID* cid, int cidCount)
{
    //Get object from node. Abort if no object.
	if (!nodePtr) 
        return NULL;

	Object* obj = nodePtr->GetObjectRef();

	if (!obj) 
        return NULL;
	
	while (obj->SuperClassID() == GEN_DERIVOB_CLASS_ID && obj)
	{
		IDerivedObject* derivedObj = (IDerivedObject*)(obj);				
			
		//Iterate over all entries of the modifier stack.
		int modStackIndex = 0;
		
        while (modStackIndex < derivedObj->NumModifiers())
		{
			//Get current modifier
			Modifier* modifier = derivedObj->GetModifier(modStackIndex);

            //Is this the right modifier
            for (int i = 0; i < cidCount; i++)
            {
			    if (modifier->ClassID() == cid[i])
				    return modifier;
            }
			modStackIndex++;
		}
		obj = derivedObj->GetObjRef();
	}

	//Not found.
	return NULL;
}

Modifier* Utilities::FindModifier(INode* nodePtr, Class_ID cid)
{
    Class_ID classIDs[1] = {cid};
    return FindModifier(nodePtr, classIDs, 1);
}

Modifier* Utilities::FindPhysiqueModifier(INode* nodePtr)
{
    return FindModifier(nodePtr, Class_ID(PHYSIQUE_CLASS_ID_A, PHYSIQUE_CLASS_ID_B));
}

Modifier* Utilities::FindSkinModifier(INode* nodePtr)
{
    static const Class_ID skinClassIDs[] =
        {
        MAYBE_OLD_SKIN_CLASS_ID,
        CURRENT_SKIN_CLASS_ID 
        };    

    return FindModifier(nodePtr, skinClassIDs, COUNT_OF(skinClassIDs));
}

Modifier* Utilities::FindMorpherModifier(INode* nodePtr)
{
    return FindModifier(nodePtr, MORPHER_MODIFIER_CLASS_ID);
}

void Utilities::SetFigureMode(ReferenceTarget* mod, bool state)
{
	IPhysiqueExport* phyExport = (IPhysiqueExport*)mod->GetInterface(I_PHYINTERFACE);
    if (phyExport != NULL)
    {
        phyExport->SetInitialPose(state);            
        mod->ReleaseInterface(I_PHYINTERFACE, phyExport);
    }
}

void Utilities::ReformatVersion(TSTR& version)
{
    const size_t PART_COUNT = 4;

    //Parse version components
    int parts[PART_COUNT] = {0};
    StringSplitter splitter(version, _T(','));
    size_t itemCount = std::min(splitter.size(), PART_COUNT);
    for (size_t itemIndex = 0; itemIndex < itemCount; itemIndex++)
    {
        Utilities::TrimWhitespace(splitter[itemIndex]);
        parts[itemIndex] = ToInt(splitter[itemIndex]);
    }

    //Format new string
    if (parts[3] <= 1)
        version.printf("%d.%d.%d", parts[0], parts[1], parts[2]);
    else
        version.printf("%d.%d.%d.%d", parts[0], parts[1], parts[2], parts[3]);
}

float Utilities::GetAnimationScale(const TSTR& scaleText, const TSTR& lengthText, Interval animationInterval)
{
    float animationScale = 1;    
    if (!IsNull(scaleText))
    {
        if (IsSameNoCase(scaleText, _T("normalize")))
        {
            if (animationInterval.Duration() > 0)
                animationScale = 1 / TicksToSec(animationInterval.Duration());
        }
        else
            animationScale = ToFloat(scaleText);
    }
    else if (!IsNull(lengthText))
    {
        if (IsSameNoCase(scaleText, _T("normalize")))
        {
            if (animationInterval.Duration() > 0)
                animationScale = 1 / TicksToSec(animationInterval.Duration());
        }
        else
        {
            float animationLength = ToFloat(lengthText);
            if (animationLength > 0 && animationInterval.Duration() > 0)
                animationScale = animationLength / TicksToSec(animationInterval.Duration());
        }
    }
    if (animationScale <= 0)
        animationScale = 1;
    return animationScale;
}

float Utilities::GetAnimationScale(const KeyedValues& properties, Interval animationInterval)
{
    TSTR scaleText, lengthText;
    properties.GetValue(PropertyKeys::SCALE, scaleText);
    properties.GetValue(PropertyKeys::LENGTH, lengthText);
    return GetAnimationScale(scaleText, lengthText, animationInterval);    
}

bool Utilities::GetCopyFirstKeyToLast(const KeyedValues& properties)
{
    TSTR text;
    if (properties.GetValue(PropertyKeys::COPY_FIRST_KEY_TO_LAST, text))
        return ToBool(text);
    else
        return false;
}

bool Utilities::HasDescendant(INode* possibleAncestorNode, INode* possibleChild)
{
    for (INode* parent = possibleChild->GetParentNode(); parent != NULL && !parent->IsRootNode(); parent = parent->GetParentNode())
    {
        if (parent == possibleAncestorNode)
            return true;
    }

    return false;
}

bool Utilities::IsInSetOrDescendantOfAny(INode* node, const std::vector<INode*>& nodes)
{
    for (size_t i = 0; i < nodes.size(); i++)
    {
        if (nodes[i] == node || HasDescendant(nodes[i], node))
            return true;
    }

    return false;
}
        
bool Utilities::HasAncestorInSet(INode* node, const std::set<INode*>& nodes)
{
    for (INode* parent = node->GetParentNode(); parent != NULL && !parent->IsRootNode(); parent = parent->GetParentNode())
    {
        if (nodes.find(parent) != nodes.end())
            return true;
    }

    return false;
}

void Utilities::GetTopLevelNodes(std::vector<INode*>& nodes, bool selected)
{
    Interface* maxInterface = GetCOREInterface();
    INode* rootNode = maxInterface->GetRootNode();
    
    nodes.clear();

    if (selected)
    {
        int selectedNodeCount = maxInterface->GetSelNodeCount();

        //Add all the selected nodes to a set
        std::set<INode*> selectedNodes;
        for (int i = 0; i < selectedNodeCount; i++)
            selectedNodes.insert(maxInterface->GetSelNode(i));

        //Add only the topmost nodes to the output collection
        for (int i = 0; i < selectedNodeCount; i++)
        {
            INode* selectedNode = maxInterface->GetSelNode(i);
            if (!HasAncestorInSet(selectedNode, selectedNodes))
                nodes.push_back(selectedNode);
        }
    }
    else
    {
        //Add the topmost nodes to the output collection
        int topNodeCount = rootNode->NumberOfChildren();
        nodes.resize(topNodeCount);

        for (int i = 0; i < topNodeCount; i++)
            nodes[i] = rootNode->GetChildNode(i);
    }
}

bool Utilities::InitializeData(IGameObject* gameObject, IGameNode* gameNode)
{
    bool result = gameObject->InitializeData();
    return result;
}

void Utilities::GetMeshAnimationSampleIntervals(const KeyedValues& properties, float& vertexAnimationSampleInterval, float& skeletonAnimationSampleInterval)
{
    vertexAnimationSampleInterval = 0;
    skeletonAnimationSampleInterval = 0;
    
    TSTR vertexAnimationSampleIntervalText;
    if (properties.GetValue(PropertyKeys::VERTEX_ANIMATION_SAMPLE_INTERVAL, vertexAnimationSampleIntervalText))
        vertexAnimationSampleInterval = ToFloat(vertexAnimationSampleIntervalText);
    else if (properties.GetValue(PropertyKeys::VERTEX_ANIMATION_SAMPLE_RATE, vertexAnimationSampleIntervalText))
    {
        float vertexAnimationSampleRate = ToFloat(vertexAnimationSampleIntervalText);
        if (vertexAnimationSampleRate > 0)
            vertexAnimationSampleInterval = 1 / vertexAnimationSampleRate;
    }

    TSTR skeletonAnimationSampleIntervalText;
    if (properties.GetValue(PropertyKeys::SKELETON_ANIMATION_SAMPLE_INTERVAL, skeletonAnimationSampleIntervalText))
        skeletonAnimationSampleInterval = ToFloat(skeletonAnimationSampleIntervalText);
    else if (properties.GetValue(PropertyKeys::SKELETON_ANIMATION_SAMPLE_RATE, skeletonAnimationSampleIntervalText))
    {
        float skeletonAnimationSampleRate = ToFloat(skeletonAnimationSampleIntervalText);
        if (skeletonAnimationSampleRate > 0)
            skeletonAnimationSampleInterval = 1 / skeletonAnimationSampleRate;
    }

    TSTR animationSampleIntervalText;
    if (properties.GetValue(PropertyKeys::MESH_ANIMATION_SAMPLE_INTERVAL, animationSampleIntervalText))
    {
        vertexAnimationSampleInterval = ToFloat(animationSampleIntervalText);
        skeletonAnimationSampleInterval = vertexAnimationSampleInterval;
    }
    else if (properties.GetValue(PropertyKeys::MESH_ANIMATION_SAMPLE_RATE, animationSampleIntervalText))
    {
        float meshAnimationSampleRate = ToFloat(animationSampleIntervalText);
        if (meshAnimationSampleRate > 0)
        {
            vertexAnimationSampleInterval = 1 / meshAnimationSampleRate;
            skeletonAnimationSampleInterval = 1 / meshAnimationSampleRate;
        }
    }
}

float Utilities::GetAnimationSampleInterval(const KeyedValues& properties)
{
    float sampleInterval = 0;
        
    TSTR sampleIntervalText;
    if (properties.GetValue(PropertyKeys::SAMPLE_INTERVAL, sampleIntervalText))
        sampleInterval = ToFloat(sampleIntervalText);
    else if (properties.GetValue(PropertyKeys::SAMPLE_RATE, sampleIntervalText))
    {
        float sampleRate = ToFloat(sampleIntervalText);
        if (sampleRate > 0)
            sampleInterval = 1 / sampleRate;
    }

    return sampleInterval;
}

void Utilities::RemoveBipedNonuniformScale(INode* maxNode, bool remove)
{
    Control* c = maxNode->GetTMController();

    if ((c->ClassID() == BIPSLAVE_CONTROL_CLASS_ID) ||
        (c->ClassID() == BIPBODY_CONTROL_CLASS_ID) ||
        (c->ClassID() == FOOTPRINT_CLASS_ID))
    {
        //Get the Biped Export Interface from the controller 
        IBipedExport* BipIface = (IBipedExport *) c->GetInterface(I_BIPINTERFACE);

        //Remove the non uniform scale
        BipIface->RemoveNonUniformScale(remove ? TRUE : FALSE);
    }
}

GMatrix Utilities::GetUniformMatrix(const GMatrix& transform)
{   
    Matrix3 result3;
    result3.SetRotate(transform.Rotation());
    result3.SetTrans(transform.Translation());

    GMatrix result(result3);
    return result;
}
     
GMatrix Utilities::GetUniformRelativeMatrix(const GMatrix& transform, const GMatrix& parentTransform)
{          
    return GetUniformMatrix(transform) * GetUniformMatrix(parentTransform).Inverse();
}

void Utilities::Replace(TSTR& result, const TSTR& source, const TSTR& replace, const TSTR& replaceWith)
{
    int sourceLength = source.Length();
    if (sourceLength == 0)
    {
        result = _T("");
        return;
    }

    int replaceLength = replace.Length();
    int replaceWithLength = replaceWith.Length();

	int sourceOffset = 0;
    int resultOffset = 0;
	
    int maxLength = sourceLength;
    result.Resize(maxLength);

    TCHAR* foundAt = NULL;
    do
	{
		foundAt = (TCHAR*)_tcsstr(&source.data()[sourceOffset], replace.data());
		if (foundAt)
		{
			//This is the length of everything BEFORE the found sequence
			int lengthOfCopy = foundAt - &source.data()[sourceOffset];

			//Copy the stuff to result that's before the found sequence
            if (resultOffset + lengthOfCopy > maxLength)
            {
                maxLength = resultOffset + lengthOfCopy;
                maxLength *= 2;
                result.Resize(maxLength);
            }
			_tcsncpy(&result[resultOffset], &source.data()[sourceOffset], lengthOfCopy);			
			resultOffset += lengthOfCopy;

			//Now copy replaceWith to result 
            if (resultOffset + replaceWithLength > maxLength)
            {
                maxLength = resultOffset + replaceWithLength;
                maxLength *= 2;
                result.Resize(maxLength);
            }
			_tcscpy(&result.data()[resultOffset], replaceWith.data());
			resultOffset += replaceWithLength;

			//Skip past the copied characters and the matched sequence
			sourceOffset += lengthOfCopy + replaceLength;
		}
	}while (foundAt != NULL);

    //Copy the remaining stuff from the source to the result
    int remainingLength = (int)_tcslen(&source.data()[sourceOffset]);
    if (resultOffset + remainingLength > maxLength)
    {
        maxLength = resultOffset + remainingLength;        
        result.Resize(maxLength);
    }
	_tcscpy(&result.data()[resultOffset], &source.data()[sourceOffset]);
}

void Utilities::AppendKeyAndValue(TSTR& text, const TSTR& key, const TSTR& value, const TSTR& newline)
{
    text += key;
    text += _T("=");
    text += value;    
    text += newline;    
}

void Utilities::AppendPrefixedKeyAndValue(TSTR& text, const TSTR& prefix, const TSTR& key, const TSTR& value, const TSTR& newline)
{
    text += prefix;
    text += _T(" ");
    AppendKeyAndValue(text, key, value, newline);
}

void Utilities::AppendMarkedText(TSTR& text, const TSTR& marker, const TSTR& data, const TSTR& newline)
{        
    text += marker;
    text += newline;
    text += data;
    text += newline;
    text += marker;
    text += newline;
}

void Utilities::AppendNumberedItem(TSTR& text, const TSTR& key, int number, const TSTR& value, const TSTR& newline)
{
    text += key;
    text += ToString(number);
    text += _T("=");
    text += value;    
    text += newline;
}

int Utilities::ParseTrailingInt(const TSTR& text)
{
    int index = text.Length() - 1;
    while (index >= 0 && isdigit(text[index]))
        index--;

    if (index < 0)        
    {
        //The text is an integer
        index = 0;
    }
    else
    {
        //Get text after last non-digit character
        index++;
    }

    return _tstoi(&text.data()[index]);
}

int Utilities::GetValueIndex(const TSTR& value, int referenceValueCount, const TCHAR* referenceValues[], int defaultValue)
{
    for (int i = 0; i < referenceValueCount; i++)
    {
        if (Utilities::IsSameNoCase(value, referenceValues[i]))
            return i;
    }

    return defaultValue;
}
