#ifndef OgreMax_Utilities_INCLUDED
#define OgreMax_Utilities_INCLUDED


//Includes---------------------------------------------------------------------
#include "KeyedValues.h"
#include "ExportSettings.h"
#include "MeshNormals.h"


//Macros-----------------------------------------------------------------------
#define COUNT_OF(x) (sizeof(x)/sizeof((x)[0]))


//Classes----------------------------------------------------------------------
namespace OgreMax
{
    enum RotationType
    {
        ROTATION_QUATERNION,
        ROTATION_ANGLE_AXIS,
        ROTATION_EULER_ANGLE
    };

    enum KeyType
    {
        POSITION_KEY = 0x1,
        ROTATION_KEY = 0x2,
        SCALE_KEY = 0x4
    };

    struct GameNodeAndObject
    {
        GameNodeAndObject(IGameNode* gameNode, IGameObject* gameObject)
        {
            this->gameNode = gameNode;
            this->gameObject = gameObject;
        }

        IGameNode* gameNode;
        IGameObject* gameObject;
    };

    /**
     * A collection of various utility functions
     */
    class Utilities
    {
    public:
        static bool IsNull(const TSTR& text);

        static bool IsSeparator(TCHAR c);

        /** Ensures the specified path ends with a trailing path separator */
        static void EnsureTrailingPathSeparator(TSTR& path);

        static TSTR JoinPath(const TSTR& path1, const TSTR& path2);

        /** Removes all the relative components from the path. */
        static TSTR MakeFullPath(const TSTR& path);

        /** Parses the file name from the specified path. */
        static TSTR GetFileName(const TSTR& path);

        static int GetFileNameOffset(const TSTR& path);

        /** Parses the file name from the specified path. The file name contains no extension. */
        static TSTR GetFileNameNoExtension(const TSTR& path);

        /** Gets the file name's extension */
        static TSTR GetExtension(const TSTR& path, bool includeDot = true);

        /** Returns the specified path without the file extension, if any. */
        static TSTR RemoveExtension(const TSTR& path);

        static void MakeExtensionLowercase(TSTR& path);

        /** 
         * Extracts the drive and directories from the file name. 
         * The returned path has a trailing path separator
         */
        static TSTR GetDirectoryPath(const TSTR& fullFilePath);

        /** Creates all the directories and subdirectories specified by the path. */
        static void CreateDirectoryRecursive(const TSTR& path);

        /** Creates all the directories and subdirectories containing the file path. */
        static void CreateFileDirectoryRecursive(const TSTR& path);

        /**
         * Creates an an export directory
         * @param outputPath [out] - The resultant output path
         * @param defaultExportPath [in] - The default export path. This path is assumed to exist 
         * on the computer
         * @param subdirectory [in] - A subdirectory, which may be either absolute or 
         * relative to the defaultExportPath. If this is empty, the the default export path
         * is returned. If the subdirectory is not empty, it is joined with the default export path, 
         * and the resultant directory is created recursively
         */
        static TSTR CreateExportDirectory(const TSTR& defaultExportPath, const TSTR& subdirectory);

        /**
         * Performs a case-insensitive comparison of the two strings, returning
         * true if they are the same
         */
        static bool IsSameNoCase(const TSTR& s1, const TCHAR* s2);
        static bool IsSameNoCase(const TSTR& s1, const TSTR& s2);
        
        /** Performs a case-insensitive comparison to determine if s1 starts with s2 */
        static bool StartsWithNoCase(const TSTR& s1, const TSTR& s2);

        static int FindCharNoCase(const TSTR& s, TCHAR c);

        static int FindDivide(const TSTR& s);
        
        /**
         * Copies the specified substring into another string
         * @param result [out] - Receives the substring
         * @param start [in] - The string from which a substring is extracted. 
         * The extraction starts at the first character position
         * @param length [in] - The length of the substring to extract
         */
        static void CopySubstring(TSTR& result, const TCHAR* start, int length);

        /** Converts an integer to a string.*/
        static TSTR ToString(int value);

        /** Converts a DWORD to a string.*/
        static TSTR ToString(DWORD value);

        /** Converts a float to a string.*/
        static TSTR ToString(float value);
        static TSTR ToString(double value);

        static TSTR ToString(Color value);

        static TSTR ToString(Point3 value);

        /** 
         * Converts a string to a boolean value 
         * @param value [in] - The string to convert to a boolean. The string is case insensitive
         * @return If the input string is 'false', 'no' or '0', false is returned.
         * Otherwise true is returned
         */
        static bool ToBool(const TSTR& value);

        /** Converts a string to float value */
        static float ToFloat(const TSTR& value);

        /** Converts a string to an unsigned int value */
        static unsigned int ToUInt(const TSTR& value);

        /** Converts a string to an int value */
        static int ToInt(const TSTR& value);

        /**
         * Parses a 3-component vector from the specified text
         * @param text [in] - 3 numerical values, each separated by a comma
         * @return Returns a 3-component vector
         */
        static Point3 ToPoint3(const TSTR& text);

        /** Returns the average of the 3 components */
        static float Average(Point3 xyz);

        /** Returns half the average size of the dimensions of box */
        static float GetBoxHalfSize(const Box3& box);

        /** Gets/sets the property value of the specified property ID as a string */
        static TSTR GetPropertyString(int propertySet, PROPID propid);
        static void SetPropertyString(int propertySet, PROPID propid, const TSTR& text);

        /** Gets/sets the scene's summary information */
        static TSTR GetSceneSummaryInfo();
        static void SetSceneSummaryInfo(const TSTR& text);

        static void GetNodeUserProperties(INode* maxNode, KeyedValues& properties);
        static INode* GetNodeValue(IGameProperty* gameProperty);

        /** Trims whitespace from the start and end of the specified string */
        static void TrimWhitespace(TSTR& text);

        /** 
         * Determines if the specified character is a newline character
         * Newline characters are \r (0x0d) and \n (0x0a)
         */
        static bool IsNewline(TCHAR c);

        /**
         * Finds the index of the first character in the next new line of text.
         * @param text [in] - The text to search
         * @param start [in] - The start index at which searching begins
         * @return The index of the first character in the next new line. If no newline
         * could be found, the length of the text is returned.
         */
        static int FindNextLineStart(const TSTR& text, int start, bool skipExtraNewlines = true);

        /**
         * Parses a key=value pair from the line starting at 'start'
         * @param text [in] - The text from which the key=value pair is parsed
         * @param start [in] - The start index in the text
         * @param nextStart [out] - The index of the next line
         * @param key [out] - The parsed key, with its whitespace trimmed
         * @param value [out] - The parsed value, with its whitespace trimmed
         * @return Returns true if a key=value pair was found, false otherwise. Regardless,
         * the nextStart output parameter is always set.
         */
        static bool ParseKeyValueLine(const TSTR& text, int start, int& nextStart, TSTR& key, TSTR& value, bool skipExtraNewlines = true);
        static bool ParseKeyValueLine(const TSTR& text, TSTR& key, TSTR& value);

        static bool ParsePrefixedKeyValueLine(const TSTR& text, TSTR& key, TSTR& value, TSTR* prefix = NULL);

        /**
         * Parses from the specified text a section user data.
         * @param text [in] - The text from which the user data is parsed
         * @param userDataMarker [in] - The marker that indicates the start and end of the 
         * user data. The first occurrence of this marker is assumed to be the specification
         * of the marker itself and is ignored
         * @param userData [out] - The parsed user data
         * @return Returns true if the user data was found, false otherwise
         */
        static bool ParseUserData(const TSTR& text, const TSTR& userDataMarker, TSTR& userData);

        /**
         * Displays a message box with the specified message
         */
        static int ShowMessage(const TSTR& message, const TSTR& title = _T("Message"), UINT type = MB_OK);
        static int ShowMessage(int messageID, int titleID, UINT type = MB_OK);
        static void ShowMeshInfo(IGameObject* gameObject, IGameNode* gameNode, const TCHAR* header, UINT type = MB_OK);
        static void ShowClassID(Class_ID cid, const TSTR& title = _T(""), UINT type = MB_OK);
        
        static void ShowAbout();
        
        static void ShowNodeErrorMessage(IGameNode* gameNode, int stringResourceID);

        /** 
         * Adds to the specified parent node an XML element with RGB attributes
         * @param parentNode [in] - The node to which the new element is added as a child
         * @param elementName [in] - The name of the new element
         * @param color [in] - The color
         */
        static TiXmlElement* AddColorElement(TiXmlNode* parentNode, const TCHAR* elementName, Point3 color);
        static TiXmlElement* AddColorElement(TiXmlNode* parentNode, const TCHAR* elementName, Color color);

        /** 
         * Adds to the specified parent node an XML element with XYZ attributes
         * @param parentNode [in] - The node to which the new element is added as a child
         * @param elementName [in] - The name of the new element
         * @param xyz [in] - The XYZ coordinates
         */
        static TiXmlElement* AddXYZElement(TiXmlNode* element, const TCHAR* elementName, Point3 xyz);
        static TiXmlElement* AddXYZElement(TiXmlNode* element, const TCHAR* elementName, Point4 xyz);
        
        static void SetXYZAttributes(TiXmlElement* element, Point3 xyz);
        static void SetXYZAttributes(TiXmlElement* element, Point4 xyz);

        static TiXmlElement* AddFaceElement(TiXmlElement* parentNode, const TCHAR* elementName, FaceEx* face);

        /** 
         * Adds to the specified parent node an XML element with quaternion attributes 
         * @param parentNode [in] - The node to which the new element is added as a child
         * @param elementName [in] - The name of the new element
         * @param q [in] - The quaternion
         * @param rotationType [in] - The type of rotation to add
         */
        static TiXmlElement* AddRotationElement
            (
            TiXmlNode* parentNode, 
            const TCHAR* elementName, 
            Quat q, 
            RotationType rotationType = ROTATION_QUATERNION
            );

        static TiXmlElement* AddRayElement(TiXmlNode* parentNode, Point3 position, Point3 direction);

        /** 
         * Adds to the specified parent node an XML element that has the specified CDATA text as its only child
         * @param parentNode [in] - The node to which the new element is added as a child
         * @param elementName [in] - The name of the new element
         * @param text [in] - The text that is placed into a CDATA text element
         * @param isCDATA [in] - Indicates whether text should be wrapped in a CDATA element
         */
        static TiXmlElement* AddTextElement
            (
            TiXmlNode* parentNode, 
            const TCHAR* elementName, 
            const TSTR& text, 
            bool isCDATA
            );

        static TiXmlElement* AddClippingElement
            (
            TiXmlNode* parentNode,
            float nearClip,
            float farClip
            );

        static TSTR LookupContactGroupInteger(const TSTR& contactGroup, const KeyedValues& sceneProperties);

        static bool GetObjectContactGroupInteger(TSTR& contactGroup, const KeyedValues& objectProperties, const KeyedValues& sceneProperties);

        static TiXmlElement* AddOpalContactGroupSetupElement(TiXmlNode* parentNode, const TSTR& contactGroupSetup, const KeyedValues& sceneProperties);

        static TiXmlElement* AddOpalContactGroupElement(TiXmlNode* parentNode, const TSTR& contactGroup);

        static TiXmlElement* AddOpalShapeMaterialElement(TiXmlNode* parentNode, const KeyedValues& properties);

        static TiXmlElement* AddOpalAxisElement
            (
            TiXmlElement* jointXmlElement, 
            int axisNumber, 
            const GMatrix& transform, 
            IPropertyContainer* propertyContainer,
            const KeyedValues& properties,
            TimeValue timeValue
            );

        /**
         * Sets into the specified element the plane's values as attributes
         * @param element [in] - The element that receives the plane values as attributes
         * @param plane [in] - The plane to set into the element's attributes
         */
        static void SetPlaneAttributes(TiXmlElement* element, Point4 plane);

        /**
         * Gets the material's bitmap tiling values
         * @param gameMaterial [in] - The material. This can be null
         * @param timeValue [in] - The time at which to evaluate the tiling parameters
         * @param tiling [out] - The bitmap tiling values
         * @return Returns true if there was a bitmap material specified, and the tiling values
         * were successfully retrieved. Otherwise, false is returned.
         */
        static bool GetTiling(IGameMaterial* gameMaterial, TimeValue timeValue, Point2& tiling);

        static bool GetSphereSegmentCount(IGameObject* gameObject, const KeyedValues& properties, TimeValue timeValue, IPoint2& segmentCount);

        static bool GetPlaneMeshSegmentCount(IGameObject* gameObject, IPoint2& segmentCount);

        /**
         * Similar to the other GetTiling() method, but it only returns the X component of 
         * the tiling values
         */
        static bool GetTiling(IGameMaterial* gameMaterial, TimeValue timeValue, float& tiling);

        /**
         * Creates a plane from the specified triangle points
         */
        static Point4 CreatePlane(Point3 p0, Point3 p1, Point3 p2);

        /**
         * Creates a normal from the specified triangle points
         */
        static Point3 CreateNormal(Point3 p0, Point3 p1, Point3 p2);

        /** Determines if the specified node has a mirrored transform */
        static bool IsMirrored(IGameNode* gameNode);

        /** 
         * Fixes (removes) any mirroring transforms applied to the specified matrix.
         * This is used by the GetNodeTM functions
         */
        static void FixMirroring(GMatrix& matrix);

        /** Gets the specified node's transformation matrix */
        static GMatrix GetNodeTM(IGameNode* gameNode, TimeValue timeValue, bool relativeToParent);

        /** Gets the specified node's transformation matrix relative to another node */
        static GMatrix GetNodeTMRelativeTo(IGameNode* gameNode, IGameNode* parentNode, TimeValue timeValue);

        static GMatrix GetNodeTMRelativeToSelf(IGameNode* gameNode, TimeValue referenceTime, TimeValue currentTime);

        static TSTR MakeModelName(const TSTR& name);
        static TSTR MakeMaterialName(IGameMaterial* gameMaterial, const TSTR& defaultName);
        static TSTR MakeMaterialName(const TSTR& name);
        static TSTR MakeMaterialFileName(const TSTR& name);
        static TSTR MakeMeshFileName(const TSTR& name, bool isXml);
        static TSTR MakeSkeletonFileName(const TSTR& name, bool isXml);

        static bool CanExportMaterial(const TSTR& materialName, const KeyedValues& sceneProperties);
        static bool CanExportMesh(const TSTR& meshName, const KeyedValues& sceneProperties);
        static TSTR GetRenamedBitmapFileName(const TSTR& fileName, const KeyedValues& sceneProperties);

        static void GetNodeUserProperties(IGameNode* gameNode, KeyedValues& properties);        
        static TSTR GetMeshName(IGameNode* gameNode, const KeyedValues& properties);

        static bool GetMeshPlane(IGameObject* gameObject, Point4& plane, bool worldSpace);

        /**
         * Gets the mesh's local bounding box, as defined by 3DS Max (not IGame).
         * This is useful for getting the local dimensions of a mesh, not the
         * dimensions that might be transformed by IGame
         * @return Returns true if the bounding box was retrieved, false otherwise
         */
        static bool GetMeshLocalBoundingBox(IGameObject* gameObject, Box3& box);        

        /** Gets the mesh's local bounding sphere */
        static float GetBoundingSphereRadius(IGameMesh* gameMesh, Point3 axes = Point3(1,1,1));

        /** Gets the mesh's local "length". This is its size along its local Z axis */
        static float GetMeshLength(IGameMesh* gameMesh);

        /** Makes all of the point's components positive */
        static void Abs(Point3& p);

        /** Restricts the specified value to fall within [minValue, maxValue] */
        static float Limit(float value, float minValue, float maxValue);

        /** Interpolates the two points */
        static Point3 Lerp(Point3 p0, Point3 p1, float t);

        static Point3 GetSampledPoint3Key(const IGameKeyTab& positionKeys, TimeValue keyTime, Point3 defaultValue = Point3(0,0,0));

        static Quat GetSampledRotationKey(const IGameKeyTab& rotationKeys, TimeValue keyTime);

        static void ExportUserData
            (
            TiXmlElement* parentXmlElement, 
            const KeyedValues& properties, 
            IGameNode* gameNode
            );
        static void ExportUserData
            (
            TiXmlElement* parentXmlElement, 
            const KeyedValues& properties, 
            const TSTR& propertyBuffer
            );

        static void ExportHardwareBufferUsage
            (
            TiXmlElement* parentXmlElement, 
            const KeyedValues& properties, 
            const TSTR& resourceName
            );

        template <class T>
        static void DeleteAll(T& collection)
        {
            for (T::iterator item = collection.begin(); item != collection.end(); ++item)
                delete *item;
            collection.clear();
        }

        static bool IsLightOrCamera(IGameObject* gameObject);

        static GMatrix Transpose(const GMatrix& matrix);
        static GMatrix InverseTranspose(const GMatrix& worldMatrix);

        static bool IsInteger(const TSTR& text);

        static bool IsPhysicsShapeMesh(const KeyedValues& properties);

        static NoteTrack* GetAnimationNoteTrack(Animatable* anim);
        static bool HasAnimation(NoteTrack* noteTrack);

        static void KeepKeysInInterval(IGameKeyTab& keys, Interval interval);
        static void GetTimesInInterval(std::vector<TimeValue>& outTimes, const std::vector<TimeValue>& inTimes, Interval interval);
        
        static void GetFloatKeyTimes(Control* control, std::vector<TimeValue>& keyTimes);

        static void GetPRSKeyTimes(IGameNode* gameNode, std::vector<TimeValue>& keyTimes, size_t flags = POSITION_KEY | ROTATION_KEY | SCALE_KEY);

        static IGameMorpher* GetMorpher(IGameObject* gameObject);

        static bool IsZero(const Point3& v);
        static bool AlmostZero(float v);
        static bool AlmostZero(const Point3& v);

        static void GetSampledKeyTimes(std::vector<TimeValue>& keyTimes, float sampleInterval, float animationLength, TimeValue baseTime);

        static int TicksToFrames(int ticks);
        static int FramesToTicks(int frames);

        static int GetClosestVertexIndex(IGameMesh* gameMesh, const Point3& vertex, float* closestDistance = NULL);
        static void ConvertRawVertex(Point3& p, UpAxis upAxis);
        static void GetVertices(Object* object, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals = NULL);

        static void EvalMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals = NULL);
        static void EvalUnskinnedMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals = NULL);
        static void EvalUnmorphedMeshVertices(INode* maxNode, TimeValue time, UpAxis upAxis, std::vector<Point3>& verts, MeshNormals* normals = NULL);

        static void SetModifiersState(Object* maxObject, const Class_ID* classIDs, int classIDCount, bool enabled, std::vector<bool>* oldStates = NULL);
        
        static void SetModifiersStates(Object* maxObject, const Class_ID* classIDs, int classIDCount, const std::vector<bool>& states);
        
        static void SetModifierState(Object* maxObject, Class_ID classID, bool enabled, std::vector<bool>* oldStates = NULL);
                
        static void GetModifierStackState(Object* maxObject, std::vector<bool>& state);
        static void SetModifierStackState(Object* maxObject, const std::vector<bool>& state);

        static bool ContainsClassID(const Class_ID* classIDs, int classIDCount, Class_ID id);

        static bool IsBone(INode* maxNode, IGameSkin* gameSkin);        
        static bool IsBone(IGameNode* gameNode, IGameSkin* gameSkin);

        static Modifier* FindModifier(INode* nodePtr, const Class_ID* cid, int cidCount);
        static Modifier* FindModifier(INode* nodePtr, Class_ID cid);
        static Modifier* FindPhysiqueModifier(INode* nodePtr);
        static Modifier* FindSkinModifier(INode* nodePtr);
        static Modifier* FindMorpherModifier(INode* nodePtr);
        static void SetFigureMode(ReferenceTarget* mod, bool state);        
        
        static TSTR GetVersionInfo(const TSTR& entry);
        static void ReformatVersion(TSTR& version);

        static float GetAnimationScale(const TSTR& scaleText, const TSTR& lengthText, Interval animationInterval);
        static float GetAnimationScale(const KeyedValues& properties, Interval animationInterval);

        static bool GetCopyFirstKeyToLast(const KeyedValues& properties);

        /** Determines if the ancestor node has the child node as a decendent */
        static bool HasDescendant(INode* possibleAncestorNode, INode* possibleChild);

        /** Determines if the specified node is in the list of nodes, or is a descendent of any of those nodes */
        static bool IsInSetOrDescendantOfAny(INode* node, const std::vector<INode*>& nodes);                            
        
        static bool HasAncestorInSet(INode* node, const std::set<INode*>& nodes);
        static void GetTopLevelNodes(std::vector<INode*>& nodes, bool selected);

        static Point3 ParseXYZMask(const TSTR& text, float foundValue, float notFoundValue);

        static bool InitializeData(IGameObject* gameObject, IGameNode* gameNode);

        static void GetMeshAnimationSampleIntervals(const KeyedValues& properties, float& vertexAnimationSampleInterval, float& skeletonAnimationSampleInterval);

        static float GetAnimationSampleInterval(const KeyedValues& properties);

        static void RemoveBipedNonuniformScale(INode* maxNode, bool remove);

        static GMatrix GetUniformMatrix(const GMatrix& transform);
        static GMatrix GetUniformRelativeMatrix(const GMatrix& transform, const GMatrix& parentTransform);

        static void Replace(TSTR& result, const TSTR& source, const TSTR& replace, const TSTR& replaceWith);

        static void AppendKeyAndValue(TSTR& text, const TSTR& key, const TSTR& value, const TSTR& newline);
        static void AppendPrefixedKeyAndValue(TSTR& text, const TSTR& prefix, const TSTR& key, const TSTR& value, const TSTR& newline);
        static void AppendMarkedText(TSTR& text, const TSTR& marker, const TSTR& data, const TSTR& newline);
        static void AppendNumberedItem(TSTR& text, const TSTR& key, int number, const TSTR& value, const TSTR& newline);
        
        static int ParseTrailingInt(const TSTR& text);

        static int GetValueIndex(const TSTR& value, int referenceValueCount, const TCHAR* referenceValues[], int defaultValue = 0);
        
    private:
        static bool IsValidMeshNameCharacter(TCHAR c);
        static void FixMeshName(TSTR& meshName);                
    };

    template <class T> 
    void PushFront(std::vector<T>& vect, const T& value)
    {
        if (vect.empty())
            vect.push_back(value);
        else
        {
            size_t length = vect.size();
            
            //Use push_back to move last item and increase size of array by one
            vect.push_back(vect[length - 1]);

            //Copy the other items as normal
            for (size_t i = length - 1; i >= 1; i--)
                vect[i] = vect[i - 1];

            //Put new item at front
            vect[0] = value;
        }
    }
}


#endif